
# fc.tcl, 17.9.98

# receiving is controlled:
# since nevot doesn't listen to a session anymore, after receiving a 
# "$sessionid receive 0" msg, we use "$sessionid talk 1 {mem1 mem2 ..}
# to get the semantics of "receivin non but {mem ..}. isc evaluated the talk 
# msg and sends
# 1. "sessionid/cname receive 0" msg to all members in an session
# with the exception of mem1, mem2, ... .
# 2. isc sends "sessionid/cname(mem) receive 1" to mem1, mem2, ../.
# Thus, the "session" msg receive" is tranformed to a set of "member" messages.

#*********** Global variables: ***************************
set msg "Welcome to Session!"

#************ address parameter **************************
# local pmm address
set pmm_controller {udp 225.2.0.1 5010 0}
set pmm_host [lindex $pmm_controller 1]
set pmm_port [lindex $pmm_controller 2]

#extern multicast address
#to be be set in commandline
set mc_controller {udp}
#otherwise the default one is used:
set def_mc_controller {udp 222.1.2.3 4400 191}

#*********** data structures *****************************
set session(id) "Conference XXX"
set session(cname) %u

#indicate if ifloor session is already established
set session(opened) 0
#index of actual topic
set session(topic_no) 0
set session(duplex) ""

#variables for moderated conferences
#default mode for conferencing is "conferencing", all participants have the
#same rigths
set session(role) "Equal"
set session(mode) "Conferencing"
set session(moderator) no
set session(heartbeater) " "
# default mode for speaker list is not hidden:
set session(hide) 0
set role "Equal"

set session(waiting_heartbeat) 0
set session(wait_heartbeat_started) 0
set session(update_req) 0
set msgList " "

set Footer(no_answer) 0
#************ VARIABLES SET AND TUNED BY USERS ! *************
set session(watch) 1

#verbose mode, switch on/off for fc control messages:
set session(logfile) "stdout"
# the speaker can switch off the micro for a while:
set session(pause) 10000
# duration a footer message is shown
set session(display) 5000
#variables for tuning ifloor msg exchange 
#number of replies for some messages to be sure they are heared everywhere:
set session(replies) 20
set session(replies_acc) 5
#delay, e.g. between repeated messages
#set session(delay) 500
#set session(beat) 5000
set session(delay) 500
set session(beat) 500
#delay, expected time, a msg needs to cross the net:
set session(net_delay) 1000
#max number of repetitions if no answer arrives:
set session(max_reply) 5
# delaying a "retouch" message according to random number; rand. max.:
set session(delay_touch) 10
# floor acceptance must be confirmed within that delay:
set session(wait_acc) 10000
# floor confirmation from remote member must be received within:

set session(wait_floorconf) 10000
# delay between two requests for "accept" floor
#set session(floorLimit) 10000
set session(floorLimit) 100000

#*
#puts stdout "fc starting1"



#********** Error messages *********************************
proc fc_message { msg } {
    global session
    set w .fc_message
    if { [winfo exists $w] } {
	destsroy $w
    }
    toplevel $w
    wm title $w "Ifloor Error Message"
    frame $w.head
    
    after $session(display) " if [winfo exists $w] {destroy $w}"
    pack \
	$w.head \
	[label $w.l -text $msg] \
	[button $w.b -text "Close" -command "destroy $w"] \
	-side top
}


#********** connection establishment *************************
#establishment of extern connections address settings

proc init_mc_udp {} {
    global mcfile mc_controller
  set host [lindex $mc_controller 1]
  set port  [lindex $mc_controller 2]
  set mcfile [udp $host $port]
    Tcl_SetSockBuffer $mcfile
    fconfigure $mcfile -ttl [lindex $mc_controller 3]
    fconfigure $mcfile -buffering none
    fconfigure $mcfile -member 1 
#    set conf [fconfigure $mcfile]
#    tracing "fc -- init_mc: config = $conf"
    fileevent  $mcfile readable {fc_read_mc $mcfile}

}

#puts stdout "fcmem2"

proc init_pmm_udp {} {
    global cfile pmm_controller
    set host [lindex $pmm_controller 1]
    set port  [lindex $pmm_controller 2]
    set cfile [udp $host $port]
    Tcl_SetSockBuffer $cfile  
    fconfigure $cfile -ttl 0
    fconfigure $cfile -buffering none
    fconfigure $cfile -member 1
    fileevent $cfile readable {fc_read_pmm $cfile}
}
#puts stdout "fcmem3"
# address parameters are changing:

proc ttl {} {	
    global mc_controller session    
    set l [llength $mc_controller]
    if { $l < 4 } {
        set mc_controller [lappend mc_controller $session(ttl)]
  	return
    } else {
	set mc_controller [lreplace $mc_controller 3 3 \
      			     $session(ttl)]
        return
    }
}

proc port {} {
    global mc_controller session    
    set mc_controller [lreplace $mc_controller 2 2 $session(port)]
    mk_sessname
    return 
}

proc host {} {
    global mc_controller session    
    set mc_controller [lreplace $mc_controller 1 1 $session(host)]
    mk_sessname
    return
}

#puts stdout "fcmem4"
#********** basic I/O-control for udp *************************	
proc fc_read_udp {sock} {
    set msg [gets $sock]
    return $msg
}

proc fc_send_udp_pmm {messg t} {
    global cfile session
    if [info exists cfile] {
	set c [lindex $messg 1]
	set trace 1
	switch $c {
	    receiving -
	    active -
	    !active -
	    alive -
	    !alive { set trace 0
	    }
	}
	if {$trace == 1} {	
#	    tracing "fc -- fc_send_udp_pmm:  $messg"
	}
	puts $cfile $messg
	return
    }
    # message is to be sent before connection establishment
    set msg " Cannot send pmm message, no connection established yet! "
    sh_msgwin "warning" $msg
}

#extern send to ifloor mc address
proc fc_send_udp_mc {messg t} {
    global mcfile session
    if [info exists mcfile] {
#	tracing "fc -- fc_sending_udp_mc: $mcfile $messg  "
       puts $mcfile $messg
    } else { 
	# menu button has been pushed before connection establishment
	set msg " Cannot send mc message, no connection established yet! "
	sh_msgwin "warning" $msg
    }
}
#puts stdout "fcmem5"

#*********** fc I/O-routines *********************************

proc fc_send {msg addr {to {}}} {
    global pmm_controller mc_controller
   set id ${addr}_controller
    fc_send_[lindex [set $id] 0]_${addr} $msg $to
}

#
# reading from ifloor-mc connection
# all messages are stored in msgList, messages in this list are not
# evaluated again, they are freed elsewhere to enable evaluation
# of same message (type, sender) again.
#
proc fc_read_mc {sock} {
    global mc_controller session msgList
    set msg [fc_read_[lindex $mc_controller 0] $sock]
    if { [string length $msg] < 0 } {
	set errmsg " fatale error, cannot read mc messages"
        sh_msgwin "error" $errmsg
        after 1000 i_close
    }
#    tracing "->fc_readmc $msg"
    #some messages will be evaluated frequently
    switch -- [lindex $msg 1] {
	heartbeat -
	requestfloor_conf -
	touch -
	get_speakers -
	hide -
	speakers -
#	moderator -
	speakerlend {
	    floor_cntrl $msg
	    return
	}
	requestfloor {
	    set mmsg [split $msg " "]
	    set mmsg [list [lindex $mmsg 0] [lindex $mmsg 1]]
	}
	moderator {
	    set mmsg [split $msg " "]
	    set mmsg  \
		[list [lindex $mmsg 0] [lindex $mmsg 1] [lindex $mmsg 2]]
	}
	default {
	    set mmsg $msg
	}
    }
    if {![info exists mmsg]} {
	set mmsg $msg
    }
    if {![info exists msgList]} {
        lappend msgList $mmsg
        floor_cntrl $msg
    } else {
        if {[lsearch -regexp $msgList "[set mmsg]( \t)*.*"] == -1} {
	    tracing "fc-->fc_read_mc: into Msglist: $msg"
            lappend msgList $msg
            floor_cntrl $msg        
        }
    }
}

proc fc_read_pmm {sock} {
    global pmm_controller errmsg session
    set msg [fc_read_[lindex $pmm_controller 0] $sock]
    if { [string length $msg] < 0 } {
        set errmsg " fatale error, cannot read pmm messages"
        sh_msgwin "error" $errmsg
        after 1000 i_close      
    }
    set c [lindex $msg 1]
    set trace 1
    switch $c {
        active -
        !active -
        alive -
        !alive { set trace 0
        }
    }
    if {$trace == 1} {
#	tracing "->fc_readpmm $msg"
    }
    cntrl_pmm $msg
}

proc open_history {} {
    #task: trace file may be opened here as well!!
    #file for history information is opened (change name here!!)
    global session history env
    if [info exists session(history)] { return }
    
    set  home $env(HOME)
    
    if {![file exists $home/.ifloor]} {
        tracing "fc --open_history: mkdir $home/.ifloor"
	set n [exec mkdir  $home/.ifloor]
	set n [exec chmod 664  $home/.ifloor]
    }
    set mchine [exec uname -n]
    set fn  $home/.ifloor/$session(id)_$mchine.history
    
    #tracing "fc-->open-history: opened $session(history)" 
    if [catch {open $fn w} session(history)] {
        fc_message \
	    " Cannot open file history:$fn and thus will not make a log file"
	unset session(history)
    } else {
	tracing "fc-->open-history: opened $session(history)" 
    }
}    
#puts stdout "fcmem6"

#************ options ********************************************
#task: check, if parameter are empty or wrong
proc cmd_line {argc argv}  {
    global myagent mc_controller def_mc_controller errmsg
    global session
    tracing " -- cmd_line - argv\[0\]: [lindex $argv 0], argc: $argc"
    for {set i 0} {$i < $argc} {incr i 1} {
	switch -regexp -- [lindex $argv $i] {
	    -C { # conference name
		set session(id) [lindex $argv [incr i 1]]
		tracing "fc, cmd_line --> opt: [lindex $argv $i], \ 
                                            confid: $session(id)"
	    }
	    -w { set session(watch) 1
	    }
	    -a { # myagent is set to agenttype and -name given in cmdline
		set agenttyp [lindex $argv [incr i 1]]
		switch -exact -- $agenttyp {
		    audio - 
 #                    wb -  
		    video {
			add_myagent $agenttype                            
			incr i 1} 
		    default { 
			sh_msgwin "error" "unknown media agent type:\
                                               $agenttyp"
		    }
                 }	          
	    }
	    -mc { # "mc_controller" is set to be the mc address for inter fc 
		# communications
	       set session(host) [lindex $argv [incr i 1]]
               set mc_controller [lappend mc_controller $session(host)]
               if {[expr $i + 1] < $argc} {
		   if {[lindex $argv [incr i 1]] != "-"} {
                       set mc_port [lindex $argv $i] 
                   } else {
                       set mc_port [lindex $def_mc_controller 2]
                   }
	       }
               set mc_controller [lappend mc_controller $mc_port]
	       #puts "fc, cmdline -->option:[lindex $argv $i], \
                  #        mc_controller; $mc_controller"
	   }
	    -r { # ifloor is configured to be in moderator modus or 
	        # behave as an ordinary conference participant
		set r [lindex $argv [incr i 1]]
	       set r [lindex $r 0]
		tracing "fc-->cmd_line: role == $r"
		if {[lsearch -exact {m o e} $r] == -1} {
		   set msg "wrong parameter for command line option \ 
        \"-r\", should be:\[-r \"m\"(oderator) | \"o\"(rdinary) | \"e\"(qual)]"
		    sh_msgwin "error" $msg
		    after 1000 i_close
	       }
		set session(role) $r
	    }
	    -halfduplex {
		set session(duplex) "halfduplex"
	    }
	    -fullduplex {
		set session(duplex) "fullduplex"
	    }
	    default { 
		set msg "usage: fc \[-C <conference name>\] \
                      \[-a <agenttype> <agentname>\] \[-mc <host port>\] \[-halfduplex\]\
		\[-r \"Moderator\" | \"Ordinary\" | \"Equal\"]"
		sh_msgwin "error"
		after 1000 i_close	       
tracing "fc, cmd_line -->argv: [lindex $argv $i],mc_controller: $mc_controller"
	    }
       }
    } 
    #if cmd-line values are omitted, default's can be set:
    # take default values if set 
    if { [llength $mc_controller] < 2} {
	set msg "Default multicast address will be used: $mc_controller"
	  set mc_controller $def_mc_controller
	sh_msgwin "warning" $msg
    }

    #create name for this session
    mk_sessname
}
#puts stdout "fcmem7"

proc mk_sessname {} {
    #set ifloor session
    global session mc_controller
    set mc_addr [join [list [lindex $mc_controller 1] \
                           [lindex $mc_controller 2]] : ]
    set session(session) [join [list $session(id) "ifloor" $mc_addr] / ]
    tracing "fc-->mk_sessnam: session(session): $session(session)"
}

#*********** register to isc ************************************

proc announcement {} {
global session
    set msg "agent fc attached [pid]"
    fc_send $msg "pmm"
}         

#*********** list of accepted messages **************************

#
# remove message of list of received messages, enable another evaluation
# of message of same type from same member
#
# proc RmMsg {cn cmd {par {}}} {
# global session
#     after 1000 "eraseMsg $cn $cmd $par"
# }


proc rmMsg {cn cmd {par {}}} {
global msgList session
    if {![info exists msgList]} { return }
 
    tracing "fc-->rmMsg: rm cn=$cn, cmd=$cmd, par=$par ?" 
    set msg "[string trimright  $session(id)/ifloor/.*/$cn]"
    set entry  "[set msg]( |t)*[set cmd].*[set par].*"
#clear:
    tracing "fc-->rmMsg: rm $entry ?"
    while {[set i \
     [lsearch -regexp $msgList "[set msg]( )*[set cmd].*[set par].*"]] >= 0} {
  	  tracing "fc-->rmMsg: removed $i. [lindex $msgList $i]"
 	  set msgList [lreplace $msgList $i $i]
    }
    while {[set i \
     [lsearch -regexp $msgList $entry]] >= 0} {
  	  tracing "fc-->rmMsg: removed $i. [lindex $msgList $i]"
 	  set msgList [lreplace $msgList $i $i]
    }

}
 
#*********** random number generator ****************************
proc randomInit {seed} {
global rand
      set rand(ia) 9301;
      set rand(ic) 49297
      set rand(im) 233280
      set rand(seed) $seed
}

proc random {} {
global rand
      set rand(seed) \
          [expr ($rand(seed)*$rand(ia) + $rand(ic)) % $rand(im)]
      return [expr $rand(seed)/double($rand(im))]
}

proc randomRange {} {
global session
     expr int([random]*$session(delay_touch))
}

#*********** control agent **************************************
#puts stdout "fcmem8"

proc agent_cntrl {cmd mode {mem ""}} {
global actagent session
#tracing "fc-->agent_cntrl: sending: $cmd $mode $mem"
    if [info exists actagent] {
	foreach s $actagent {
	    tracing "fc-->agent_cntrl: $s "	   
            if {[string compare $mem  ""] != 0} {
		set msg [list $s $cmd $mode $mem]
            } else {
		set msg [list $s $cmd $mode]
	    }	
	    fc_send $msg "pmm"
	}
    }
}

proc init_agents { } {
   agent_cntrl receive  -1
}

#********* agent lists and arrays *************************
# operations on agent lists and arrays of agentnames
#puts stdout "fcmem9"

proc add_actagent {s} {
#add media agent to list of agents actually under control
global actagent speakerl session   
    if {! [info exists actagent]} {
	lappend actagent $s
        tracing "fc-->add_agent1: controlling $s"
    } else { 
        set i [lsearch -exact $actagent $s]
        if {$i >= 0} {
            tracing "fc-->add_agent2: $s already under controll"
            return
	}
        tracing "fc-->add_agent3:controlling $s"
        lappend actagent $s
    }
    cntrl_speakers
}



proc add_myagent {agent} {
#add media agent to list of agent to be controlled when attached
global myagent
    if {! [info exists myagent]} {
	lappend myagent $agent
        tracing "fc-->add_myagent1:agent=$agent"
	return
    } 
    set i [lsearch -exact $myagent $agent]
    if {$i >= 0} {
        tracing "fc-->add_myagent2:agent=$agent"
	    return
    }
    tracing "fc-->add_myagent3:agent=$agent"
    lappend myagent $agent
}

proc rm_myagent {agent} {
global myagent
    if {![info exists myagent]} { return }
    set i [lsearch -exact $myagent $agent]
    tracing "fc-->rm_myagent1:agent=$agent"
    if {$i >= 0} {
	lreplace $myagent $i $i
	if {[llength $myagent] == 0} {
	    tracing "fc-->rm_myagent2: myagent is empty"
	    unset myagent
	}
    }
    set l [llength $myagent]
    for {set i 0} {$i < $l} {incr i } {
        set a [lindex $myagent $i]
        tracing "fc-->rm_myagent:$i $a"
    }
}  

proc rm_actagent {s} {
global actagent 
#one media session is removed
    if {![info exists actagent]} { return }
    set i [lsearch -exact $actagent $s]
    tracing "fc-->rm_actagent1:agent=$s, index = $i"
    if {$i >= 0} {
	set actagent [lreplace $actagent $i $i]
        set l [llength $actagent]
        tracing "fc-->rm_actagen2t: length = $l"
	if {$l == 0} {	   
	    tracing "fc-->rm_actagent3: actagent is empty"
	    unset actagent
	} else {
            for {set i 0} {$i < $l} {incr i } {
                set a [lindex $actagent $i]
                tracing "fc-->rm_actagent4:index=$i $a"
            }
        }
    }
#    cntrl_speaker
#session is left uncontrolled:
    fc_send [list $s send 0] "pmm"
}
 
proc rm_actagents {agent} {
global actagent
    if {![info exists actagent]} { return }
    set len [llength $actagent]
    set i 0
    while {$i < $len} {
        set s [lindex $actagent $i]
	set ss [split $s "/"]
   	set a [lindex $ss 1]
      	     if {$a == $agent} {
		 lreplace $actagent $i $i
                 #leave session uncontrolled
                 fc_send [list $s send 0] "pmm"                
	     } else {
		 incr i
	     }
	if {[llength $actagent] == 0} {	   
	    tracing "fc-->rm_actagent2: actagent is empty"
	    unset actagent 
	}
    }
}


#********* speaker list *************************************

proc compose_sl {i mem stamp topic {stamp_wait "-1"}} {
# ifloor protocol: collect all accepted elements of speakerlist
global sl tmpsp_attr session
tracing "fc-->compose_sl:index: $i, mem: $mem, time: $stamp, topic: $topic" 
    set sl($i) $mem
    set tmpsp_attr($mem,stamp) $stamp
    set tmpsp_attr($mem,topic) $topic
    set tmpsp_attr($mem,stamp_wait) $stamp_wait
}

proc sl_okay {len} {
#is received list of speakers complete?
global sl
    for {set i 0} {$i < $len} {incr i 1} {
	if {![info exists sl($i)]} {
	    return -1
	}
    }
    return 1
}

proc save_sl {} {
#received list of speakers is inserted in speakerlist
global sl tmpsp_attr 
    foreach i [array names sl] {
        insert_speaker $sl($i) $tmpsp_attr($sl($i),stamp) \
        $tmpsp_attr($sl($i),topic) $tmpsp_attr($sl($i),stamp_wait)
    }
 #   cntrl_speakers
 #   put_speakers
}

proc send_speakers {} {
global speakerl sp_attr session
#send speakerlist to conference member
    if {![info exists speakerl]} {
	set len 0
    } else {
	set len [llength $speakerl]
    }
    for {set i 0} {$i < $len} {incr i 1} {
	set mem [lindex $speakerl $i]
	set m [join [list $i $mem $sp_attr($mem,stamp) \
			 $sp_attr($mem,topic)] "/"]
	switch -exact $i {
	    0 -
	    1 {set m [join [list $m $sp_attr($mem,stamp_wait)] "/"]
	    }
	}
	set msg [list $session(session)/$session(cname) "speakers" $m]
	fc_send $msg "mc"
    }
    set msg [list $session(session)/$session(cname) "speakerlend" $i]
    fc_send $msg "mc"
    rmMsg ".*" "touch"
}


proc put_speakers {} {
global speakerl sp_attr session
    if {$session(hide) == 1} {
        if {![myself $session(moderator)]} {
            return
        }
    } 
    clear_speakerl
    if [info exists speakerl] {
	set length [llength $speakerl]
	for {set i 0} {$i < $length} {incr i 1} {
            set sp [lindex $speakerl $i] 
	    tracing "fc->put_sp:$i. $sp"
	    switch -exact -- $i {
		0 {
		    set txt [list [duration $sp_attr($sp,stamp_wait)] \
				 "min" "speak"]
		    flush_speaker $sp $txt $sp_attr($sp,topic) 1
		}
		1 {
		    set txt [list [duration $sp_attr($sp,stamp_wait)] \
			       "min" "wait"]
		    flush_speaker $sp $txt $sp_attr($sp,topic) 2
		}
		default {
		    set txt "                       "
		    flush_speaker  $sp $txt $sp_attr($sp,topic)
		}
	    }
	}
    highl_speaker
    }
}
#puts stdout "fcmem11"

#
# tim1 < tim2 => 1, else 0
proc before {{tim1 99:99:99:99} {tim2 99:99:99:99}} {
global session
  set tim1 [split $tim1 :]
  set tim2 [split $tim2 :]
#  tracing "fc --> before $tim1 $tim2 ?"
  for {set i 0} {$i <= 3} {incr i} {
      set lower [string compare [lindex $tim1 $i] [lindex $tim2 $i]]
      if {$lower == -1} {
###	  tracing "fc --> before:$lower"
        return 1
      } else {
	  if {$lower == 1} {
	      return -1   
	  }
      }
  }
    return 0
}

proc 1st_speaker {} {
global speakerl
    if [info exists speakerl] {
	set sp [lindex $speakerl 0]
	if {$sp == {}} {
	    set sp " "
	}
	return [lindex $speakerl 0]
    } else {
	return "no"
    }
}

proc speaker1st {mem} {
global speakerl
   if {![info exists speakerl]} { return 0 }
    set frst [lindex $speakerl 0]
#    tracing "fc_speaker1st: $frst"
    if {$frst == $mem} {
	    return 1
    } else { return -1
    }  
	
}

# indication if mem is in speaker list
proc inlist {mem} {
global speakerl
    if [info exists speakerl] {
	if {[lsearch -exact $speakerl  $mem] >= 0} {
	    return 1
	}
    }
    return -1
}
    

proc insert {mem time i} {
#insert a speaker according to timestamp
global speakerl errmsg sp_attr
#    tracing "fc -- insert, params: mem = $mem, 
#                                   time = $time, length = $i"
   tracing "fc -- insert, speakerlentry: \
                                   [lindex $speakerl [expr $i-1]] "
    set lower [before $sp_attr([lindex $speakerl [expr $i-1]],stamp) $time] 
    if {$lower == 0} {
#footer:      
    set_footer "Error-" "Same time stamps in two \"RequestFloor\" messages! "
   
#Same time stamps in two messages, now compare lexically on member name
	set lower [before [lindex $speakerl $i] $mem]
    }
    if {$lower} {
      set speakerl [linsert $speakerl $i $mem]
      return
    } else {
	  insert $mem $time [expr $i-1]
    }
}

proc insert1st {mem topic} {
# a speaker is inserted at first place. 
global session speakerl sp_attr
#insert member in speaker list at first place
    tracing "fc_insert1st "
    if [info exists speakerl] {	
	set i [llength speakerl] 
	for {set i [expr  - 1]} {$i > 0} {set i [expr $i - 1]} {
	set speakerl [linsert $speakerl [expr $i + 1] \
			  [lindex $speakerl $i]]
	}
	set speakerl [linsert $speakerl 0 $mem]
        #first is the second now and waits from now on
        set sp_attr([lindex $speakerl 1],stamp_wait) [mk_timestamp]
    } else {
	set speakerl $mem
    }
#speaker will not be interrupted anymore, even if someone else should be first:
    set sp_attr($mem,stamp) "0:0:0:0"
#speaker is speaking from now
    set sp_attr($mem,stamp_wait) [mk_timestamp]
    set sp_attr($mem,topic) $topic
}

proc insert_speaker {mem stamp topic {stamp_wait "-1"}} {
global speakerl sp_attr session
# if mem not already in list, insert mem according to "timestamp",
# so each speaker will be in the list only once
#    tracing "fc -->ins._speak"
    if [info exists speakerl] {
	set len [llength $speakerl]
#	tracing "fc -->ins._speak: speakerl length: $len"
	if {[lsearch -exact $speakerl  $mem] >= 0} { return }
	insert $mem $stamp [llength $speakerl] 
    } else {
#       tracing "fc -->ins._speak:insert1.= $mem "
       set speakerl $mem
    }
# save speaker attributes
    set sp_attr($mem,stamp) $stamp
    set sp_attr($mem,topic) $topic
	tracing "fc -->ins._speak:$mem, $topic"
    set sp_attr($mem,stamp_wait) [mk_timestamp]
}

proc rm_speaker {mem} {
global speakerl session sp_attr
#    tracing "fc_rm_speaker"
    if {![info exists speakerl]} { return }
    set i [lsearch $speakerl $mem]
    if {$i >= 0} {
	unset sp_attr($mem,stamp)
	unset sp_attr($mem,topic)
	unset sp_attr($mem,stamp_wait)
	
	tracing "fc-->rmspeaker: $i . == [lindex $speakerl $i]"
	if {[llength $speakerl] > 1 } {
	    set speakerl [ lreplace $speakerl $i $i ] 
	} else { 
	    unset speakerl 
	}
    }
}

#
#media agents are controlled according to actual speaker list
#
proc cntrl_speakers {} {
global speakerl session 
#init new session:

    if {![info exists speakerl]} { return }

#otherwise: receiving is controlled
    set first [1st_speaker]
    if {[myself $session(moderator)]} {
	tracing "fc-->cntrl_speakers: myself moderator"
         agent_cntrl send 1   
	if  {[myself $first]} {
	    agent_cntrl receive -1
	} else  {
	    agent_cntrl talk 1 $first
	    # prevent VAT in halfduplex mode from continuing to send
	    tracing "fc-->cntrl_speakers: CNTRL_SPEAKERS returning to sending mode for halfduplex"
	    if { "$session(duplex)" == "halfduplex" } {
		agent_cntrl send 0  
		tracing "fc-->cntrl_speakers: halfduplex mode"
	    }
	}
	return
    } 



    if {[myself $first]} {
	tracing "fc-->cntrl_speakers: myself speaking"   
        agent_cntrl send 1
#        agent_cntrl talk -1
	if {$session(mode) == "Moderated"} {
	    agent_cntrl talk 1 $session(moderator)
	} else {
	    agent_cntrl receive -1
	}
    } else {
	tracing "fc-->cntrl_speakers: myself listening"
        agent_cntrl send 0  
#        agent_cntrl talk 1
# receive non, but one
	if {$session(mode) == "Moderated"} {
	    agent_cntrl talk 1 "$session(moderator) $first"
	} else {	
	    agent_cntrl talk 1 $first
	} 
	# prevent VAT in halfduplex mode from continuing to send
	if { "$session(duplex)" == "halfduplex" } {
	    tracing "fc-->cntrl_speakers: halfduplex mode"
	    agent_cntrl send 0  
	}

    }


    put_speakers


}


#
# invoked whenever the actual speaker may change (releasefloor,..)
# the actual speaker must confirm the speaking permission, allothers are
# waiting for this confirmation and release if not accepted
# media are controlled
#
proc nextspeaker {} {
global speakerl session sp_attr Frame msgList 
# no speaker list: receiving is enabled from everyone
    tracing "fc_nextspeaker"
    if {![info exists speakerl]} {
	agent_cntrl receive 1
	if { "$session(duplex)" == "halfduplex" } {
	    agent_cntrl send 0
	} else {
	    agent_cntrl send 1
	}
	return
    }
#I am speaker?
    set mem [1st_speaker]
    set confirm 0
    if {[myself $mem]} {
	tracing "fc-->nextspeaker: myself speaking!"

# must be accepted by user; then control media, send confirmation;
# else release 
  	bell
	bell
	set msg "It is your turn to speak now. Will you accept?"
        set Frame(Req,state) disabled
        $Frame(Req,but) config -state $Frame(Req,state)
        Dialogs $msg "accFloor"
        tracing "fc-->nextspeaker: accfloor $Frame(Dialog,accFloor)"
#	unselect $mem
	enable "sel"
        if {$Frame(Dialog,accFloor) == 1} {
	    set confirm 1
            ch_srequest
            set Frame(Pause,state) normal
            $Frame(Pause,but) config -state $Frame(Pause,state)
	 
# periodically confirm floor
        } else {
	    tracing "nextspeaker: myself has not accepted, going to leave!"
            ReleaseFloor
	    return
        }      

    } else {
# does the actual member accept floor? otherwise return

	# regardless who is chosen to send, stop sending local audio when in halfduplex mode
	# prevent VAT in halfduplex mode from continuing to send
	if { "$session(duplex)" == "halfduplex" } {
	    tracing "fc-->cntrl_speakers: halfduplex mode"
	    agent_cntrl send 0  
	}


	foreach l $msgList {
	    tracing "fc-->nextspeaker: msgList = $l"
	}
	set session(wait_requestfloor_conf,$mem) 0
	wait_requestfloor_conf $mem
	tkwait variable session(wait_requestfloor_conf,$mem)
	switch -exact -- $session(wait_requestfloor_conf,$mem) {

	    0 {
		tracing "fc-->nextspeaker: got no confirmation from $mem"
		if [info exists session(selected,$mem)] {
		    bell
		   set_footer " " "The selected member $mem does not accept!!"
		# if mem doesn't exist in session yet it cannot release
		# so clean MSg list now to enable later selection of $mem
		    unselect $mem
		}
		if [info exists speakerl] {
		    releasefloor $mem
		}
		return
	    }
	    1 {
	tracing "fc-->nextspeaker: got confirmation from [1st_speaker]!!"
	
		if [info exists session(selected,$mem)] {
		    #		stopMsg "select" $mem
		    # rmMsg ".*" "select" $mem
		    enable "sel"
		    # unset session(selected,$mem)
		}
		# If we are still in sending mode, stop sending another participant is talking
		if { "$session(duplex)" == "halfduplex" } {
		    tracing "fc-->cntrl_speakers: halfduplex mode"
		    agent_cntrl send 0  
		}
		

	    }
	    2 {
		tracing "nextspeaker: stopped for $mem"
		 unselect $mem
		return
	    }
	}
    }

#speaker will not be interrupted, even if one other has older time stamp:
    set sp_attr([1st_speaker],stamp) "0:0:0:0"
    set sp_attr([1st_speaker],stamp_wait) [mk_timestamp]
    # media agents are controlled:
    cntrl_speakers
#a new history entry is created
   set msg \
      "\n $session(topic_no).[1st_speaker] \"$sp_attr([1st_speaker],topic)\""
    writeComment $msg
    incr session(topic_no)
    if {$confirm == 1} {
	Confirm "requestfloor"
    } 

    if {[myself $mem]} {
    } else {
	waitFloorLimit [1st_speaker]
    }
}
#puts stdout "fcmem12"

#******** some functions *******************************

proc myself {name} {
global session
    if {$session(cname) == $name} {
        return 1
    } else {
        return 0
    }
}

#******** ifloor protocol **********************************
#******** message repetitions, confirmations **********

proc start_heartbeat {} {
global session
    tracing "fc-->Heartbeat started by $session(cname)"
    stopMsg "panic"
    set session(heartbeater) $session(cname) 
    Heartbeat
   if [info exists session(starting_heartbeat)] {    
	unset session(starting_heartbeat)
    }
}

proc Heartbeat {} {
#moderator sends heartbeat messages
global session Id
     tracing "@@@Heartbeat@@@"
     if [info exists Id(send_heartbeat)] {
#	 puts "!! Id(send_heartbeat) = $Id(send_heartbeat)] !! still exists"
	 after cancel $Id(send_heartbeat)
     }
set msg [list $session(session)/$session(cname) heartbeat $session(moderator) \
	     [1st_speaker] $session(hide)]

    fc_send $msg "mc"
    set Id(send_heartbeat) [after $session(beat) Heartbeat]
#clear:
    tracing "****Heartbeat:Id = $Id(send_heartbeat)"
}

proc heartbeat {cn mod speaker hid} {
global session Id
    tracing "fc-->heartbeat"
#    set session(waiting_heartbeat) 1
    set waiting 0
# after recover: myself ready for sending heartbeat, but other does first?
    if {[info exists Id(starting_heartbeat)]} {
	after cancel $Id(starting_heartbeat)
    }
    if {[info exists session(starting_heartbeat)]} {
	unset session(starting_heartbeat)
    }
    set waiting 1
    
    stopMsg "moderator" "$cn"
    
    if {[myself $session(heartbeater)]} {
	if {![myself $cn]} {
	    if [myself $session(moderator)] {
		# the other one must stop!
		return
	    } else {
		if {$cn == $session(moderator)} {
		    tracing \
		    " heartbeat:!!!!foraign heartbeat message from moderator"
		    stop_heartbeat
		} else {

		    tracing " heartbeat:!!!!foraign heartbeat message"
		    stop_heartbeat
		    set session(panic) 1
		    recover
		    return
		}
	    }
	} else {
	   # [myself $session(heartbeater)] && [myself $cn]
	    return
	}
    }
  
    set session(heartbeater) $cn
    if [info exists Id(beat)] {
	after cancel $Id(beat)
#	puts "cancelling Id(beat) = $Id(beat)"
    }
    wait_heartbeat
    
    if {$mod == "no"} {
	if {$session(mode) == "Moderated"} {
	    equal
	}
    }
    if {$mod != $session(moderator)} {
	if {$mod != "no"} {
	    moderator "xx" $mod "00.00.00.00"
	}
    }
    if {$speaker != [1st_speaker]} {
	if {$speaker != "no"} {
	    select $speaker 0
	}
    }
    if {$hid != $session(hide)} {
	hide $hid
    }
    if {$waiting == 1} {
	start_wait_heartbeat
    }

}

proc stop_heartbeat {} {
global session Id
    tracing "fc-->stop_heartbeat"
    set session(heartbeater) no
    if [info exists Id(send_heartbeat)] {
	after cancel $Id(send_heartbeat)
    }
 #     if {[info exists Id(send_heartbeat)]} {
# 	after cancel $Id(send_heartbeat)
#    }
}

proc start_wait_heartbeat {} {
global session
#    set session(waiting_heartbeat) 1
    set session(wait_heartbeat_started) 1
    wait_heartbeat
}

proc wait_heartbeat {} {
global session Id
#    tracing "wait_heartbeat beginning"
    if [info exists Id(beat)] {
	after cancel $Id(beat)
    }
    set Id(beat) [after [expr 8*$session(net_delay)] \
		      no_heartbeat]
    tracing "****wait_heartbeat: Id = $Id(beat)"
}
    
proc  no_heartbeat {} {
   global session 
    tracing "No heartbeat messages!!!."
    if {![info exists session(lateComer)]} {
	set_footer " " \
	    "Warning- uncontrolled leaving of a conference member!?!"
    } else {
	unset session(lateComer)
    }
    if {$session(mode) == "Moderated"} {
	equal
    }
    releasefloor $session(heartbeater)
    stopMsg "moderator" $session(heartbeater)
    recover
}


proc stop_wait_heartbeat {} {
global session Id
    tracing "fc-->stop_wait_heartbeat"
    set session(wait_heartbeat_started) 0
    if {[info exists Id(beat)]} {
	after cancel $Id(beat)
    }
}


proc recover {} {
global session Id
    set session(starting_heartbeat) 1
    stop_wait_heartbeat
    set delay [randomRange]
    set delay [expr 500*$delay]
    tracing "fc-->touch: delay == $delay"
    set Id(starting_heartbeat) [after $delay "start_heartbeat"]
    if [info exists session(panic)] {
	Panic
	unset session(panic)
    }
}

proc Panic {{i 1}} {
global session
#    set session(send_panic,$session(cname)) 1
    set msg "$session(session)/$session(cname) panic"
    fc_send $msg "mc"
    if {$i < 5} {
	after $session(beat) "Panic [incr i]"
    }
}

#
# warning; the speaker list may be inconsistent. Allow the evaluation of
# (possibly older) requestfloor msgs to rebuild the list.
#
proc panic {} {
    set_footer "warning-" "panic, loss of heartbeat messages"
    rmMsg ".*" "requestfloor"
}


#
#set timer: if no alive msg arrives from speaker, s/he is released 
#
proc waitFloorLimit {mem} {
global session Id
    tracing "waitFloorLimit beginning $mem"
    if [info exists  Id(floorlimit)] {
#	puts "canceling  Id(floorlimit) = $Id(floorlimit)"
	after cancel $Id(floorlimit)
    }
#    set session(wait_floor_conf,$mem) 1
    set Id(floorlimit) [after [expr 2*$session(wait_floorconf)]\
                no_FloorLimit $mem]
    tracing "*****waitFloorLimit: Id = $Id(floorlimit)"
}

proc no_FloorLimit {mem} {
global session Id
    tracing "no_FloorLimit: no req_conf msg from $mem"
    set_footer " " "uncontrolled leaving of member $mem ?"
#    if [info exist session(wait_floor_conf,$mem)] 
#    unset session(wait_floor_conf,$mem)
    releasefloor $mem
}


proc stop_waitFloorL {mem} { 
global session Id
    tracing "stop_waitFloorLimit "
#   set session(wait_floor_conf,$mem) 2
    if {[info exists Id(floorlimit)]} {
	after cancel $Id(floorlimit)
    }
}
			    
proc alive {mem} {
global session
#    set session(wait_floor_conf,$mem) 1
}

proc Confirm {cmd} {
# confirms speeking permission (requestfloor_conf)
# no increasing delay between repeated messages
global session
    set cmd ${cmd}_conf
    tracing "fc-->Reply: $cmd"
     append name "send_" $cmd 
    set session($name,$session(cname)) 1
    set msg [list $session(session)/$session(cname) $cmd]
    Repeat $msg $session(cname) 4
}

proc setSendConf {n m v} {
global session
   set session($n,$m) $v
} 

#
# repeats a message until stopped by deleting "session(wait_" $c "_conf)" 
# variable.
# ?? assume n just for tests: unlimited number of repetitions!!
# increment i in each call to Repeat to increase delays between 
# msg up to factor d ( delay = ($n * $session(delay), with $n <= $d)
#
proc Repeat {msg mem {d 10} {i 1} {to "all"}} {
    global session Id
    #repeat message to be heart everywhere, if not confirmed

    set c [lindex  [split $msg "  "] 1]

    append name "send_" $c 
    if {[info exists session($name,$mem)]}  {
	tracing ">>>Repeat_conf: sending $msg once more"
	fc_send $msg "mc" $to
	set Id($name,$mem) [after [expr $i * $session(delay)] \
				    sendMsg [list $msg] $mem $d $i $to]
	tracing "******Repeat: Id =$Id($name,$mem)"
    }
}

proc sendMsg {msg mem d i to} {
global session Id
    set c [lindex  [split $msg "  "] 1]
    append name "send_" $c 
    if [info exists session($name,$mem)] {
	if {$session($name,$mem) == 0} {
	    unset session($name,$mem)
	    bell
	    tracing " ***********stopped:[list $msg] $mem "
	    return
	} else { 
#	    fc_send $msg "mc" $to
	    if {$i < $d} { set i [incr i] }
	    
	    Repeat $msg $mem $d $i $to
	}
    }
} 

proc stopMsg {cmd {mem "?"}} {
global session Id
# you cant use variables in proc arguments, so:
    if {$mem == "?"} {set mem $session(cname)}

    tracing "fc-->stopMsg: cmd=$cmd ; mem=$mem" 
    append name "send_" $cmd  
    if {$mem == "*"} {
	foreach e [array names session $name,*] {
	    	tracing "***************> stopMsges: $e"
	    if {[info exists Id($e)]} {
#		puts "canceling Id($e) = $Id($e)"
		after cancel $Id($e)
	    }
	}
	return 1
    } else {
	if {[info exists session($name,$mem)]}  {
	    tracing "***************> stopMsg: $name, $mem"
	    unset session($name,$mem) 
	    if {[info exists Id($name,$mem)]} {
#		puts " canceling Id($name,$mem) = $Id($name,$mem)"
		after cancel $Id($name,$mem)
	    }
	    return 1
	}
    }
    return 0
}

#puts stdout "fcmem13"  
  
#***********************************************************
proc i_open {} {
global session mc_controller
    set mmsg [list $session(session) "opened"]
    fc_send $mmsg "pmm"
    #initialization, all data are available just now!
    if {$session(opened) == 1} { return }
    set session(opened) 1
    init_mc_[lindex $mc_controller 0]
    init_agents
    open_history
    frame_init
    set mmsg [list $session(session)/$session(cname) "hello"]
    Repeat $mmsg $session(cname)	      
    set session(wait_touch) 1
    Touch $session(replies)
#    start_wait_heartbeat
    return
}

proc opened {} {
tracing "fc-->opened"
}

proc Touch {n} {
#sending "touch" till answer arrives, max. n-times
global session Id
    if  [info exists session(wait_touch)] {
        tracing "fc-->Touch: $n -times more"
	if {$n > 0} {
	    set mmsg [list $session(session)/$session(cname) "touch"]
	    fc_send $mmsg "mc"
	    set Id(Touch) [after $session(net_delay) Touch [expr $n - 1]]

	} else {
#	set d [expr $session(delay_touch) + 1]
	    set wait [expr $session(net_delay) ]
	    tracing "fc-->Touch: waiting $wait"
	    set Id(cancel_Touch) [after $wait cancel_Touch]
	}
    }
}

proc cancel_Touch {} {
#time-out, no "retouch" received
global session
      tracing "fc-->cancel_Touch"
      if [info exists session(wait_touch)] {
	  unset session(wait_touch)
      }
# assume there is no 'hearbeater" start sending heartbeat messages:
      tracing "cancel_Touch: firts member? start hartbeat"
#      recover
      if {![myself $session(heartbeater)]} {
	    set session(lateComer) 1
            start_wait_heartbeat
      }
}

proc touch {mem} {
global session requests speakerl Id
#my own message?
    tracing "fc-->touch1: "
    if {$mem == $session(cname) || [info exists session(send_wait)]} {
	tracing "fc-->touch:send_wait set"
    	return
    }
    hello $mem
#moderator shows up to new member
    if {$session(role) == "Moderator"} {
	Moderator_norep $session(cname) $mem
	set msg [list $session(session)/$session(cname) "hide" $session(hide)]
	fc_send $msg "mc"
#slaves doesn't answer at all
    } else {
	if {$session(role) == "Ordinary"} {
	    return
	}
    }
	
#any speakerlist to send?
#    if {![info exists speakerl]} {
#	tracing "fc-->touch: no speakerl!"
#	return}

#set random delay and wait
    if {![info exists session(wait_touch)]} {
	set session(send_retouch) 1
#	set requests($mem) 1
	set delay [randomRange]
	set delay [expr 100*$delay]
	tracing "fc-->touch: delay == $delay"
	set Id(Retouch) [after $delay "Retouch $mem"]
    }
}


proc Retouch {mem} {
#send "retouch", will be canceled, if other member sends the message first
global session 
 tracing "fc-->Retouch"
    if [info exists session(send_retouch)] {
	set mmsg [list $session(session)/$session(cname) "retouch"]
	tracing "fc-->Retouch: m=$mmsg" 
#	fc_send $mmsg "mc"
	Repeat $mmsg $mem
    }
}

proc retouch {mem} {
global session  Id
   if {$session(cname) == $mem} { return }
#I am going to "retouch", but another member does first
    if [info exists session(send_retouch)] {
       after cancel $Id(Retouch)
#       unset session(send_retouch)
	return
   }
#got answer from the "nearest" conference member, only the first "retouch"
#message is considered
    if {[info exists session(wait_touch)]} {
	after cancel $Id(Touch)
	unset session(wait_touch)
	set session(nearest) $mem
	set session(wait_sl) 1
        tracing "fc-->floorcntrl,retouch"
	Get_speakers $session(replies)
    }
}

proc Get_speakers {n} {
#ask a conference member to send a speakerlist
global session Id
  if {![info exists session(wait_sl)]} {
      tracing "fc-->Get_speakers: not waiting for sl"
      return
  }
  tracing "fc-->Get_speaker from $session(nearest); $n times more"
  if {$n < 0} {
      set_footer "Warning-" "Cannot get speakerlist, problems in mc connection?"
      bell
  } else {
      set mmsg \
     [list $session(session)/$session(cname) "get_speakers" \
	  $session(nearest)]
      fc_send $mmsg "mc"
      set Id(Get_speakers) [after $session(delay) Get_speakers [expr $n-1]]
  }
}

proc get_speakers {cn mem} {
global session Id
#a conference member asks me to send the speakerlist
    tracing "fc-->get_speakers"
    stopMsg "retouch"
    if {($mem == $session(cname))} {
	if [info exists Id(Retouch)] {
	    after cancel $Id(Retouch)
	}
#         if {$session(role) == "Moderated"} {
#             Moderator_norep $session(moderator)
#         }
#         if {$session(hide) == 1} {
#             Hide
#         } 
        
    send_speakers 
    }
}

proc speakers {mem sp} {
global session Id
# accepting an element of the speakerlist - if I am waiting for
    tracing "fc-->speakers: accepting"
    if {![info exists session(wait_sl)]} {
	tracing "fc-->speakers: wait_sl not exists"	
	return
    }
    if [info exists Id(Get_speakers)] {
	after cancel $Id(Get_speakers)
    }
    set speaker [split $sp "/"]
    compose_sl [lindex $speaker 0]  [lindex $speaker 1] \
    [lindex $speaker 2] [lindex $speaker 3] [lindex $speaker 4]

}

proc speakerlend {mem index} {
global session Id
    if {![info exists session(wait_sl)]} { return }
# the complete speakerlist is already accepted - if I am waiting for
    if [info exists Id(Get_speakers)] {
	after cancel $Id(Get_speakers)
    }
    unset session(wait_sl)
    rmMsg $mem "speakerlend"
    rmMsg $mem "speakers"
    if {$index == 0} {
	tracing "fc-->speakerlend: empty speaker list"
	if {$session(wait_heartbeat_started) == 0} {
	    set session(wait_heartbeat_started) 1
	    tracing "speakerlend: wait_heartbeat_starting"
	    set session(lateComer) 1
	    start_wait_heartbeat
	}

	return
    }
    if {[sl_okay $index] > 0} { 
	save_sl
	cntrl_speakers
	put_speakers
	if {$session(wait_heartbeat_started) == 0} {
	    set session(wait_heartbeat_started) 1
	    tracing "speakerlend: wait_heartbeat_starting"
	    set session(lateComer) 1
	    start_wait_heartbeat
	}
	if [info exists Id(floorlimit)] {
	    after cancel $Id(floorlimit)
	}
	waitFloorLimit [1st_speaker]
    } else {
#speakerlist is not complete, ask for another one 
	tracing "fc-->speakerlend: more Touches!"
	Touch $session(replies)
    }
}
#puts stdout "fcmem14"

#******** moderated session ********************************
#modul variables:
set role Equal 
set mode Conferencing

#
#moderator selects a session member for sending
#
proc Select {mem} {
global session Frame 
    if {[1st_speaker] == $mem} {
	set_footer " " "$mem is already talking!"
	return
    }
    foreach m [array names session selected,*] {
	set_footer " " "$session($m) exists!"
 	stopMsg "select" $session($m)
     }
    set session(selected,$mem) 1
    set  session(send_select,$mem) 1
    stopMsg "discard" $mem	
    set msg [list $session(session)/$session(cname) select $mem]
    disable "sel"
    Repeat $msg $mem 
}

#
# selected mem is inserted at 1st place in list. If it's myself,confirm.
# moderator waits for the confirmation
#
proc select {mem {conf 1}} {
global session sp_attr msgList Frame speakerl Id
    tracing "fc-->select: $mem"
    if {[1st_speaker] == $mem} { return } 

# enable another confirmation message of actual speaker
    if [info exists speakerl] {
	set old [1st_speaker]
	if {[myself $old] == 1} {
#	    stopMsg "requestfloor"
	    stopMsg "requestfloor_conf"
	    set Frame(Req,state) normal
	    if {[info exists Id(update_req)]} {
		after cancel $Id(update_req)
	    }
	    set session(update_req) 0
	    ch_request
	    disable "Pause"
	} else {
	    # don't wait for old confirmations anymore
	    stop_waitReqConf $old
	    stop_waitFloorL $old
	    rmMsg  $old "requestfloor_conf"
	    rmMsg  ".*" "select" $old
	}
    }
#a conference member could be selected not already in the speaker list.
#no topic is defined then. 
    if { ![info exists sp_attr($mem,topic)] } {
        set sp_attr($mem,topic) "?"
    }

    if [myself $mem] {
	stopMsg "releasefloor"
	stopMsg "requestfloor"
	stopMsg "srequest"
# change buttons to enable release floor if speaker was not in sp. list before
	set Frame(Req,state) disabled
	$Frame(Req,but) config -state $Frame(Rel,state)
	set Frame(Rel,state) normal
	$Frame(Rel,but) config -state $Frame(Rel,state)
	ch_request
    }
 tracing " select: topic = $sp_attr($mem,topic)"
    set top $sp_attr($mem,topic)
  
# allow evaluation of more messages:
    # enable "releasefloor" after a while to pretend evaluation of old message
    after $session(net_delay) "rmMsg $mem \"releasefloor\""
    rmMsg ".*" "discard" $mem
#    rmMsg $mem "discard_conf"
    rmMsg $mem "requestfloor_conf"
    rm_speaker $mem  
    insert1st $mem $top
    if [myself $session(moderator)] {
	set session(selected,$mem) $mem

    }
    foreach l $msgList {
	tracing "msglist: $l"
    }
    if {$conf == 1} {
	nextspeaker 
	put_speakers
    } else {
	cntrl_speakers
    }
}


#
# if I have selected a member, enable next "select" 
#
proc unselect {mem} {
global session
    if [info exists session(selected,$mem)] {
	stopMsg "select" $mem
	enable "sel"
	unset session(selected,$mem)
    }
    rmMsg ".*" "select" $mem
}

#
#moderator discards a member from speaker list
#
proc Discard {m} {
global session speakerl
    tracing "fc-->Discard:$m"
    if {![info exist speakerl]} { return }
    set i [lsearch -exact $speakerl $m]
    tracing "fc-->Discard: [lindex $speakerl 0]"
    if {$i < 0} {
	set_footer " " "The member $m is not in speaker list!"
	tracing "fc_Discard:i = $i, $m not in slist"
	return
    }
    unselect $m
    disable "dis"
    set session(send_discard,$m) 1
    set msg [list $session(session)/$session(cname) "discard" $m]   
    Repeat $msg $m 
#    Repeat_conf $msg $m 
}
 
proc discard {mem} {
global session Frame Id
    tracing "fc-->discard:$mem"
    
    if [myself $mem] {
        tracing "fc - discard: myself "
	if {[info exists Id(update_req)]} {
	    after cancel $Id(update_req)
	}
	set session(update_req) 0
	if {[info exists Id(floorlimit)]} {
	    after cancel $Id(floorlimit)
	}
#  set button states
	ch_srequest

	set Frame(Req,state) normal
	$Frame(Req,but) config -state $Frame(Req,state)
	set Frame(Rel,state) disabled 
	$Frame(Rel,but) config -state $Frame(Rel,state)
	set Frame(Pause,state) disabled 
	$Frame(Pause,but) config -state $Frame(Rel,state)

	stopMsg "requestfloor"
	stopMsg "requestfloor_conf"	
	stopMsg "srequest"
	stopMsg "releasefloor"
	set_footer " " "Your speaking permission is revoked by the moderator!"
    } else {
	stop_waitReqConf $mem
	stop_waitFloorL $mem
    }
    if {![myself $session(moderator)]} {
# enable another select/request msg from mem
	rmMsg ".*" "select" $mem 
    }
    rmMsg $mem "requestfloor"
#    rmMsg $mem "select"
    rmMsg ".*" "interrupt" $mem

    if [myself $session(moderator)] {

	enable "dis"
    }
    releasefloor $mem
}


proc Moderator {{m " "} {to "broad"}} {
#moderator announces herself or new one
global session
    tracing "fc--> Moderator"
    if {$m == " "} {
	set m $session(cname)
    }
  
    set session(send_moderator,$m) 1 
    set t [mk_timestamp]
    set msg [list $session(session)/$session(cname) "moderator" $m \
		 $t ]
    Repeat $msg $m
    moderator "xx" $m $t 
#    frame_init
}

proc Moderator_norep {{m " "} {to "broad"}} {
#part of heartbeat msg!!
global session
    tracing "fc--> Moderator norep"
    if {$m == " "} {
	set m $session(cname)
    }
    set msg [list $session(session)/$session(cname) "moderator" \
             $m [mk_timestamp]]
    fc_send $msg "mc"

 }
#
# new(?) moderator shows up
#
proc moderator {cn mem stamp} {
global session

 # moderator is still the same as before
    if {[string compare $mem $session(moderator)] == 0} { return }
    if {[myself $cn]} { return }
# variable ch_frame indicates, that the frame must change
    set ch_frame 0
#    set start_waitbeat 0
    set start_beat 0
    tracing "fc-->moderator:new = $mem"
    if [myself $session(heartbeater)] {
	stop_heartbeat
    } else {
	stop_wait_heartbeat
    }
# if moderator is first moderator, we must be sure, there was no
# older moderator message.
    tracing "fc-->moderator:1" 
    if [info exists session(frstMod)] {
        unset session(frstMod)
	set low [before $session(modStamp) $stamp]
	if {$low == 0} {
	    set low [before $session(moderator) $mem]
	}
        if {$low == -1} { 
	    tracing "moderator: older msg arrived from $cn"
	    tracing "moderator: actual = $session(modStamp), new = $stamp"
#	    return 
	} else { 
	    if {$cn != $session(moderator)} {
		tracing "moderator: unallowed moderator msg from $cn"
		return
	    }
	}
    }
    tracing "fc-->moderator:2" 
# conference changes to moderated mode. 
    if {$session(mode) == "Conferencing"} {
	tracing " moderator: conference changes to moderated mode"
        set session(frstMod) 1
        set session(mode) "Moderated"
	stopMsg "equal"
    }
    tracing "moderator:stopMsg moderator $session(moderator)"
  	stopMsg "moderator" $session(moderator)           
#I am loosing moderatorship! sending no heartbeat anymore!
    if [myself $session(moderator)] {
	tracing "moderator: I am loosing moderatorship!"
#        stop_heartbeat
	stopMsg "moderator" $session(moderator)
	stopMsg "moderator" 
	stopMsg "requestmod" "*"
        set session(role) "Ordinary"
        set ch_frame 1
    }

# remove older messages from leaving moderator from received message list:
    if {![info exists session(frstMod)]} {
	rmMsg ".*" "moderator" $session(moderator)
    }
#    rmMsg ".*" "moderator" $session(moderator)
# now set new moderator
    tracing "moderator: set session(moderator) $mem"
    set session(moderator) $mem
    set session(modStamp) $stamp
# enable cmds for this moderator
    rmMsg $session(moderator) "equal"
    rmMsg $session(moderator) "requestmod" 
 
# stop older rejects, if there are some:
#    stopMsg "rejectmod"

#Myselve the new moderator?
    if {$session(cname) == $mem} {
#task: role already set for Moderator ?
	tracing "moderator: Myselve the new moderator?"
	set session(role) "Moderator"

        set ch_frame 1
	stopMsg "acceptmod"
        set start_beat 1
    } else {
	tracing "fc-->moderator:_role = Ordinary"
	set session(role) "Ordinary"
        if {$ch_frame == 0} {
#I am still an "ordinary". Only the moderator label must change
            set_modLab "Conference is moderated by: $session(moderator)"
            sh_conf_modes
        }
    }
    cntrl_speakers
#if the role has changed, the interface looks different:
    if {$ch_frame == 1} {
        frameBuild
	enable "sel"
	enable "dis"
	enable "mod"
    }
    if {$start_beat == 1} {
	tracing "moderator:start_beat"
	stop_wait_heartbeat
	start_heartbeat
    } else { 
	if {$session(wait_heartbeat_started) == 0} {
	    start_wait_heartbeat
	}
    }
}

#
#moderator asks session member to accept moderatorship
#
proc RequestMod {to} {
global Frame session
#repeat, wait for confirmation
#    tracing "RequestMod:"
    set to $Frame(entry,param)
    if [myself $to] {
	set_footer " " "You are already the moderator!"
	return }
    rmMsg $session(cname) "requestmod"
    set session(send_requestmod,$session(cname)) 1
    set msg \
    [list $session(session)/$session(cname) "requestmod" $to [mk_timestamp]]
#    Repeat $msg
    disable "mod"
    Repeat $msg $session(cname) 1 10
}

proc requestmod {mem} {
global session Frame msgList
# starts window that accepts user decision
    foreach e $msgList {
	tracing "msgList = $e"
    }
    rmMsg "$mem" "acceptmod" 
    rmMsg "$mem" "rejectmod" 
    tracing "fc-->requestmod: request $mem"
    if [myself $session(moderator)] {
	wait_requestmod_conf $mem
	tracing "fc-->requestmod: end waiting"
	
    } else {
	stopMsg  "acceptmod"
	stopMsg  "rejectmod"
    
	if {[string compare $session(cname) $mem] == 0} {
	    bell
	    bell
	    Dialogs "Do you accept Moderatorship?" accMod
	    if {$Frame(Dialog,accMod) == 1} {
		Acceptmod
	    } else {
		Rejectmod
	    }
	}
    }

    tracing "fc-->requestmod:end"    
}


# moderator waits for confirmation; may be requested member is not conference
# member (anymore)?
# 
 proc wait_requestmod_conf {mem } {
 global session Frame Id
     tracing "wait_requestmod_conf:$mem"
     set Id(requestmod_noconf) [after \
	 $session(wait_floorconf) "requestmod_noconf $mem"]
#clear: 
     tracing "****wait_requestmod_conf: Id = $Id(requestmod_noconf) "
}

proc requestmod_noconf {mem} {
 global session Frame Footer
    stopMsg "requestmod" "*"
    tracing "requestmod_noconf: $mem does not answer to your request!"
    set_footer " " "$mem does not answer to your request!"
    set Footer(no_answer) 1
    stopMsg "moderator" 
    stopMsg "requestmod" 
    if [info exists Frame(mod,but)] {
	enable "mod"
    }
}
	

proc Acceptmod {} {
global session
    set session(send_acceptmod,$session(cname)) 1
    set msg [list $session(session)/$session(cname) "acceptmod"]
    Repeat $msg $session(cname)
}

proc acceptmod {mem} {
global session Id
    tracing "fc-->acceptmod: from $mem"
 #   rmMsg ".*" "requestmod" $mem
    if [myself $session(moderator)] { 
# moderatorship has been accepted, stop repeating request
	tracing "fc-->acceptmod: set session(wait_requestmod_conf) 1"
	set session(wait_requestmod_conf) 1
	if {[info exists Id(requestmod_noconf)]} {
	    after cancel $Id(requestmod_noconf)
	}
	stopMsg "moderator" 
	stopMsg "requestmod" 
	enable "mod"
	Moderator $mem "broad"
    }
}

proc Rejectmod {} {
global session
    set session(send_rejectmod,$session(cname)) 1
    set msg [list $session(session)/$session(cname) "rejectmod"]
# repeat "rejectmod' msg only 10 times, if not heard by the moderator
# he is not waiting for further one's anymore
    Repeat $msg $session(cname) 1 10
#    Repeat_conf $msg $session(cname)
}

proc rejectmod {mem} {
global session Id Footer 
    tracing "fc-->rejectmod from $mem"
#    rmMsg ".*" "requestmod" $mem
    if [myself $session(moderator)] { 
# moderatorship has been rejected, stop repeating request
	tracing "fc-->rejectmod: set session(wait_requestmod_conf) 0"
	set session(wait_requestmod_conf) 0
	if {[info exists Id(requestmod_noconf)]} {
	    after cancel  $Id(requestmod_noconf)
	}
	stopMsg "requestmod" "*"
	enable "mod"
	if {$Footer(no_answer) == 0} {
	    set msg "$mem has rejected your request for Moderatorship!"
	    set_footer "Message- " $msg
	} else {
	     set Footer(no_answer) 0
	} 
    }
# enable receiving of later requestmod message, but wait for a while
# otherwise the same message may be evaluated twice.
    after [expr 2* $session(net_delay)] "rmMsg \".*\" \"requestmod\" $mem"
}

proc Equal {} {
#change to conference mode
global session
    stopMsg "moderator"
    set session(send_equal,$session(cname)) 1
    set msg [list $session(session)/$session(cname) "equal"]
    Repeat $msg $session(cname)

}

#
# Conference mode changes to "conferencing"
# The frame must change
#
proc equal {{d "dummy"}} {
global session 
    tracing "fc-->equal"
# mode really changes?
    if {$session(role) == "Equal"} { return }

    set session(role) "Equal"
    if [info exists session(frstMod)] {
	unset session(frstMod)
    }
    set session(hide) 0
    set session(mode) "Conferencing"
    if [myself $session(moderator)] {
 #      stop_heartbeat
    } else {
#	stop_wait_heartbeat
    }
    stopMsg "moderator" ".*"
    rmMsg ".*" "moderator" $session(moderator)
    set session(moderator) "no"
    frameBuild
    cntrl_speakers
    put_speakers
}

proc Hide {{mem "broad"}} {
global session
    set session(send_hide,$session(cname)) 1
    set msg [list $session(session)/$session(cname) \
                 "hide" $session(hide)]
    Repeat $msg $session(cname)
}

proc hide {{h {}}} {
global session  .li
#whenever session(hide) is changed, the ifloor main window is reinitialized
    if {$h == {}} { return }

    if {$h != $session(hide)} {
    tracing "fc_hide: hide mode changed: hide = $h"
	rmMsg {} "hide" $h
	stopMsg "hide" $h
        set session(hide) $h
        if {$session(moderator) != $session(cname)} {
            if {$session(hide) == 1} {
                destroy .li
            } else {
                frameBuild
            }
        }
#?        put_speakers
    }  
}

proc bye {mem} {
global session Disable
    if [myself $mem] return
    set Disable($mem) 1
    if {[string compare $mem $session(moderator)] == 0} {
#	stop_wait_heartbeat
	set_footer " " "The moderator is leaving!"
#	recover
        equal
    }
  if {[string compare $mem $session(heartbeater)] == 0} {  
      stop_wait_heartbeat
      recover
  }

    releasefloor $mem
}
#puts stdout "fcmem15"

#******** timing control **********************************

proc mk_timestamp {} {
# Warning: uses "date", may need maintenance
# extracts "day hour minute second" from date in Greenwich-time
    set tim [split [exec \date -u] " "]
    tracing "fc -- mk_timstamp: time1 = $tim"
# removing blancs from "date"
    set i 0
    set time 00:00:00
    foreach num $tim {
	if {[string match {*:*:*} $num] } {
	    set time $num
	}
    }
    set time [string trim $time ,]
    set time [string trim $time ;]
    set time [string trim $time  ]
    
    set date [split $time ":"]
    if {[lindex $time 0] == "00"} {
	set time  [join [list "99" $time] ":"]
    } else {
	set time  [join [list "00" $time] ":"]
    }


   tracing "fc -- mk_timstamp: time = $time"
    return $time
}


proc duration {t1} {

    if {$t1 >= 0} {
        set t2 [mk_timestamp]
# number of seconds from time stamp        
        set t1sec [seconds  $t1]

        set t2sec [seconds $t2]
 
        set diffsec [expr $t2sec-$t1sec]
        if {$diffsec < 0} {
            set diffsec 0
        }

        set diff [stamp $diffsec]

        return $diff

    } else {
#for "empty" time
        return " "
    }
   
}

proc seconds {time} {
    set time [split $time ":"]
    set j 3

    for {set i 0} {$i < 4} {incr i} {
	set t($i) [lindex $time $j]
	set t($i) [string trimleft $t($i) 0]
	if {$t($i) == {}} {
	    set t($i) 0
#	    tracing "fc-->seconds: empty factor $t($i)" 
	}
	set j [expr $j - 1]
    }
#time to seconds:
    set m 1
    set tim $t(0)

    for {set i 1 } {$i < 3} {incr i} {
	set m [expr 60 * $m]
	set tim [expr $tim + [expr $t($i) * $m]]
#    tracing "fc-->seconds:for time, factor:$m, $tim"
    }
# date to houres to seconds:
    set m [expr 24 * $m]
    set tim [expr $tim + [expr $t(3) * $m]]
#    tracing "fc-->seconds:for hour, factor:$m, $tim"

    return $tim
}

#create time_stamp format out of amount of seconds
proc stamp {t} {
    set time $t
    for {set i 1} {$i< 4} {incr i} {
	set sec [expr $i * 60]
	set tim($i) [expr $time % 60]

	set time [expr $time / $sec]

    }
    set tim(4) $time
    for {set i 1} {$i<= 4} {incr i} {
	if {[string length $tim($i)] <= 1} {
	    set time 0
	    set tim($i) [append time $tim($i)]

	}
    }
    set timestamp [join [list $tim(4) $tim(3) $tim(2) $tim(1)] ":"]
    return $timestamp
}

#puts stdout "fcmem16"
	
#********** speaker control *********************************

proc hello {mem} {
global speakerl Disable
    tracing "hello: $mem"
    if [info exists Disable($mem)] {
	unset Disable($mem)
    }
    rmMsg $mem "."
    stopMsg ".*" $mem
    if [info exists speakerl] {
#	agent_cntrl talk 0 $mem
	cntrl_speakers
      }
}

proc RequestFloor {} {
#Warning : touching speakerl!
global session speakerl Frame
     tracing "fc --> RequestFloor"
 
if {[inlist $session(cname)] == 1} { return }
#     $Frame(Req,but) configure -background IndianRed
     if [info exists session(wait_touch)] {
	  tracing "fc ---> RequestFloor: wait_touch set"
	  tkwait variable session(wait_touch)
      }
      stopMsg "releasefloor"
      set session(send_requestfloor,$session(cname)) 1
      set msg [list $session(session)/$session(cname) \
                        "requestfloor" [mk_timestamp] $Frame(Req,entry)]
      Repeat $msg $session(cname) 
}

proc requestfloor {mem tim {rmsg "?"}} {
global session speakerl Frame
    tracing "fc --> requestfloor:topic = $rmsg"
#    $req(but) configure -background LightGrey
    if [myself $session(moderator)] {
	stopMsg "discard" $mem
    }
    insert_speaker $mem $tim $rmsg
#    rmMsg $mem "discard_conf"
    if [myself $mem] {
        set Frame(Rel,state) normal
        $Frame(Rel,but) config -state $Frame(Rel,state)
        ch_request
    } else {
#	alive $mem
    }
    if {$session(hide) == 1} {
        if {$session(moderator) != $session(cname)} {
            return
        }
    }
    rmMsg $mem "release"
    rmMsg ".*" "discard" $mem
    rmMsg $mem "requestfloor_conf"
    put_speakers
  
    if { [speaker1st $mem] > 0 } {
	    nextspeaker
    }
#    put_speakers
}

#
# every member must wait for confirmation of floor token; may be actual member 
# is not conference member (anymore)
# 
# proc wait_requestfloor_conf {mem} {
# global session Id
#     tracing "fc-->wait_requestfloor_conf from $mem"
#     set  session(wait_requestfloor_conf,$mem) 2
#     set Id(wait_requestfloor_conf) \
#      [after [expr 2*$session(wait_floorconf)] "set \
#                             session(send_requestfloor_conf,$mem) 0"]
#     tkwait variable session(wait_requestfloor_conf,$mem)
#     after cancel $Id(wait_requestfloor_conf)
#     if [info exists session(wait_requestfloor_conf,$mem)] {
# 	if {$session(wait_requestfloor_conf,$mem) == 0} {
# 	    tracing "fc-->wait_requestfloor_conf $mem timed-out or released"
# 	    unset session(wait_requestfloor_conf,$mem)
# 	    return 0
# 	} else {
# 	    tracing "fc-->wait_requestfloor: $mem set to 1"	
# 	    unset session(wait_requestfloor_conf,$mem)
# 	    return 1
# 	}
#     }
# }

#
# accepting confirmation message stops sending more request  messages, 
# but not the confirmation message itself
# confirmes a select message from the moderator as well.
#
# proc requestfloor_conf {m} {
# global session Id
#     tracing "fc-->requestfloor_conf: mem = $m"
#     if {![myself $m]} {
# #	alive $m
# 	if [info exists Id(wait_requestfloor_conf)] {
# 	    after cancel $Id(wait_requestfloor_conf)
# 	}
# 	if [info exists session(wait_requestfloor_conf,$m)] {
# 	    set session(wait_requestfloor_conf,$m) 1 
# 	    tracing "fc-->requestfloor_conf: "
# 	    tracing "  set session(wait_requestfloor_conf,$m) 1"
# 	}
#     } else {
# #	if [info exists session(confirming)] {
# 	    stopMsg "requestfloor"
# #	    updateReq
# #	}
#     }
# }

# #
# # accepting release message stops waiting for floor confirmation from this
# # member
# #
# proc requestfloor_noconf {m}  {
# global session Id
#     tracing "fc-->requestfloor_noconf: mem = $m"
#     if [info exists Id(wait_requestfloor_conf)] {
# 	after cancel $Id(wait_requestfloor_conf)
#     }
   
#     if [info exists session(wait_requestfloor_conf,$m)] {
# 	set session(wait_requestfloor_conf,$m) 0 
# 	tracing "fc-->requestfloor_noconf $m "
#     }
#     unselect $m
# }


proc wait_requestfloor_conf {m} {
 global session Id
    tracing "wait_requestfloor_conf $m"
  set Id(wait_floor_conf,$m) \
     [after $session(wait_floorconf) "requestfloor_noconf $m"]
    tracing "****wait_requestfloor_conf: Id = $Id(wait_floor_conf,$m)"
}

proc requestfloor_noconf {m}  {
 global session 
    tracing "fc-->requestfloor_noconf $m timed-out or released"
    set session(wait_requestfloor_conf,$m) 0
#    set session(wait_floor_conf,$m) 0
#    unselect $m
    stop_waitFloorL $m
    return 0
} 

proc stop_waitReqConf {mem} {
global session Id
    set session(wait_requestfloor_conf,$mem) 2
    if {[info exists Id(wait_floor_conf,$mem)]} {
	after cancel $Id(wait_floor_conf,$mem)
    }
    unselect $mem
}
	
#
# accepting confirmation message stops sending more request  messages, 
# but not the confirmation message itself
# confirmes a select message from the moderator as well.
#
proc requestfloor_conf {m} {
global session Id
    tracing "fc-->requestfloor_conf: mem = $m"
    if {![myself $m]} {
	alive $m
	set session(wait_requestfloor_conf,$m) 1
#	set session(wait_floor_conf,$m) 1
	if [speaker1st $m] {
	    if [info exists Id(floorlimit)] {
		after cancel $Id(floorlimit)
	    }
	    waitFloorLimit $m
	}
	tracing "fc-->requestfloor_conf: "
	tracing "  set session(wait_requestfloor_conf,$m) 1, now cancel"
#???, still valid?:

	if {[info exists Id(requestmod_noconf)]} {
	    after cancel $Id(requestmod_noconf)
	}
    } else {
	    stopMsg "requestfloor"
	    if {$session(update_req) == 0} {
#		start_updateReq
	}
    }
}


 
proc SRequest {} {
    global session speakerl Frame
    # stop older requests
    stopMsg "srequest"

    $Frame(Req,but) configure -background IndianRed -text $Frame(Req,text)
    set session(send_srequest,$session(cname)) 1
    set cmd "srequest"
    set msg [list $session(session)/$session(cname) $cmd [mk_timestamp]]
    tracing "fc->SRequest: $msg"
    Repeat $msg $session(cname)
}
  
proc srequest {mem} {
global session Frame 

    tracing "fc --> srequest"
    $Frame(Req,but) configure -background LightGrey
    if {$mem == $session(cname)} {
        set_footer " " "Your request has been raised!"
	return
    }
 
    if {[speaker1st $session(cname)] == 1} {
	bell
	bell
       set msg "$mem would like to talk NOW! Continue?"
	Dialogs $msg "accSreq"
        if {$Frame(Dialog,accSreq) == 0} {
            ReleaseFloor
        }
    } else {
	set msg  "$mem would like to talk NOW!"
	set_footer "Request: " $msg
    }
}
    
   
proc Pause {} { 
global Frame session Id
    set me [speaker1st $session(cname)]
    if { $me < 0 } { return }
    agent_cntrl send 0
    set Frame(Pause,text) "Resume"
    set Frame(Pause,cmd) Resume
   $Frame(Pause,but) configure -background IndianRed -text $Frame(Pause,text) \
	       -command $Frame(Pause,cmd)
    set Id(tim_p1) [after $session(pause) Pause1] 
}

proc Pause1 {} {
global Frame session Id
#footer:
    set_footer  " " "Please come on!"
    bell
    $Frame(Pause,but) flash
    set Id(tim_p2) [after $session(pause) Pause2]
}   

proc Pause2 {} {
global Frame 
#footer:
#  destroy_win
    set Frame(Pause,text) "Pause"
    set Frame(Pause,cmd) Pause
   $Frame(Pause,but) configure -background LightGrey -text $Frame(Pause,text) \
    -command $Frame(Pause,cmd)
    set_footer  " " "Long delay; speaking permission is passed"
    ReleaseFloor
}

proc Resume {} {
global Frame Id
    tracing "fc-->Resume"
    if {[info exists Id(tim_p1)]} {
	after cancel $Id(tim_p1)
    }
    if {[info exists Id(tim_p2)]} {   
	 after cancel $Id(tim_p2)
     }
    set Frame(Pause,text) "Pause"
    set Frame(Pause,cmd) Pause
   $Frame(Pause,but) configure -background LightGrey -text $Frame(Pause,text) \
    -command $Frame(Pause,cmd)
    cntrl_speakers
    clear_footer
}

proc ReleaseFloor {} {
global session speakerl Frame Id
    tracing "fc-->ReleaseFloor"

    stopMsg "srequest"
    stopMsg "request"
    stopMsg "request_conf"
    set mem $session(cname)
    
    if {![info exists speakerl]} { return }
    if {[lsearch -exact $speakerl $mem] >= 0} {
# set button states
	set Frame(Rel,state) disabled 
	$Frame(Rel,but) config -state $Frame(Rel,state)
	set Frame(Pause,state) disabled 
	$Frame(Pause,but) config -state $Frame(Rel,state)
	set Frame(Req,state) normal
	$Frame(Req,but) config -state $Frame(Req,state)
        tracing "fc - release: cname = $session(cname) "
	ch_srequest
	set session(send_releasefloor,$session(cname)) 1
        set msg [list $session(session)/$session(cname) "releasefloor"]
	Repeat $msg $mem
# 	if [info exists Id(update_req)] {
# 	    after cancel $Id(update_req)
# 	}
 	set session(update_req) 0
	if {[info exists Id(floorlimit)]} {
	    after cancel $Id(floorlimit)
	}
    }
} 

proc releasefloor {mem} {
global session Frame
    if {$mem == {} } { return }
    tracing "fc-->releasefloor:$mem"
    if [myself $session(moderator)] {
	unselect $mem
    } else { 
# enable another select/request msg from mem
	rmMsg ".*" "select" $mem 
    }
    rmMsg $mem "request"
    rmMsg $mem "request_conf"
#    rmMsg $mem "select_conf"
    rmMsg $mem "srequest" 
# stop msg repetition

    if [myself $mem] {
	stopMsg "requestfloor"
	stopMsg "requestfloor_conf"
	$Frame(Rel,but) config -state $Frame(Rel,state)
	set Frame(Pause,state) disabled 
	$Frame(Pause,but) config -state $Frame(Rel,state)
	set Frame(Req,state) normal
	$Frame(Req,but) config -state $Frame(Req,state)
    } else {
# stop waiting for floor confirmation from this member 
	stop_waitReqConf $mem
	stop_waitFloorL $mem
    }
    set f 0
    if {[speaker1st $mem]==1} { 
        tracing "fc-->releasefloor: 1st speaker $mem is leaving"
        set f 1}
    rm_speaker $mem
    tracing "fc-->releasefloor2"
    if {$f == 1} {
        nextspeaker
# RequestFloor button is redefined
#        ch_srequest
    }
    put_speakers 
}

#puts stdout "fcmem17"

#******** Interface ****************************************

#
# just for tests;clear
#
proc logvar {varname index op} {
global Frame session 
upvar $varname var 

 
}

proc frame_init {} {
global session Frame 
    set Frame(Footer,lab) " "
    set Frame(Footer,msg) " "   
    set Frame(Dialog,labtext) " "
    set Frame(Dialog,buttext1) "  "
    set Frame(Dialog,buttext2) "  "
    set Frame(Dialog,entrytext) " "
 
    set Frame(Dialog,msg) " "
    set Frame(Dialog,topic) " "
    set Frame(Dialog,entrystate) disabled
    set Frame(Dialog,but1state) disabled
    set Frame(Dialog,but2state) disabled
    set Frame(Req,text) "Request Floor"
    set Frame(Req,state) normal
    set Frame(Req,cmd) "activateRequest"
    set Frame(Pause,text) "Pause"
    set Frame(Pause,cmd) "Pause"
    set Frame(Pause,state) disabled
    set Frame(Rel,text) "Release Floor"
    set Frame(Rel,cmd) ReleaseFloor
    set Frame(Rel,state) disabled
    set Frame(sel,state) normal
    set Frame(dis,state) normal
    set Frame(mod,state) normal
    set frame(entry,param) " "
    set Frame(Comment,text) Comment
    set Frame(Comment,cmd) "activateComment"
#    set Frame(Comment,entry) " "
    set Frame(slist,state) disabled

    frameBuild
#    trac variabele Frame(Req,text) wru logvar
}

#
# builds the ifloor window. can be recreated during the session. when the 
# conferencing modes changes, sometimes the window must change as well 
#
proc frameBuild {} {
global session Frame 
    if [catch {destroy .bottom .header} result] {
	tracing "fc-->frame_init: destroy error: $result"
    }
    catch {destroy .li}
    catch {destroy .cmds}
    catch {destroy .modecom}
    catch {destroy .comments}
    catch {destroy .footnote}
    catch {destroy .dialog}

    tracing "fc-->frameBuild"
    wm title . "FloorControl for $session(id)"

#frame for footer
    set Frame(Footer,win) [frame .footnote -bd 2]
   pack [label $Frame(Footer,win).image1 -bg white -fg white -bitmap warning] \
              -side left -fill both -expand true
    pack [label $Frame(Footer,win).lab -text $Frame(Footer,lab) -anchor w \
	      -bg white \
              -font *-*-*-bold-n-*-10-*-*-*-*-*-*-*] \
              -side left  -fill both -expand true
  pack [label $Frame(Footer,win).image2 -bg white -fg white -bitmap warning] \
              -side right -fill both -expand true
    pack [label $Frame(Footer,win).msg -text $Frame(Footer,msg) -anchor w\
	      -bg white \
              -font *-*-*-bold-n-*-10-*-*-*-*-*-*-*] \
              -side right -fill both -expand true
 
    pack $Frame(Footer,win) -side bottom -fill x

#frame for dialogs
    set Frame(Dialog,win) [label .dialog -bd  0 -relief flat]
    set Frame(Dialog,win1) [frame .dialog.win1 -bd 0 -relief flat ]
    set Frame(Dialog,win2) [frame .dialog.win2 -bd 0 -relief flat]
    set Frame(Dialog,lab) [label $Frame(Dialog,win1).lab \
			       -text $Frame(Dialog,labtext) \
			       -anchor e \
      -bg LightGrey -font *-*-*-bold-n-*-12-*-*-*-*-*-*-* -bd 0 -relief flat]
    set Frame(Dialog,entry) [entry $Frame(Dialog,win1).msg  -bg LightGrey \
				 -width 1 \
				 -textvariable Frame(Dialog,msg) \
				-relief flat -state $Frame(Dialog,entrystate) \
				 -highlightcolor Lightgrey]
    set Frame(Dialog,but1) [button $Frame(Dialog,win2).but1 -bg LightGrey \
				-relief flat \
			-bd 0 -relief flat -state $Frame(Dialog,but1state) \
				-highlightcolor Lightgrey] 
    set Frame(Dialog,but2) [button $Frame(Dialog,win2).but2 -bg LightGrey \
				-relief flat\
                         -bd 0 -relief flat -state $Frame(Dialog,but2state)  \
				-highlightcolor Lightgrey]
    pack $Frame(Dialog,lab) $Frame(Dialog,entry)  -side left -fill both \
    -expand true 
    pack $Frame(Dialog,but1) -anchor w -side left -fill both -expand true 
    pack $Frame(Dialog,but2) -anchor e -side right -fill both -expand true

    pack $Frame(Dialog,win2) -side bottom -fill both -expand true
    pack $Frame(Dialog,win1) -side bottom -fill both -expand true
    pack $Frame(Dialog,win) -side bottom -fill x



#frame for mode and comment buttons
    set MC [frame .modecom]
    pack $MC -side bottom -fill both

#frame for conference modes
    set Frame(Mode,frame) [frame $MC.modes -borderwidth 2 -relief groove]
    set Frame(Mode,lab) [label $Frame(Mode,frame).lab\
			     -text "Conference Modes:"] 
    set Frame(Mode,but) [menubutton $Frame(Mode,frame).modes \
			     -text $session(mode) \
			     -bg LightGrey -activebackground Slategrey \
			     -menu $Frame(Mode,frame).modes.menu \
			     -bd 4 -relief raised]
    pack $Frame(Mode,lab) $Frame(Mode,but) -side left 
    pack $Frame(Mode,frame) -side left -fill x -ipadx 10 -ipady 10  

#menu for conference control modes
    sh_conf_modes


#frame for comment button
    set Frame(Comment) [frame $MC.com -borderwidth 2 -relief groove ]
    set Frame(Comment,cmd) "$Frame(Comment).but configure \
                           -bg IndianRed ; activateComment"
    set Frame(Comment,but) [button $Frame(Comment).but \
		-text $Frame(Comment,text) -bg LightGrey \
		-activebackground Slategrey -command $Frame(Comment,cmd)]
    pack $Frame(Comment,but) -padx 10 -pady 10 -anchor e
    pack $Frame(Comment)  -side left -fill both -expand true 


#frame for moderator commands
    if {[string compare $session(role) "Moderator"] == 0} {
        frame .cmds -relief sunken -borderwidth 1

        set ce [frame .cmds.entry  -borderwidth 4] 
        set cc [frame .cmds.choice  -borderwidth 4]
        set che [frame .cmds.head  -borderwidth 4]

#header
        pack [label $che.msg -anchor center \
                  -text "Moderator Commands" \
                  -font *-*-*-bold-i-*-12-*-*-*-*-*-*-*]
	pack $che -side top -fill x

#entry for cmd parameter    
        set Frame(entry) [entry $ce.par -width 30 -relief sunken \
                   -textvariable Frame(entry,param) -bg white]
        pack $ce.par -side left
#clear button
        button $ce.clear -text Clear -relief sunken \
	    -command "set Frame(entry,param) \" \""
        pack $ce.clear -side right -expand true

        pack $ce -side top 


#select command
        set Frame(sel,but) [button $cc.select -text Select -command \
               "eval_cmd \"Select\"" -width 9 -bg LightGrey \
		-state $Frame(sel,state) -activebackground Slategrey] 
        pack $Frame(sel,but) -side left -fill x 

#discard command
        set Frame(dis,but) [button $cc.discard -text Discard -command \
                      "eval_cmd Discard"  -width 9 -bg LightGrey \
                     -state $Frame(dis,state) -activebackground Slategrey]
        pack $Frame(dis,but) -side left 

#pass command
        set Frame(mod,but) [button $cc.pass -text "Moderator" \
		  -command "eval_cmd \"RequestMod\"" -width 9 -bg LightGrey \
		  -state $Frame(mod,state) -activebackground Slategrey]
        pack $Frame(mod,but) -side left 

        pack $cc -side top
        pack .cmds -side bottom -fill x
    }

#frame forbasic buttons
    frame .bottom -borderwidth 2 

    pack .bottom -side bottom -anchor w -fill both -expand true 

#command buttons
    set Frame(Req,but) [button .bottom.request -text $Frame(Req,text) \
                -command $Frame(Req,cmd) -state $Frame(Req,state) \
	        -bg LightGrey -activebackground Slategrey]
    set Frame(Pause,but) [button .bottom.pause -text $Frame(Pause,text) \
		       -command $Frame(Pause,cmd) \
                       -state $Frame(Pause,state)\
			      -bg LightGrey -activebackground Slategrey]

  set Frame(Rel,but) [button .bottom.release -text $Frame(Rel,text)  -command \
	   $Frame(Rel,cmd) -bg LightGrey -activebackground Slategrey \
                        -state $Frame(Rel,state) ]
    set Ebut [button .bottom.exit -text Exit -command i_close \
                        -bg LightGrey -activebackground Slategrey]
 
    pack .bottom.request .bottom.release .bottom.pause \
            .bottom.exit -side left
    bind  <Control-c> ReleaseFloor

#frame for list must be defined, even if hidden
    set Frame(slist) [frame .li -width 400]
    text $Frame(slist).list -width 35 -height 12 \
         -borderwidth 2 -relief sunken -setgrid true -bg LemonChiffon1 \
         -selectbackground Lightgrey -selectforeground  DarkSeaGreen\
          -xscrollcommand {.li.xscroll set} -font *-times-medium-i*-12-* \
          -yscrollcommand {.li.yscroll set} -font *-times-medium-i*-12-* \
          -wrap none -setgrid true
    scrollbar .li.xscroll -command [list .li.list xview] \
		    -bg LemonChiffon1 -borderwidth 0.2m -orient horizontal
    scrollbar $Frame(slist).yscroll -command [list .li.list yview] \
		    -bg LemonChiffon1 -borderwidth 0.2m -orient vertical
    $Frame(slist).list tag configure indent -tabs "80 left 200 left"
    $Frame(slist).list configure -state $Frame(slist,state)

    if [myself $session(moderator)] {
        $Frame(slist).list config -cursor hand2
    }

    bind $Frame(slist).list <Button-1> {

       set tkPriv(selectMode) line
       tkTextSelectTo %W %x %y
       catch {%W mark set insert sel.first}
       cp_item %W %y
    }

#pack speaker list if not hidden
#  packing_list
    tracing "fc--> frame_init: hide = $session(hide)"
    if {$session(role) != "Ordinary"} {
       pack $Frame(slist).xscroll -side bottom -fill x
       pack $Frame(slist).yscroll -side right -fill y
       pack $Frame(slist).list -side left -fill both -expand true
       pack $Frame(slist) -side bottom -anchor e -fill both
   } else {
       if {$session(hide) == 0} {
           pack $Frame(slist).xscroll -side bottom -fill x
           pack $Frame(slist).yscroll -side right -fill y
           pack $Frame(slist).list -side left -fill both -expand true
           pack $Frame(slist) -side bottom -anchor e -fill both
       } else { 
           tracing "fc_frame_init: list is hidden"
       }
   }
#selecting in list moves entries into entrybox
  

#moderator label
      tracing "fc--> frameinit: role == $session(role)"
      switch $session(role) {
          "Moderator" {
              set Frame(ModLab,text) "Conference Moderator" 
          }
          "Ordinary" {
              set  Frame(ModLab,text) \
                  "Conference is moderated by : $session(moderator)"
          }
          "Equal" {
              set  Frame(ModLab,text) "Conferencing Mode "
          }
      }
      set Frame(ModLab,frame) [frame .header -relief flat -borderwidth 2]
      set Frame(ModLab,lab) [label .header.mod -anchor center \
				 -text $Frame(ModLab,text) \
				 -font  *-*-*-bold-i-*-12-*-*-*-*-*-*-*]
      pack $Frame(ModLab,lab)
      pack $Frame(ModLab,frame) -side bottom -fill x
#      $ModLab.mod config -text "new text"
 
      tracing "fc-->frame_init: frame built"

#tag for "speaker"
  $Frame(slist).list tag configure speak -background LightGrey
  put_speakers
  highl_speaker
}


#***************************** end of frame_init ***************************

#puts stdout "fcmem18"
proc sh_conf_modes {} {
global Frame session
    catch {destroy $Frame(Mode,frame).modes.menu}

    set ConfMenu [menu $Frame(Mode,frame).modes.menu]
    switch -exact -- $session(role) {
        "Ordinary" {
            $Frame(Mode,but) configure -activebackground Slategrey\
		-text $session(mode)
        }
        "Moderator" {
            $ConfMenu add command -label Conferencing -command Equal
#                "$Frame(slist).list configure -cursor xterm ; Equal" 
            $ConfMenu add check -label Hide -command Hide \
                -variable session(hide)
	    $Frame(Mode,but) configure -text $session(mode)
        }
        "Equal" {
            $ConfMenu add command -label Moderated -command Moderator 
		$Frame(Mode,but) configur -text $session(mode)
      
 #            "$Frame(slist).list configure -cursorhand2 ; Moderator"  
	}

    }
}

proc getComment {} {
global Frame
    tracing "fc-->getComment: get $Frame(Dialog,msg)"
    $Frame(Dialog,but1) configure -bg Lightgrey
    writeComment $Frame(Dialog,msg)
    set  Frame(Dialog,msg) " "
    tracing "fc-->getComment: set $Frame(Dialog,msg)"    
    deactivateDialog
}

#
#the dialog box in . frame is activated. A comment can be written to a file.
#
proc activateComment {} {
global Frame
    tracing "fc-->activateComment"
    $Frame(Comment).but configure -bg IndianRed

    $Frame(Dialog,win) configure -bd 0 -relief flat
    set Frame(Dialog,labtext) "Write Comment"
    
    $Frame(Dialog,lab) configure -text $Frame(Dialog,labtext) -bd 0 \
          -relief flat
    set Frame(Dialog,entrystate) normal
    $Frame(Dialog,entry) configure -bg white -width 30 \
          -state $Frame(Dialog,entrystate) -bd 2 -relief raised
    tracing "fc-->activateComment:  entry = $Frame(Dialog,msg)"
    set Frame(Dialog,buttext1) Write
    set Frame(Dialog,but1state) normal
    $Frame(Dialog,but1) configure -text $Frame(Dialog,buttext1) -bd 2 \
                -command getComment -relief raised \
                -state $Frame(Dialog,but1state)

    set Frame(Dialog,buttext2) Dissmiss
    set Frame(Dialog,but2state) normal
    $Frame(Dialog,but2) configure -text $Frame(Dialog,buttext2) -bd 2 \
                -relief raised -state $Frame(Dialog,but2state) \
                -command "$Frame(Comment).but configure -bg Lightgrey ;\
                          set  Frame(Dialog,msg) \" \" \ ;
                          deactivateDialog"
}

proc getRequest {} {
global Frame
#    $Req(but) configure -bg IndianRed
    tracing "fc-->actRequest: get topic $Frame(Dialog,msg)"    
    set Frame(Req,entry) $Frame(Dialog,msg) 
    set Frame(Dialog,msg) " "
    tracing "fc-->actRequest: set topic $Frame(Dialog,msg)"   
    deactivateDialog
    RequestFloor
}

#
# the dialog box in . frame is activated. The member can write the topic he
# wants to talk about. Then RequestFloor is called.
#
proc activateRequest {} {
global session Frame
    tracing "fc--> activateRequest"
    $Frame(Req,but) configure -bg IndianRed

    $Frame(Dialog,win) configure -bd 0 -relief flat

#    catch {destroy $Frame(Dialog,win1)}
#    set Frame(Dialog,win1) [frame .dialog.win1 -bd 2]

    set Frame(Dialog,labtext) "Write Topic"

    $Frame(Dialog,lab) configure -text $Frame(Dialog,labtext) -bd 0 \
                 -relief flat
    set Frame(Dialog,entrystate) normal
    $Frame(Dialog,entry) configure -bg white -width 30 -relief raised \
                 -state $Frame(Dialog,entrystate) -bd 2

    set Frame(Dialog,buttext1) Write
    set Frame(Dialog,but1state) normal
    $Frame(Dialog,but1) configure -text $Frame(Dialog,buttext1) -bd 2 \
                -relief raised \
                -command getRequest -state $Frame(Dialog,but1state)

    set Frame(Dialog,buttext2) Dissmiss
    set Frame(Dialog,but2state) normal
    $Frame(Dialog,but2) configure -text $Frame(Dialog,buttext2) -bd 2 \
                -relief raised  -state $Frame(Dialog,but2state) \
          -command "$Frame(Req,but) configure -bg Lightgrey ; deactivateDialog"
}
#
#The dialog box in . frame is deactivated again.
#
proc deactivateDialog {} {
global session Frame
    tracing "fc-->deactivateDialog"
    $Frame(Dialog,win) configure -bg Lightgrey
    $Frame(Comment).but configure -bg Lightgrey

    $Frame(Dialog,win) configure -bd 0 -relief flat
    set Frame(Dialog,labtext) " "

    $Frame(Dialog,lab) configure -text $Frame(Dialog,labtext) -bg Lightgrey \
                  -bd 0 -relief flat
    set Frame(Dialog,entrystate) disabled
    set Frame(Dialog,msg) " "
    $Frame(Dialog,entry) configure -bg Lightgrey -width 1 \
                  -state $Frame(Dialog,entrystate) \
                  -bd 0 -relief flat -state disabled
    tracing "fc-->deactivateDialog: entry = $Frame(Dialog,msg)"
    set Frame(Dialog,buttext1) "  "
    set Frame(Dialog,but1state) disabled 
    $Frame(Dialog,but1) configure -text $Frame(Dialog,buttext1) -command ";" \
                  -relief flat -state $Frame(Dialog,but1state) -bd 0 -bg Lightgrey

    set Frame(Dialog,buttext2) " "
    set Frame(Dialog,but2state) disabled
    $Frame(Dialog,but2) configure -text $Frame(Dialog,buttext2) -command ";" \
                   -relief flat -state $Frame(Dialog,but2state) -bd 0 -bg Lightgrey
}


#
#For all dialogs with members: accept floor; accept moderatorship
#
proc Dialogs {msg RetValue} {
global session timeout Frame 
    $Frame(Dialog,win) configure -bg IndianRed
    set Frame(Dialog,$RetValue) 0
    $Frame(Dialog,win) configure -bd 0 -relief flat
    set Frame(Dialog,entrystate) disabled 
    $Frame(Dialog,entry) configure -textvariable Frame(Dialog,msg) \
        -relief raised  \
        -bg LightGrey -width 1 -state $Frame(Dialog,entrystate) -bd 0     
    set Frame(Dialog,labtext) "$msg"
 
    $Frame(Dialog,lab) configure -text $Frame(Dialog,labtext) -bg IndianRed \
        -bd 0 \
        -relief flat  

#    set $Frame(Dialog,entrytext) 
    set Frame(Dialog,buttext1) Yes
    set Frame(Dialog,but1state) normal
    set Frame(Dialog,but1cmd) "set Frame(Dialog,$RetValue) 1; \
                               set timeout 1;deactivateDialog"
    $Frame(Dialog,but1) configure -text $Frame(Dialog,buttext1) -bd 2 \
         -command $Frame(Dialog,but1cmd) -state $Frame(Dialog,but1state) \
         -relief raised 
                        
    set Frame(Dialog,buttext2) No
    set Frame(Dialog,but2state) normal
    set Frame(Dialog,but2cmd) "set timeout 0; \
                               deactivateDialog" 
    $Frame(Dialog,but2) configure -text $Frame(Dialog,buttext2) -bd 2 \
         -command  $Frame(Dialog,but2cmd) -state $Frame(Dialog,but2state) \
         -relief raised 

    set timeout 2
    set id [after $session(wait_acc) \
		"set timeout 0"]
    tkwait variable timeout
    after cancel $id
    tracing "fc-->Dialogs: return $timeout"
    deactivateDialog
 
#    deactivateDialog
}


proc clear_speakerl {} {
global Frame
   $Frame(slist).list configure -state normal
    if  { ![info exists Frame(slist)] }  { return }
    if { [winfo exists $Frame(slist).list] } {
	$Frame(slist).list delete 1.0 end
    }
   $Frame(slist).list configure -state disabled
}

proc clear_speaker {i} {
global Frame
   $Frame(slist).list configure -state normal
    $Frame(slist).list delete $i.0 [expr $i+1].0
   $Frame(slist).list configure -state disabled
}


proc flush_speaker {sp txt topic {i 0}} {
global Frame
    tracing "fc-->flush: i=$i"
    $Frame(slist).list configure -state normal
    if {[string length $topic] > 50} {
	set topic [string range $topic 0 49]
    }
    
    set ttxt "$sp \t $txt \t \"$topic"
    if {$i > 0} {
	$Frame(slist).list insert $i.0  "$ttxt \n"
    } else {
	$Frame(slist).list insert insert "$ttxt \n" 
    }
	$Frame(slist).list tag add indent "insert linestart" "insert lineend"
    $Frame(slist).list configure -state disabled
}

proc highl_speaker {} {
global Frame
    $Frame(slist).list configure -state normal
    $Frame(slist).list tag add speak 1.0 1.end
    $Frame(slist).list configure -state disabled
}

proc cp_item {src y} {
#cpoy entry from speakerlist to entry
global Frame
    set w [focus]
    tracing "focus = $w"
    focus $src
    if {![catch { set selection [selection get]}]} {

	set selection [string trimright $selection "\n"]
	set selection [string trim $selection]  
    tracing "fc-->cp_item: selection = $selection"
    }
    set selection [split $selection " "]
    set Frame(entry,param) [lindex $selection 0]
}

#
# enable menue buttons
# some buttons (moderator buttons) may not exist anymore)
#
proc enable {but} {
global Frame session
	set Frame($but,state) normal
    if [myself $session(moderator)] {

	$Frame($but,but) configure -state $Frame($but,state)
    }
}

#
# disable menue buttons
#
proc disable {but} {
global Frame session

    if [myself $session(moderator)] {
	tracing "fc-->disable Frame($but,state)"
	set Frame($but,state) disabled
	$Frame($but,but) configure \
	    -state $Frame($but,state) -background LightGrey
    }
}

proc ch_request {} {
global Frame
#requestfloor button is redefined: 2nd klick is strong request!
    tracing "fc-->ch_request"
    set Frame(Req,text) "Interrupt Speaker"
    set Frame(Req,cmd) SRequest
    $Frame(Req,but) configure -background LightGrey -text $Frame(Req,text) \
    -command $Frame(Req,cmd) -state $Frame(Req,state)
}

proc ch_srequest {} {
global Frame
#strong request button becomes request floor button again
    tracing "fc-->ch_srequest"
    set Frame(Req,text) "Request Floor"
    set Frame(Req,cmd) activateRequest
    $Frame(Req,but) configure -background LightGrey -text $Frame(Req,text) \
    -command $Frame(Req,cmd) -state $Frame(Req,state)
}



proc eval_cmd {cmd} {
global Frame
 #   tracing "eval_cmd:  cmd = $cmd $Frame(entry,param)"
    set entry $Frame(entry,param)
    set entry [string trim $entry]
    if {[string length $entry] == 0} {
	set_footer " " "Insert a members name first!"
	return
    }
    if {[string length $cmd] == 0} {

	return
    }
    eval [list $cmd  $entry]
 #    switch -regexp -- $entry {
# 	{( |\t)+} {
# 	    set_footer " " "Insert a members name first!"
# 	    return
# 	}
# 	default {
# 	    eval [list $cmd  $entry]
# 	}
#     }
}


#
# sets the label indicating the moderator
#
proc set_modLab {text {fsize 12}} {
global Frame
    set Frame(ModLab,text) $text
    $Frame(ModLab,lab) config -text $text \
    -font *-*-*-bold-i-*-$fsize-*-*-*-*-*-*-*
}

#puts stdout "fcmem19"

#******** "exit" ******************************************
proc tcl_exit {} {exit}	

proc i_close {} {
global session 
# stop sending messages ??
 
# media sessions are left uncontrolled:
    if {"$session(duplex)" == "halfduplex" } {
	agent_cntrl send 0
    } else {
	agent_cntrl send 1
    }
    agent_cntrl receive 1
# close protocol
    if [info exists session(history)] {
        catch  { flush $session(history)}
        catch { close $session(history)}
    }
# inform local isc
    set msg [list $session(session) "closed"]
    fc_send $msg "pmm"
# inform conference
    set msg [list $session(session)/$session(cname) "bye" ]
    fc_send $msg "mc"
#  tcl_exit
    destroy .
    exit
}

#******** traces and logs, comments ******************************

proc writeComment {msg} {
global session
    tracing "fc-->writeComment: $msg"
    if [info exists session(history)] {
        puts $session(history)  $msg
    } else {
	set_footer " " "No history file available"
    }
    deactivateDialog
}


#
# messages are shown in a seperated window, when ifloor main window has not
# been created now.
#
proc sh_msgwin {label msg} {
    global session
  tracing "fc-->sh_msg: $label - $msg"
    if [winfo exists .label] {catch {destroy .label}}
    set f [toplevel .label -borderwidth 10]
     wm title $f $label
    frame $f.buttons -bd 10
    message $f.msg -text $msg -bg LightGrey -relief flat -borderwidth 4 \
                   -aspect 1000 -pady 10 -padx 10
    button $f.buttons.ok -text Continue -command {catch {destroy .label}} \
    -underline 0 -relief raised -activebackground Slategrey
    button $f.buttons.abort -text Abort -command { i_close } -underline 0 \
           -relief raised -activebackground Slategrey
    pack $f.buttons.ok  -side right -fill x
    pack $f.buttons.abort -side right -fill x
    pack $f.msg $f.buttons -side top
    after $session(display) clear_footer
}   

#
# a message is shown as footer
#
proc set_footer {label msg} {
    global Frame session
    set $Frame(Footer,lab) $label
    set $Frame(Footer,msg) $msg 
    $Frame(Footer,win).image1 configure -fg IndianRed 
    $Frame(Footer,win).image2 configure -fg IndianRed 

    $Frame(Footer,win).lab configure -text $label  
    $Frame(Footer,win).msg  configure -text $msg
    update idletasks
    after $session(display) clear_footer
}

proc clear_footer { } {
    global Frame
    tracing "fc-->clear_footer"
    $Frame(Footer,win).image1 configure  -fg white
    $Frame(Footer,win).image2 configure  -fg white
    $Frame(Footer,win).lab configure -text " "
    $Frame(Footer,win).msg  configure -text " " 
    set $Frame(Footer,lab) " "
    set $Frame(Footer,msg) " " 
}    


proc tracing {msg} {
global session
    if [info exists session(watch)] {
	#puts $session(logfile) $msg
    }
}
#puts stdout "fcmem20"

#************ message evaluation *********************************

proc cntrl_pmm {msg} {
#evaluating messages from pmm
 global session errmsg mc_controller actagent myagent update_int speakerl
  set op [lindex $msg 0]
#  set  sessnam [split [lindex $msg 1] / ]
  set param [lindex $msg 2]
  switch -exact -- $op {
      session  {
	  tracing "fc, cntrl-->session: $session(session) - \
                                    session: [lindex $msg 1]" 
	  if {[string compare $session(session) [lindex $msg 1]] == 0} { 
              set mmsg [list $session(session) "created"]
              fc_send $mmsg "pmm" }
	  after $update_int update_active
          return
      }
      agent -
      +      { return
      }
  }
  set op [lindex $msg 1]
  set s [lindex $msg 0]
  set sessnam [split $s / ]
#my conference? select just one conference!
  if {$session(id) != [lindex $sessnam 0]} { return }
  if {$op != "active"} {
      if {$op != "!active"} {
	  tracing "fc-->cntrl_pmm1: op = $op"
      }
  }

# listen to pmm commands 
  switch -exact -- $op {
      hello {
	  eval [list $op $param]
      }

  }
# evaluate fc control commands

#evaluate initialization commands first
   if {[string match "*/ifloor/*" [lindex $msg 0]]} {
      switch -exact -- $op {
	  tracing "fc-->control_pmm2: op == $op"
        network -
        name -  
        loc -
	source -
        cname { set session($op) $param
	    return
        }
        ttl -
	host -
	port  { 
	    set session($op) $param
	    eval $op
        } 
        open  {
            i_open
            return
        }
        update_interval {
            set update_int $param
        }
      }
   }
#most commands can be evaluated only, if session is already "open"
 if {[string match "*/ifloor/*" [lindex $msg 0]]} {
     if {$session(opened) == 1} {
      switch -exact -- $op {
	  tracing "fc-->control_pmm2: op == $op"
#if agent closes and comes up again, it is still under controll!  
	addmedia {
#all sessions with media agent of type "param" will be controlled 
           
	      add_myagent $param
        }
	rmmedia { 
	    rm_myagent $param
	    rm_actagents $param
	    return
        }
	s_control {
              tracing "fc--cntrl_pmm: eval  add_actagent $param"
	      add_actagent $param
	}
	s_uncontrol {
	      rm_actagent $param
        }
        close { 
	  tracing "fc-->cntrl_pmm4: op = $op"
            i_close
        }
      }
     } 
  } else {
# evaluate control commands from agents
      set agent [lindex $sessnam 1]
#controll/uncontroll a particular media session
      switch -exact -- $op {
	  close -
	  closed { 
#one of my active agents close/d? closing agent is removed from actagent
	      rm_actagent $s
	  } 
      }
#control media sessions of a particular type 
#one of my agents?
      if { ![info exists myagent] } { return }
      if {[lsearch -exact $myagent $agent] < 0} { return }
      switch -exact -- $op {
	  created {
	      tracing "fc-->session_created: $s "
# session is "created" , agent is assumed to be already attached
	  }
          opened -
	  open {
# one of my agents opened? Add to actagent
	      add_actagent $s 
	      cntrl_speakers
	 }
      }
  }
}

proc floor_cntrl {msg} {
global next speakerl session errmsg Disable
# reading cmds from extern mc address for ifloor communication
# there may be several ifloors within one conference all communicating 
# via the same mc address !!
    tracing "fc-->control_mc:msg = $msg"
    set sessnam [split [lindex $msg 0] "/" ]
    set cn [string trim [lindex $sessnam 3]]
#???task: always param?? otherwise??
    set param [string trim [lindex $msg 2]]
#    set param2 [string trim [lindex $msg 3]]
    if { [lindex $sessnam 0] != $session(id) } {
	tracing "fc-->control_mc: 	$session(id) != [lindex $sessnam 0]"
	return }
    if { [lindex $sessnam 1] != "ifloor" } {
        set errmsg "fc--> warning: messages for unknown agent on fc mc address"
        flusherr
        return }

    set op [lindex $msg 1]
# A member wants to register (again). Must be heard even if "disabled".
    switch -exact -- $op {
	hello -
        touch {
	    if [info exists Disable($cn)] {
		unset  Disable($cn)
	    }
	    eval [list $op $cn]
	    return
      }
    }

    tracing "fc-->control_mc:op = $op"
    if [info exists Disable($cn)] {
	tracing "fc-->control_mc: refusing msg from disabled mem:$msg"
	return
    }
    switch -exact -- $op {
#heartbeat heared:
        heartbeat {
          tracing "fc-->control_mc:eval heartbeat"
	    eval $op $cn $param [lindex $msg 3] [lindex $msg 4] }
	hello -
#  member leaves, remove from speakerlist if in:
      bye {
	  eval [list $op $cn]
	  }
      speakers -
      speakerlend -
# a conference member asks me to send the speakerlist
      get_speakers {
	  eval [list $op $cn $param]
      }
      requestfloor {
#  time stamp is part of message
	  eval [list $op $cn $param [string trim [lindex $msg 3]]]
      }
      srequest -
      releasefloor -
      touch -
#      discard_conf -
#      select_conf -
      requestfloor_conf -
      acceptmod -
      rejectmod -
      retouch {
	  tracing "fc-->control_mc2:op = $op"
	  eval [list $op $cn]
      }
#      heartbeat -
      select - 
      hide -
      discard -
      equal -
      requestmod {
# moderated session? msg comes from moderator? otherwise ignore
	  if {$session(mode) == "Moderated"} {
	      if {[string compare $session(moderator) $cn] != 0} {
                      tracing "fc-->control_mc: $op msg not from moderator"
                      return
              }         
	  tracing "fc-->floor_cntl: eval $op $param"
	  eval [list $op $param]
	  }
      } 
      moderator {
# msg comes from moderator? otherwise ignore
	  if {$session(mode) == "Moderated"} {
# if session(frstMod) exists, conference has just changed to 'moderated'.
#  Thus, the time stamps must  be evaluated for correct establishment of 
# moderatorship.
              if {![info exists session(frstMod)]} {
                  if {[string compare $session(moderator) $cn] != 0} {
                      tracing "fc-->control_mc: $op msg not from moderator"
                      return
                  }
              }
          }
	  tracing "fc-->floor_cntl: eval $op $cn $param [lindex $msg 3]"
	  eval [list $op $cn $param [string trim [lindex $msg 3]]]
      }
#      close { leave $cn }
     closed { bye $cn }
  } 
}

randomInit [pid]
init_pmm_[lindex $pmm_controller 0]
cmd_line $argc $argv
announcement

#
# initial value of activity inirtval reports is set to 2000, the actual value
# to be used will then be delivered by isc
#
set update_int 2000

#
# signal the activity of the tool
#
proc update_active { } {
    global session update_int
    fc_send "$session(session) active" "pmm"
    tracing "fc--> update_active: update_int = $update_int" 
    if { $update_int > 0 } {
       after $update_int update_active
	# update the time values od the speakers and waiting
	put_speakers

    tracing "session(wait_heartbeat_started)=$session(wait_heartbeat_started)"
	set events [after info]
	tracing ">>>>>>pending events: $events"
    }
}


