#
# conference -- display group of members, each with several media
# Member frames are placed (using the table geometry manager) into
# a frame, .conference.members, of variable size.  The member
# positions are counted starting at 0, top-to-bottom, left to right
# (newspaper column style).
#
# $Id: conference.tcl,v 1.40 1998/10/06 09:39:53 dor Exp $
#=======
# $Id: conference.tcl,v 1.40 1998/10/06 09:39:53 dor Exp $
#


##########################################################################
# the convention for the name of the coding is rather confusing
# in the mailcap the names are in the form (jpeg, pcmu, l16)
# however we wrote nevit, nevot to accept coding values in the shape of
# JPEG, PCMU 1 8000
# therefore we convert all video types toupper before using it as 
# media type. The audio types are a little more of a problem. nevot needs 
# the channel value as well as the frequency. We conver between the mailcap 
# entry and the actual needed value in code_2_name. A close look at the 
# way entries are specified in the mailcap needs therefore to be made to reduce
# this confusion
##############################################################################


# Per-conference state:
# Initial conference receiving flag.
set conf(rx) 1

# List of free positions.
set conf(free) {}

set numbers {0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20}

	
#
# Invoked when pressing session "Receive" button. Sends message
# to pmm for all current sessions within the conference.
#
proc conference_receive {} {
  global conf

  foreach e $conf(sessions) {
    control_send "$conf(id)/$e receive $conf(rx)"
  }
}

#
# control if all the required local information are present
#
proc get_locals {} {
    if [catch {open "~/.RTPdefaults" r} id] {
	
    }
}
#
# Iconify the isc window and show configuration window
#
proc conference_init {{width 200} } {
    global conf     
    wm title    . $conf(name)
    wm iconname . $conf(name)
    wm iconify .
    #update idletasks
    conference_init_1
    # if [catch {open "~/.RTPdefaults" r} id] {
# 	destroy .conference_1
# 	wm deiconify .
# 	conference_init_2
# 	configure_init
#     } else {
# 	close $id
#     }
}
#
# enable the entry fields for this media
# 

proc choose_network_enabled {w row column type} {
    
    global media_list
    lappend media_list $type

    $w.$row$column configure \
	-command "choose_network_disabled $w $row $column $type"
    enable_row $w $row
}

#
# disable all the entries to prevent the user from specifying new sessions
# during the opening phase
#
proc disable_entries {w row column type} {
    global ${type} Session Default

     #test if this row already exists -i.e we only want to reactivate the media
    set test_column1 [expr $column+1]
    if ![winfo exists $w.$row$test_column1] {
	
	if { ![info exists ${type}(host)] } {
	    #intialize the media list and network parameters
	    set host $Default(host)
	    set ${type}(host) \
		[host2ip  \
		     [join [lreplace [split  $host .] 3 3 \
				[expr [lindex  \
					   [split $host .] 3]+$row + $column ]] .]]
	}
	if { ![info exists ${type}(port)] } {
	    set ${type}(port) $Default(port)
	}
	if { ![info exists ${type}(local)] } {
	    set ${type}(local) $Default(local)
	}

	set column [expr $column+1]
	grid configure \
	    [entry $w.$row$column  -width 15 -textvariable ${type}(host) \
		 -relief groove -state disabled -foreground grey \
		 -background white] \
	    -row $row -column $column -sticky w
	set column [expr $column+1]
	grid configure \
	    [entry $w.$row$column  -width 15 -textvariable ${type}(local) \
		 -relief groove -state disabled -foreground grey \
		 -background white ] \
	    -row $row -column $column -sticky w
	
	set column [expr $column+1]
	grid configure \
	    [entry $w.$row$column  -width 6 -textvariable ${type}(port) \
		 -relief groove -state disabled -foreground grey \
		 -background white] \
	    -row $row -column $column -sticky w
    } else {
	disable_row $w $row
	$w.$row$column configure \
	    -command "choose_network_enabled $w $row $column $type"
	$w.$row$column  configure -relief raised -foreground black \
	    -state normal	

    }	
    return $column
} 

#
# set network initialization part but with disabled entries, they are only 
# enabled when the user clicks on the media icon
#	 

proc choose_network_disabled {w row column type} {
    global ${type} media_list

    #test if this row already exists -i.e we only want to reactivate the media
    set test_column1 [expr $column+1]
    if ![winfo exists $w.$row$test_column1] {
	if { [lindex [grid size $w] 1] ==1 } {
	    #intialize the media list and network parameters
	    set media_list ""
	}
    } else {
	set media_list [remove_entry $media_list $type]
    }	
    return [disable_entries $w $row $column $type]
	
}

#
# procedures for configuring the protocols
#
proc config_protocol_RTP {{win ""}} {
}

proc config_protocol_VAT {{win ""}} {
}

proc config_protocol_UDP {{win ""}} {
}



# changed by mkr
proc config_connect_env {{win ""}} {
    
    set w $win.env
    if {[winfo exists $w]} {return}
    toplevel $w
    #put the window in a fixed place
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
    tixLabelFrame $w.net -label "Net"
    set frame [$w.net subwidget frame]

    pack \
	[radiobutton $frame.env_lan -variable connect_env(net) \
	     -text "LAN" -value LAN] \
	[radiobutton $frame.env_wan -variable connect_env(net) \
	     -text "WAN" -value WAN]
    pack \
	[button $frame.apply -text "Apply"  -relief raised\
	     -command "destroy $w" ]\
	-fill x -expand yes -anchor nw -side left
    
    pack $w.net -side left -anchor w -expand yes -fill both

    
}
#

proc config_protocol_ATM {{win ""}} {
    global Session
    set w $win.atm
    if {[winfo exists $w]} {return}
    toplevel $w
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
   wm title $w "ATM Network Configuration"
    set lw [string length "Peak Bandwidth Minimum:     "]
    pack \
       [hscale $w.peak_bandwidth_m "Peak Bandwidth Minimum: " $lw 0 10000  10 \
	    kbit/s Session(peak_bandwidth_minimum) 100 ] \
	[hscale $w.peak_bandwidth_t "Peak Bandwidth Target: " $lw 0 10000  10 \
	     kbit/s Session(peak_bandwidth_target) 100 ] \
	[hscale $w.mean_bandwidth_m "Mean Bandwidth Minimum: " $lw 0 10000 10 \
	     kbit/s Session(mean_bandwidth_minimum) 100 ] \
	[hscale $w.mean_bandwidth_t "Mean Bandwidth Target: " $lw 0 10000 10  \
	     kbit/s Session(mean_bandwidth_target) 100 ] \
	[hscale $w.mean_burst_m "Mean Burst Minimum: " $lw 0  20 0.1  \
	     kbit/s Session(mean_burst_minimum) 100 ] \
	[hscale $w.mean_burst_t "Mean Burst Target: " $lw 0  20 0.1  \
	     kbit/s Session(mean_burst_target) 100 ] \
	[field_str $w.device       "Device"        $lw 20 "" Session(device)] \
	-side top -anchor w   
    config_buttons $w "" config_accept
}

#
# choose the application transport protocol (RTP or VAT for example) 
# and the network transport protocol (ATM or UDP
#
proc choose_protocol {type transport network win row}  {
    global ${type}

    set w $win._$type
    if {![winfo exists $w ]} { 
	toplevel $w
	set x [expr [winfo screenwidth .] /2 ]
	set y [expr [winfo screenheight .] /2 ]
	wm geometry $w +$x+$y
	tixNoteBook $w.all
	pack $w.all -side left
    } else {
	destroy $win._$type
	return 
    }
    set w $w.all
    $w add transport -label "Transport" -underline 0
    $w add network   -label "network"   -underline 0
    
    set frame [$w subwidget transport]
    
    for { set i 0} {$i <[llength $transport]} {incr i} {
	radiobutton $frame._[lindex $transport $i] \
	    -variable ${type}(transport) -text [lindex $transport $i] \
	    -command "config_protocol_[lindex $transport $i] $w;\
                set  ${type}(transport) [lindex $transport $i];\ 
                 $win configure \
                   -text {[lindex $transport $i] [set ${type}(network)]}"\
	    -value [lindex $transport $i] 
	pack $frame._[lindex $transport $i] -side top -fill x 
    }
    set ${type}(network) [lindex $network 0]

    set frame [$w subwidget network]
    for { set i 0} {$i <[llength $network]} {incr i} {
	radiobutton $frame._[lindex $network $i] \
	    -variable ${type}(network) -text [lindex $network $i]\
	    -value [lindex $network $i] \
	    -command "config_protocol_[lindex $network $i] $w;\
                set  ${type}(network) [lindex $network $i];\ 
                 $win configure \
                    -text {[set ${type}(transport)] [lindex $network $i]}"
	pack $frame._[lindex $network $i] -side top -fill x 
    }
    set ${type}(network) [lindex $network 0]
     
}

#
# choose the audio packet size
#
proc choose_audio_frame {win} {
    global audio
    
    set w $win._frame
    if {![winfo exists $w ]} { 
	toplevel $w
	set x [expr [winfo screenwidth .] /2 ]
	set y [expr [winfo screenheight .] /2 ]
	wm geometry $w +$x+$y
    } else {
	destroy $win._frame
	return 
    }
    pack \
	[field_str $w.bpm        "Audio per packet"  15 6  "sec" \
	     audio(frame)] \

    set audio(frame) 0.02
}


#
# set audio initialization part
#
proc choose_audio {w row column} {
    global conf member Session image SIP audio
    
    set audio(transport) RTP
    set audio(network) UDP
    grid configure \
	[button $w.$row$column  -relief raised  \
	     -padx 4 -width 100 -anchor w \
	     -command "choose_network_enabled $w $row $column audio" ] \
	-row $row -column $column -sticky w
    compound_win $w.$row$column audio "Audio"

    iscBalloon $w.$row$column audio1
    #add a menu for the audio packet size
    bind $w.$row$column <ButtonRelease-3> "choose_audio_frame $w.$row$column"

    set column [choose_network_disabled $w $row $column audio]
    set column [expr $column + 1]
    grid configure \
	[button $w.$row$column -text "$audio(transport) $audio(network)" \
	     -width 10 -relief groove -stat disabled \
	     -command "choose_protocol audio {RTP VAT} {UDP} $w.$row$column $row"] \
	-row $row -column $column -sticky w

    iscBalloon $w.$row$column audio2

    # add menu codec
   set column [expr $column + 1]
    grid configure \
	[tixOptionMenu $w.$row$column -variable audio(media) \
	     -width 15 -state disabled -options {
		 menubutton.width 10
	         menubutton.relief raised }] \
    -row $row -column $column -sticky w
    set m $w.$row$column 

    iscBalloon $w.$row$column audio2
    # fim
    global mailcap_subtypes
    set type_list ""
    foreach i $mailcap_subtypes(audio) {
	if { [lsearch -exact $type_list $i ] < 0 } {
	    lappend type_list $i
	    $m add command [code_2_name $i]
	}
    }
    unset $type_list

    #set audio(media) [code_2_name pcmu]
    set audio(media) $Session(audio)
    disable_win $w.$row$column 
    return $row
}


#
# set video initialization part
#
proc choose_video {w row column} {
    global conf member Session image SIP video
    
    set video(transport) RTP
    set video(network) UDP
    grid configure \
	[button $w.$row$column  -relief raised  \
	     -padx 4  -width 100 -anchor w \
	     -command "choose_network_enabled $w $row $column video" ] \
	-row $row -column $column -sticky w 
    compound_win $w.$row$column video "Video"
    
    iscBalloon $w.$row$column video1

    set column [choose_network_disabled $w $row $column video]
    set column [expr $column + 1]
    
    grid configure \
	[button  $w.$row$column   \
	     -text "$video(transport) $video(network)"\
	     -width 10 -relief groove -state disabled \
	     -command "choose_protocol video {RTP} {UDP ATM} $w.$row$column $row"] \
	-row $row -column $column -sticky w 


    iscBalloon $w.$row$column video2
    # add menu codec
   set column [expr $column + 1]
    grid configure \
	[tixOptionMenu $w.$row$column -variable video(media) \
	     -width 20 -state disabled -options {menubutton.width 10}] \
    -row $row -column $column -sticky w
    set m $w.$row$column 

    iscBalloon $w.$row$column video3
    # fim
    global mailcap_subtypes
    set type_list ""
    foreach i $mailcap_subtypes(video) {
	if { [lsearch -exact $type_list $i ] < 0 } {
	    lappend type_list $i
	    $m add command [string toupper $i]
	}
    }
    unset $type_list
    set video(media) $Session(video) 

    disable_win $w.$row$column 
    return $row
}
#
# set image initialization part
#
proc choose_picture {w row column} {
    global conf member Session picture SIP image
    set picture(transport) RTP
    set picture(network) UDP
    grid configure \
	[button $w.$row$column  -relief raised  \
	     -padx 4  -width 100 -anchor w \
	     -command "choose_network_enabled $w $row $column picture" ] \
	-row $row -column $column -sticky w 
    compound_win $w.$row$column picture "Picture"
    iscBalloon $w.$row$column picture1
 
    set column [choose_network_disabled $w $row $column picture]
    set column [expr $column + 1]
    
    grid configure \
	[button  $w.$row$column  -text "$picture(transport) $picture(network)"\
	     -width 10 -relief groove -state disabled \
	     -command "choose_protocol picture {RTP} {UDP ATM}\
                           $w.$row$column $row"] \
	-row $row -column $column -sticky w 
    iscBalloon $w.$row$column picture2


    # add menu codec
   set column [expr $column + 1]
    grid configure \
	[tixOptionMenu $w.$row$column -variable picture(media) \
	     -width 20 -state disabled -options {menubutton.width 10}] \
    -row $row -column $column -sticky w
    set m $w.$row$column 
    $m add command "[string toupper jpeg]" 
    set picture(media) [string toupper jpeg]
    iscBalloon $w.$row$column picture3

    disable_win $w.$row$column 
    return $row
}
# netghost configuration
#
proc choose_netghost {w row column} {
    global conf member Session netghost SIP image
    set netghost(transport) RTP
    set netghost(network) UDP
    grid configure \
        [button $w.$row$column  -relief raised  \
             -padx 4  -width 100 -anchor w \
             -command "choose_network_enabled $w $row $column netghost" ] \
        -row $row -column $column -sticky w 
    compound_win $w.$row$column netghost "NetGhost"
    iscBalloon $w.$row$column netghost1
 
    set column [choose_network_disabled $w $row $column netghost]
  set column [expr $column + 1]
    
    grid configure \
        [button  $w.$row$column   -text "   " \
             -width 10 -relief groove -state disabled \
             -command " "] \
        -row $row -column $column -sticky w 
    iscBalloon $w.$row$column netghost2
 
 
 
    set column [expr $column + 1]
    grid configure \
        [tixOptionMenu $w.$row$column -variable netghost(mode) \
             -width 20 -state disabled -options {menubutton.width 10}] \
    -row $row -column $column -sticky w
    iscBalloon $w.$row$column netghost3
    set m $w.$row$column 
    $m add command "sender" 
    $m add command "receiver" 
    set netghost(mode) receiver
 
 
 
    disable_win $w.$row$column 
    return $row
}

proc choose_ifloor {w row column} {
    global conf member Session image SIP ifloor Ifloor
    set ifloor(media) audio
    grid configure \
	[button $w.$row$column  -relief raised  \
	     -padx 4  -width 100 -anchor w\
	     -command "choose_network_enabled $w $row $column ifloor" ] \
	-row $row -column $column -sticky w 
    compound_win $w.$row$column ifloor "Ifloor"
    iscBalloon $w.$row$column ifloor1
 
    set column [choose_network_disabled $w $row $column ifloor]
    set column [expr $column + 1]
    
    grid configure \
	[button  $w.$row$column   -text "   " \
	     -width 10 -relief groove -state disabled \
	     -command " "] \
	-row $row -column $column -sticky w 

   set column [expr $column + 1]

#add menue

   grid configure \
	[menubutton  $w.$row$column  \
	     -width 14  -relief groove -state disabled \
	     -text "Control Media" \
	     -menu $w.$row$column.menu] \
	-row $row -column $column -padx 5 -sticky w 
    set m [menu $w.$row$column.menu]
    iscBalloon $w.$row$column ifloor2
	
    $m add check -label Video -variable Ifloor(video) 
    $m add check -label Audio -variable Ifloor(audio)
    disable_win $w.$row$column 
    return $row
}

proc choose_reservation {w row column} {
    global conf member Session image Rsvp rsvp
    
    grid configure \
	[button $w.$row$column  -relief raised  \
	     -padx 4  -width 100 -anchor w \
	     -command "choose_network_enabled $w $row $column reservation" ] \
	-row $row -column $column -sticky w 
    compound_win $w.$row$column reservation "RSVP"
    iscBalloon $w.$row$column rsvp1
    set column [choose_network_disabled $w $row $column reservation]
    set column [expr $column + 1]
    
    grid configure \
	[button  $w.$row$column   -text "   " \
	     -width 10 -relief groove -state disabled \
	     -command " "] \
	-row $row -column $column -sticky w 

   set column [expr $column + 1]
    grid configure \
	[button  $w.$row$column   -text "   " \
	     -width 12 -relief groove -state disabled \
	     -command " "] \
	-row $row -column $column -padx 5 -sticky w 

    disable_win $w.$row$column 
    return $row
}

# fim
proc choose_vote {w row column} {
    global conf member Session image SIP vote

    grid configure \
	[button $w.$row$column  -relief raised  \
	     -padx 4  -width 100 -anchor w\
	     -command "choose_network_enabled $w $row $column vote" ] \
	-row $row -column $column -sticky w 
    compound_win $w.$row$column vote "Vote"
    iscBalloon $w.$row$column vote1
 
    set column [choose_network_disabled $w $row $column vote]
    set column [expr $column + 1]
    
    grid configure \
	[button  $w.$row$column   -text "   " \
	     -width 10 -relief groove -state disabled \
	     -command " "] \
	-row $row -column $column -sticky w 

   set column [expr $column + 1]
    grid configure \
	[button  $w.$row$column   -text "   " \
	     -width 12 -relief groove -state disabled \
	     -command " "] \
	-row $row -column $column -padx 5 -sticky w 

    disable_win $w.$row$column 
    return $row
}

#
# after choosing all the medias to create start them all and prevent the user
# from adding any new ones untill all the current ones were created
#
proc open_medias {w row} {
    global media_list conf Session all_media_list Ifloor
    
    set column 0
    disable_win [winfo parent $w]

    # start the members window and set the number of required medias
    wm deiconify .
    conference_init_2
    if { ![info exists media_list] } {
	return
    }
    set limit [llength $media_list]

    # start each needed media
    for { set i 0} {$i <$limit  } {incr i} {
	set type [lindex $media_list 0]

	global ${type}
	set conf(started) 0
	set Session(type) $type
	foreach e [array names ${type}] {
	    if { $e == "host" } {
		set ${type}($e) [host2ip [set ${type}($e)]]
	    }
	    set Session($e) [ set ${type}($e)]
	}
	set ses "$Session(type)/$Session(host):$Session(port)"
	set session($ses,media) $Session(media)

	# create the session 
	set Ifloor(session) [session_create]

	#remove the add media from media_list
	set media_list [remove_entry $media_list $type]
	footer ""
    }
    # enable all entries in the window
    enable_win [winfo parent $w].header 
    enable_win [winfo parent $w].end 

    # enable/disable the appropriate entries in the body
    for { set i 1} {$i <= $row} {incr i} {
	#disable the network, protocol and codes .. entries
	disable_row $w $i
	
	
	# enable the medias buttons
	$w.$i$column  configure -relief raised -foreground black \
	    -state normal
	$w.$i$column configure \
	    -command "choose_network_enabled $w $i $column \
                [lindex $all_media_list [expr $i -1]]"
    }
}
#
# disable all children windows of win
#
proc disable_win { win } {
    catch {$win configure -state disabled\
	       -foreground grey -relief groove -borderwidth 2}
    for { set i 0} {$i < [ llength [winfo children $win]]} {incr i} {
	disable_win [lindex [winfo children $win] $i]
    }
    
}

#
# disable all windows in row belongin to master win
#
proc disable_row { win row} {

    for { set i 0} {$i < [lindex [grid size $win] 0]}  {incr i} {
	disable_win $win.$row$i
    }
}
#
# enable all children windows of win
#
proc enable_win { win } {
    global mailcap_subtypes mailcap_rules


    for { set i 0} {$i < [ llength [winfo children $win]]} {incr i} {
	set but [winfo class [lindex [winfo children $win] $i]]
	switch -glob $but {
	    Button -
	    Menubutton {
		catch { [lindex [winfo children $win] $i] configure \
			    -state normal -relief raised -foreground black}
	    }
	    Entry {
		catch {[lindex [winfo children $win] $i]  configure \
			   -state normal -relief sunken -foreground black}
	    }
	    default  {
		catch {[lindex [winfo children $win] $i]  configure \
			   -state normal -relief groove -foreground black}
	    }
	}
	enable_win [lindex [winfo children $win] $i]
    }
    
}
#
# enable all windows in row belongin to master win
#
proc enable_row { win row} {

    for { set i 0} {$i < [lindex [grid size $win] 0]}  {incr i} {
	enable_win $win.$row$i
	set but [winfo class $win.$row$i]
	switch -exact $but {
	    Button -
	    Menubutton {
		catch {$win.$row$i configure -state normal \
		       -relief raised -foreground black}
	    }
	    Entry {
		catch {$win.$row$i configure -state normal \
			   -relief sunken -foreground black}
	    } 
	    default  {
		catch {$win.$row$i configure -state normal \
			   -relief groove -foreground black}
	    }
	}
    }
}

#
# create the compound image text window. due to some strange behavior 
# in tix we need to destroy all the compound windows before destroying the
# toplevel window. therefore we keep track of all compound windows in comp_win
#
proc compound_win {w type text {state -1}} {
    global image env comp_win Session
    lappend comp_win $w
    if {[info exists env(ISC_IMG)]} {
	set img [image create compound -window $w]
	$img add line -anchor w
	$img add image -image $image($type)
	$img add space -width 10p
	$img add text -text $text -justify left
	$w config -image $img
    } else {
	set img [image create compound -window $w]
	$img add line -anchor w
	# fim
	catch [$img add bitmap -bitmap $type] 
	$img add space -width 10p
	$img add text -text $text -justify left
	$w config -image $img  -height 25
    }	
    # activate or deactvate the buttons in accordance to the choosen server 
    # mode or the explicitlt given state (help should always be active)
    if { $state < 0 } {
	$w config -state $Session(server)
    } else {
	$w config -state $state
    }
}

#
# add a new media row to the list
#
proc conference_add_row { w type } {
    global Ifloor    
    set column 0
    set row [lindex [grid size $w] 1]
    set row [choose_$type $w $row $column]
    lappend all_media_list $type
    if {$type == "ifloor" } {
        set Ifloor(row) $row
        set Ifloor(rowwin) $w
        set Ifloor(column) $column
    }
}

#
# choose the media and their parameters
# write all shown media in all_media_list
#
proc choose_media {{version 1} } {
    global conf member Session image SIP all_media_list connect_env mailcap_subtypes mailcap_types mailcap_rules


    if [winfo exists .media_select] {destroy .media_select} 
    set window .media_select
    toplevel  $window
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $window +$x+$y
    wm title $window "Initialize Media"
   
    set w [frame $window.header  -relief groove  -borderwidth 1]
    
    #changed by mkr
    menubutton $w.mb -text "Preferences" -menu $w.mb.menu -relief raised
    set m [menu $w.mb.menu -tearoff 1]

    $m add radio -label Default -variable connect_env(type) \
	-command "load_config; set_config" 

    $m add radio -label User -variable connect_env(type) \
	-command "load_config; set_config"

    $m add separator
    $m add radio -label LAN -variable connect_env(net) \
	-command "load_config; set_config" 

    $m add radio -label WAN -variable connect_env(net) \
	-command "load_config; set_config" 

    #

    pack \
	[field_str $w.title  "Conference Title" 18 30 "" conf(name) ] \
	[field_num $w.ttl "TTL"     6 6  "" Session(ttl)]  \
	$w.mb \
	-side left -anchor w -ipadx 5
    
    if { [llength $conf(sessions)] !=0 } {
	disable_win $w.title
	$w.title.in configure -foreground black 
	$w.ttl.in configure -foreground black 
    }
    # add the help balloons
    iscBalloon $w.title title
    iscBalloon $w.ttl ttl
    iscBalloon $w.mb preferences

    pack $w -expand yes -side top -fill both -ipady 1m -pady 1m
    
    
    set w [frame $window.body  -relief groove ]
    pack $w -expand yes -side top -fill both -ipady 1m -pady 1m
 
    set i 0
    set row 0
    set column 0

    foreach e {Media Destination Local Port Protocol Coding} {
	grid  configure \
	    [label $w.i$e -text $e] \
	    -row $row -column $i -sticky w 
	iscBalloon $w.i$e $e
	set i [incr i]	
    }

    # fim
    mailcap_load
    set mailcap_types [array names mailcap_rules]
    foreach t $mailcap_types {
	set mailcap_subtypes($t) ""
	set list $mailcap_rules($t)
	set flag($t) 0
	foreach element $list {
	    set mailcap_subtype [string tolower [lindex $element 0]]
	    if {[lsearch [join $element] interactive] != -1} {
		if { [lindex $element 1] == "none"} {
		    lappend mailcap_subtypes($t) $mailcap_subtype
		    set flag($t) 1
		}
	    }
	}
	if {$flag($t) == 1} {
	    catch {conference_add_row $w $t} e
	    if {[string first "invalid" $e] < 0} {
		#puts "Mailcap type `$t' added"
	    } else {
		#puts "Warning: Mailcap type `$t' not supported!"
	    }
	}
    }

    set w [frame $window.end  -relief groove -borderwidth 1 ]
    pack \
	[button $w.cancel -text Cancel -command "destroy $window;  \
          conference_init_$version " -relief raised]\
	[button $w.add -text "Create Session " \
	     -command "open_medias $window.body $row; destroy $window" \
	     -relief raised]\
	-side left -expand yes -fill both
    pack $w -expand yes -side top -fill both -ipady 1m -pady 1m
    iscBalloon $w.add create_session
    iscBalloon $w.cancel cancel
}

#
# mint introducion window
#
proc conference_help {} {
    global env version
    
    if [winfo exists .conference_help] {destroy .conference_help}
    set w .conference_help
    toplevel $w
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
    wm title $w "Multimedia Internet Terminal"
    frame $w.text -relief groove -borderwidth 2
    set t [text $w.text.help -wrap word -width 72 -height 5]
    $t insert end {
Welcome to the Multimedia Internet Terminal:
MInT is a flexible conferencing tool offering audio, video, 
floor contol and invitation services. To use MInT you need 
to configure your mailcap as desribed in the handbook.
To get more information use the help functionality of MInT.
This ISC is $version
    }
    $t configure -state disabled
   
# create the help frame
    set w_help [frame $w.frame]
    pack \
	[button $w_help.copy -text "Copyright" -command "about $w_help.copy"]\
	[button $w_help.book -text "Handbook"  -command "help $w_help.book"]\
	[checkbutton $w_help.on -text "Online Help" -variable Session(help) -relief raised ]\
	-side left -fill both -expand yes

    pack $t
    pack $w.text \
	$w_help \
	[button $w.destroy -text Close -command "destroy $w"]\
	-side top -fill both -expand yes
}
#
# Show configuration window
#
proc conference_init_1 {{width 200} } {
    global conf member Session image SIP

    if [winfo exists .conference_1] {destroy .conference_1}
    set w .conference_1
    toplevel .conference_1
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
    wm title $w "Initialize Conference" 
    frame $w.title -width 500
    pack \
	[button $w.title.im -image $image(mint) -text MINT -command conference_help -anchor center ] \
	[label $w.title.title -text "Multimedia Internet Terminal" -height 3] \
	-side left -anchor w -fill both -expand yes 
    pack $w.title -side top -fill both -expand yes

   # activity buttons
    
    pack \
	[button $w.add \
	     -command "destroy $w; choose_media 1"  -height 30]\
	[button $w.invite -text "Invite"  \
	     -command "destroy $w;conference_init_2;wm deiconify .;get_active_session_parameters" -height 30]\
	-side left -anchor w -fill both -expand yes 
    compound_win $w.add add_media "Create Session"
    compound_win $w.invite invite "Invite"


    iscBalloon $w.add add_media 
    iscBalloon $w.invite invite

}


#
#  build the notebook for the configuration, starting with the source 
#  description
#
proc configure_init {{s ""} {type sdes} {w_child sdes}} {
 global config_modified config_modified_levels Session Levels session config_modified_rsvp conf rsvp Levels
    # configuration menu
    set w .configure
    if [winfo exists $w] {
	destroy $w
	return
    }
    if [winfo exists $w_child] return
    if {![winfo exists $w ]} { 
	toplevel $w
	set x [expr [winfo screenwidth .] /2 ]
	set y [expr [winfo screenheight .] /2 ]
	wm geometry $w +$x+$y
	wm title $w  "$conf(name): General Configuration"
	tixNoteBook $w.all
	pack $w.all -side top
    }
    set lw [string length "User e-mail (user@host)"]

#    $w.all add sdes -label "Source Description" -underline 0

    # changed by mkr
    #$w.all add connect_env -label "User preferences" -underline 0
    #

    # not needed currenntly so just ignore it -dor
    #set frame [$w.all subwidget sdes ]
    #config_sdes $s $frame
    $w.all add mail -label "Mailcap" -underline 0
    set frame [$w.all subwidget mail ]
    config_mailcap $s $frame

    #changed by mkr
    #set frame [$w.all subwidget connect_env]
    #config_profile $s $frame
    #

    #config_buttons $w "" config_accept
    config_cancel $w
    #trace variable session w config_trace
    #trace variable rsvp w config_trace_rsvp
    trace variable Levels w  config_trace_levels
    set config_modified ""
    set config_modified_levels ""
    set config_modified_rsvp ""
    update idletasks
}


#
# main window look as follows
# .conference -------------------------
#             |            |          |    
#             |f_audio     |   body   | 
#             |------------|----------|      
#             | f_sessions | control  |      
#             |            |----------|      
#             |            | message  |      
#             --------------------------
#

# Initialize conference display and member list.
# 
proc conference_init_2 {{width 270} } {
    global conf member image conf_width Session play record monitor
    if [winfo exists .conference] {return; destroy .conference}
    
    wm title    . $conf(name)
    wm iconname . $conf(name)
    
    frame .conference
    # frame for the sessions
    frame .conference.control
    set w_sessions [tixLabelFrame  [frame .conference.control.sessions \
					-relief groove -borderwidth 1 \
					-width 100 ].label\
			-label "Sessions"   -options {
			    frame.width 100}]
    iscBalloon $w_sessions sessions
    set Session(f_sessions) [$w_sessions subwidget frame]

    # frame for the audio control
    set w_audio [tixLabelFrame [frame .conference.control.audio  \
			      -borderwidth 4 -width 200 -height 170].label\
			      -label "Audio control" -options {
				  frame.width 100}]

    set Session(f_audio_control) [$w_audio  subwidget frame] 

    audioctl

     pack .conference.control.audio $w_audio .conference.control.sessions $w_sessions  \
	-side top -anchor nw -fill both -expand yes
 
    #frame for the members, messages, control
    set w_body [frame .conference.body -relief groove -borderwidth 2 ]
    tixScrolledWindow  $w_body.members \
	-relief groove -borderwidth 1 
    # create the conatiner window for the members
    set Session(f_members) [frame [$w_body.members subwidget window].frame \
				    -bg LemonChiffon1 -height 100]
    pack $Session(f_members) -anchor w -expand yes -fill both

    # create window for the message
    set Session(f_message) [message $w_body.message -aspect 10000 -anchor w -bg white]
    
    set w_control [frame $w_body.control -relief groove -borderwidth 2]
    set Session(f_control) $w_control
    pack  \
	[button $w_control.add  -relief raised  \
	     -command "choose_media 2"  -anchor w] \
	[button $w_control.invite  -relief raised  \
	     -command "get_active_session_parameters" -anchor w] \
	[button $w_control.handler  -relief raised  \
	     -command "sip_handler" -anchor w] \
        [button $w_control.config -text "Preferences" -relief raised \
	     -image $image(configure) -command configure_init -anchor w] \
       	[button $w_control.help -command conference_help \
	       -relief raised  -image $image(help) -anchor w] \
	[menubutton $w_control.control  -menu $w_control.control.menu \
	      -relief raised  -anchor w] \
        [menubutton $w_control.leave  -menu $w_control.leave.menu \
	      -relief raised -anchor w] \
	-side left  -fill x -anchor w 
    

    compound_win $w_control.add add_media "Add Media" ; iscBalloon $w_control.add add_media
    compound_win $w_control.invite invite "Invite" ; iscBalloon $w_control.invite invite
    compound_win $w_control.handler sip_handler "Invitation Handler" ; iscBalloon $w_control.handler invitation_handler 
    compound_win $w_control.config configure "Settings" ; iscBalloon $w_control.config configure_session
    compound_win $w_control.leave exit "Leave" normal; iscBalloon $w_control.leave leave
    compound_win $w_control.help help "Help" normal
    compound_win $w_control.control control "Control" ; iscBalloon $w_control.control control_session

    pack $w_body.members $w_control $w_body.message -side top -fill both -expand yes -anchor w

    #pack the big window
    pack .conference.control .conference.body  \
	-side left  -anchor w

    pack .conference
    bind .conference <Configure> "conference_resize %w"
    wm minsize . [winfo width .] [winfo height .]
    # only width can be changed manually
    wm resizable . false false
    set conf(cols)  1
    set conf_width [winfo width .conference]
    
     
    # leave/quit menu
    set m [menu $w_control.leave.menu -tearoff 0]
    $m add command -label "All (Quit)" -command "conference_close"
    
   
    # control menu
    set m [menu $w_control.control.menu -tearoff 0]
    $m add check -label "Receive" -variable conf(rx) \
	-command "conference_receive"
    $m add radio -label "Conferencing" -value 1 \
	-variable conf(conference) \
	-command "conference_start 1"
    $m add radio -label "Teleteaching" -value 0 \
	-variable conf(conference) \
	-command "conference_start 0"
    control_send "+ agent"
}

# toggle between conference mode and teleteaching mode
proc conference_start { mode } {
    global conf conference_id
    set conf(conference) $mode
    if { $mode==0} {
	set conference_id [ after 10000 member_active]
	return
    }
    after cancel  member_active
    after cancel $conference_id
}
#
# Right mouse button pressed anywhere: push-to-talk for all media.
#
proc conference_button3 {on} {
  global conf member

  foreach e $conf(sessions) {
    set member([lindex $conf(us) 0],$e,rx) $on
    control_send "$conf(id)/$e send $on"
	#make sure that the control buttons and the control icons 
	#show thesame state for transmission
	set session($e,send) $member([lindex $conf(us) 0],$e,rx)
  }
}

proc rc {pos} {
  global conf

  return "[expr $pos / $conf(cols)],[expr $pos % $conf(cols)]"
}


#
# Master window has been (manually) resized in width; thus, resize all
# active columns.  Should never happen due to member join/leave.
#
proc conference_resize {x} {
    return
}


# 
# destroy all child windows of win
proc destroy_all { win } {
    for { set i 0} {$i < [ llength [winfo children $win]]} {incr i} {
	 destroy_all [lindex [winfo children $win] $i]
    }
    destroy $win
}

# delete all the compound windows first to make sure that all fonts 
# and colors are freed -a tix problem
proc conference_close_comp_win { } {
    global comp_win
    for { set i 0} {$i < [ llength $comp_win]} {incr i} {
	catch {destroy [lindex $comp_win $i]}
    }
}

#
# Closes all media sessions in a conference and removes them from list.
#
proc conference_close {} {
    global conf
    
    # cancel any pending starts to avoid hanging
    set conf(started) 0
    
    # save actual configuration in .mintrc
    save_config

    foreach s $conf(sessions) {
	# send a close message and cleanup the session
	session_delete $s
    }
    # close the compund window first
    conference_close_comp_win
    exit
    
}


# checks every update_interval which sessions have reported them selves as 
# active the non active sessions are then deleted 
proc conference_activity { } {
    global conf Session session

    foreach s $conf(sessions) {
	if { $session($s,active) == 0 } {
	    session_delete $s
	} else {
	    set session($s,active) 0
	}
    }
    after [expr $Session(update_interval) + 1000 ] conference_activity
}
