#
# Horizontal floating-point scale widget
# w:     window name
# label: label text
# lw:    label width (if < 0, use length of label text)
# res:   resolution
# v:     variable name 
# $Id: hscale.tcl,v 1.4 1998/08/17 14:59:09 stb Exp $
proc hscale {w label lw from to res unit v {length 100} {command ""} {bigincrement 0} } {
  upvar #0 $v value
  frame $w

  if {$lw < 0} {
    set lw [string length $label]
  }

  set logarithmic 0
  if {$from > 0} {
    if {$to / $from > 100} {
      set logarithmic 1
    }
  }
  if {$to - $from > 500} {
    set logarithmic 1
  }

  pack \
      [label $w.l -text $label: -width $lw -anchor nw] \
      [scale $w.s  -resolution $res -showvalue yes \
	   -from $from -to $to -orient h -variable $v -length $length \
	   -command $command] \
      -side left

  if {$unit != ""} {
    pack [label $w.u -text $unit -anchor nw] -side right
  }

  if {[info exists value] == 0 || $value == ""} {
    set value $from
  }
  $w.s set $value
  return $w
}

#
#reconfigure the textof the label parameter
# 
proc hscale_label {w text} {
    $w.l configure -text $text
}

#
#reconfigure the unit of the scaleparameter
# 
proc hscale_unit {w {unit ""}} {
    $w.u configure -text $unit
}

#
# activate/inactivate a scale entry
#
proc hscale_active { w active } {
    if { $active ==0 } {
	$w.l configure -fg grey
	$w.s configure -state disabled -fg grey
	$w.u configure -fg grey
    }
    if { $active == 1} {
	$w.l configure -fg black
	$w.s configure -state normal -fg black
	$w.u configure -fg black
    }
}	
