#
# isc_rsvp: contains most the of the needed procedures for handling the 
# reservation and setting of RSVP specs
#
#
# for unicast we have two sessions 
#   reservation/localhost:port,reservation/remotehost:port
# the window is called reservation/remotehost:port
# but as the only changes that happen during a reservation session are the 
# addition and removal of flow entries we will use the id of the reservation
# session as entry point for the interface (reservation/localhost:port)

# to connect an rsvp session to a media session we use the
#   session($s,rsvp) variable where $s indicates the media session
#
#
# as a common win id both the sedner and receiver sessions use rsvp($s,win) 
# with s here either ssender or sreceiver


# reserve a little bit more than the actual coding used
set reservation_bonous 0.1
#
#
# Create buttons for accepting/dismissing parameter changes.
#
proc config_buttons_rsvp {w s command} {
    global config_modified Session session config_modified_rsvp

    if [winfo exists $w.action] return
    pack \
	[button $w.cancel -text "Close" -relief raised -borderwidth 1\
	     -command \
	     " set config_modified {}; destroy $w;"]\
	-fill both -expand yes -anchor nw -side left

    update idletasks
}


#
# update the entries for the rate of the rsvp session when changin max 
# bandwidth or audio coding
#
proc update_rsvp {s v} {
    global tspec session reservation_bonous
    set s "reservation/$session($s,host):$session($s,port)"
    if { [info exists tspec($s,r)] } {
	set tspec($s,r) [expr [expr $v *1000 / 8 ] +[expr $v *1000 * $reservation_bonous/ 8 ]  ]
	set tspec($s,p) [expr [expr $v *1000 / 8 ] +[expr $v *1000 * 2 * $reservation_bonous/ 8 ]  ]
	set tspec($s,b) [expr $v *1000 * 2 * $reservation_bonous/ 8 ] 
    }
}

#
# Trace for dynamically changeable variables. (rsvp)
# Strangely enough, upvar fails (but without error indication) 
#   if called through the backspace binding.
#
proc config_trace_rsvp {name element op} {
  global config_modified_rsvp rsvp

  #
  # append to list of modified variables
  #
  if {$element != ""} {
    set name ${name}($element)
  }
  set config_modified_rsvp $name
  ladd config_modified_rsvp $name
}
#
# Changes parameters for session 'cs'.  Invoked by accept button;
# checks for modified values.
#
proc config_accept_rsvp {{s ""}} {
  global conf config_modified_rsvp rsvp

    if {$s != ""} {
	set cs $conf(id)/$s
	if { $config_modified_rsvp != ""} {
	    foreach c $config_modified_rsvp {	    
		set e [lindex [split $c ",()"] 2]
		if {$s != ""} {
		    if {[regexp -nocase source $e]} {
			if { $rsvp($s,filtertype) == 1 } {
			    set value $rsvp($s,$e)
			    set source [string range $e 6 end]
			    set e "reserve $value"
			    regsub -all {_} $source {.} source 
			    regsub -all {\/} $source {.} source
			    set value $source
			    control_send "$cs $e [list $value]"
			}
		    } else {
			if {[regexp -nocase flowspec $e]} {
			    set source [string range $e 8 end]
			    regsub -all _ $source . source 
			    regsub -all \/ $source . source
			    set e "flowspec $source"
			}
			set value [set $c]
			control_send "$cs $e [list $value]"
		    }
		}
		set config_modified_rsvp ""
	    }
	}
    } else {
	foreach s $conf(sessions) {
	    config_accept_rsvp $s
	}
    }
}

#
# Reservation-specific settings.
#
proc config_tspec {{s ""} {w .reservation} title} {
    global Tspec Session limit tspec session delay

    set w $w.tspec
    if [winfo exists $w] return
    toplevel $w
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
    wm title $w "TSPEC $title"
    set lw [string length "Max :  "] 
    
    tixLabelFrame $w.up -label "QoS Parameter"
    set wup [$w.up subwidget frame]
#    frame $w.up

#    set wtype [frame $w.up.type -relief groove  -borderwidth 1]
    tixLabelFrame $w.type -label "Serviceclass"
    set wtype [$w.type subwidget frame]


#    [hscale $w.up.type "Type"  $lw 0 6 1 Number \
#     tspec($s,type) 200] 

#    pack \
#    [button $w.up.type -text "Select Service Class" \
#	 -relief raised -borderwidth 1 \
#	 -command "choose_tspec_serviceclass $s $w.up.type 1"] \
#	-side top -anchor n
    pack \
	[radiobutton $wtype.generic -text "QoS TSpec (generic)" -anchor w \
	     -relief flat -variable tspec($s,type) -value 0 \
	     -command "set tspec($s,type) 0" ]\
	[radiobutton $wtype.cntlload -text "Controlled Load" -anchor w \
	     -relief flat -variable tspec($s,type) -value 1 \
	     -command "set tspec($s,type) 1" ]\
	[radiobutton $wtype.guaranteed -text "Guaranteed Service" -anchor w \
	     -relief flat -variable tspec($s,type) -value 2 \
	     -command "set tspec($s,type) 2" ]\
	-side top -anchor nw	
    
    pack \
    [hscale $wup.b "b"  $lw 1 [expr $limit * 1000 /8] 100 byte \
      tspec($s,b) 200 "" 1000] \
    [hscale $wup.r "r"  $lw 1 [expr $limit * 1000 /8] 100 byte/s \
      tspec($s,r) 200 "" 1000] \
    [hscale $wup.p "p"  $lw 1 [expr $limit * 1000 /8] 100 byte/s \
      tspec($s,p) 200 "" 1000] \
    [hscale $wup.m "m"  $lw 0 $limit 10 byte \
      tspec($s,m) 200 "" 100] \
    [hscale $wup.mM "M"  $lw 0 20000 10 byte \
      tspec($s,M) 200 "" 100] \
    -side top -anchor nw

    
    #iscBalloon $w.up.type t
    iscBalloon $wtype.generic t
    iscBalloon $wtype.cntlload t
    iscBalloon $wtype.guaranteed t
    iscBalloon  $wup.b b 
    iscBalloon  $wup.r r
    iscBalloon  $wup.p p
    iscBalloon  $wup.m m 
    iscBalloon  $wup.mM M 
    
	pack $w.type $w.up  -side top -fill both
  frame $w.action
  pack \
    [button $w.action.dismiss -text "Close" -relief raised\
	 -command "destroy $w;"] \
    [button $w.action.accept  -text "Accept"  -relief raised\
	 -command "update_tspec $s; destroy $w;"] \
    [button $w.action.apply  -text "Apply"  -relief raised\
	 -command "update_tspec $s;" ] \
	-fill both -expand yes  -side left   
  pack $w.action

    #after $delay(rsvp_service_class) choose_tspec_serviceclass $s $w.up.type 0

}

#
# s session name
# bt button widget in rsvp window
# change = 0 just set the correct text for the button
# change = 1 set text, select next service class 
# and reinvoke choose_serviceclass
#
proc choose_tspec_serviceclass {s bt change} {
    global tspec Rsvp debug

    
    switch -exact -- $tspec($s,type) {
	0 {
	    $bt config -text "QoS TSpec"
	    if {$change=="1"} {
		set tspec($s,type) 1
		choose_tspec_serviceclass $s $bt 0
	    }
	}
	1 {
	    $bt config -text "Controlled Load"
	    if {$change=="1"} {
		set tspec($s,type) 2
		choose_tspec_serviceclass $s $bt 0
	    }
	}
	2 {
	    $bt config -text "Guaranteed Service"
	    if {$change=="1"} {
		set tspec($s,type) 3
		choose_tspec_serviceclass $s $bt 0
	    }
	}
	3 {
	    $bt config -text "Experimental 1"
	    if {$change=="1"} {
		set tspec($s,type) 4
		choose_tspec_serviceclass $s $bt 0
	    }
	}
	4 {
	    $bt config -text "Experimental 2"
	    if {$change=="1"} {
		set tspec($s,type) 5
		choose_tspec_serviceclass $s $bt 0
	    }
	}
	5 {
	    $bt config -text "Experimental 3"
	    if {$change=="1"} {
		set tspec($s,type) 0
		choose_tspec_serviceclass $s $bt 0
	    }
	}
    }
    if { $debug == "pmm" } {
	puts stdout "$s isc-->rsvp serviceclass $tspec($s,type)"
    }

}

proc config_flowspec {s w src title} {
    global Flowspec Session limit session flowspec delay

    regsub -all {\.} $src _ tmp
    set w $w.flowspec$tmp
    if [winfo exists $w] return
    toplevel $w
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
    wm title $w "FLOWSPEC $title"
    set lw [string length "Max :  "] 
    

    tixLabelFrame $w.up -label "QoS Parameter"
    set wup [$w.up subwidget frame]
#    frame $w.up

#    set wtype [frame $w.up.type -relief groove  -borderwidth 1]
    tixLabelFrame $w.type -label "Serviceclass"
    set wtype [$w.type subwidget frame]


#    pack \
#	[button $w.up.type -text "Select Service Class" \
#	     -relief raised -borderwidth 1 \
#	     -command "choose_flowspec_serviceclass $s $src $w.up.type 1"] \
#	-side top -anchor n
    


#    [hscale $w.up.type "Type"  $lw 0 6 1 Number \
#      flowspec($s,$src,type) 200] 
    pack \
	[radiobutton $wtype.generic -text "QoS TSpec (generic)" -anchor w \
	     -relief flat -variable flowspec($s,$src,type) -value 0 \
	     -command "set flowspec($s,$src,type) 0" ]\
	[radiobutton $wtype.cntlload -text "Controlled Load" -anchor w \
	     -relief flat -variable flowspec($s,$src,type) -value 1 \
	     -command "set flowspec($s,$src,type) 1" ]\
	[radiobutton $wtype.guaranteed -text "Guaranteed Service" -anchor w \
	     -relief flat -variable flowspec($s,$src,type) -value 2 \
	     -command "set flowspec($s,$src,type) 2" ]\
	-side top -anchor nw	



    pack \
	[hscale $wup.b "b"  $lw  1 [expr $limit * 1000 /8]  100 byte \
	     flowspec($s,$src,b) 200] \
	[hscale $wup.r "r"  $lw 1 [expr $limit * 1000 /8]  100 byte/s \
	     flowspec($s,$src,r) 200] \
	[hscale $wup.p "p"  $lw 1 [expr $limit * 1000 /8]  100 byte/s \
	     flowspec($s,$src,p) 200] \
	[hscale $wup.m "m"  $lw 0 $limit 10 byte \
      flowspec($s,$src,m) 200] \
	[hscale $wup.mM "M"  $lw 0 20000 10 byte \
	     flowspec($s,$src,M) 200] \
	[hscale $wup.rR "R"  $lw 0 $limit 10 byte/s \
	     flowspec($s,$src,R) 200] \
	[hscale $wup.sS "S"  $lw 0 $limit 10 ms \
	     flowspec($s,$src,S) 200] \
	-side top -anchor nw
    
    #iscBalloon  $w.up.type t
    iscBalloon $wtype.generic t
    iscBalloon $wtype.cntlload t
    iscBalloon $wtype.guaranteed t
    iscBalloon  $wup.b b 
    iscBalloon  $wup.r r
    iscBalloon  $wup.p p
    iscBalloon  $wup.m m 
    iscBalloon  $wup.mM M 

    pack $w.type $w.up -side top -fill both
    frame $w.action
    pack \
	[button $w.action.dismiss -text "Close" -relief raised\
	     -command "destroy $w;"] \
	[button $w.action.accept  -text "Accept"  -relief raised\
	     -command "update_flowspec $s $src; destroy $w;"] \
	[button $w.action.apply  -text "Apply"  -relief raised\
	     -command "update_flowspec $s $src;" ] \
    -padx 10 -side left
    pack $w.action

    #after $delay(rsvp_service_class) choose_flowspec_serviceclass $s $src $w.up.type 0

}

proc choose_flowspec_serviceclass {s src bt change} {
    global flowspec Rsvp debug

    
    switch -exact -- $flowspec($s,$src,type) {
	0 {
	    $bt config -text "QoS TSpec"
	    if {$change=="1"} {
		set flowspec($s,$src,type) 1
		choose_flowspec_serviceclass $s $src $bt 0
	    }
	}
	1 {
	    $bt config -text "Controlled Load"
	    if {$change=="1"} {
		set flowspec($s,$src,type) 2
		choose_flowspec_serviceclass $s $src $bt 0
	    }
	}
	2 {
	    $bt config -text "Guaranteed Service"
	    if {$change=="1"} {
		set flowspec($s,$src,type) 3
		choose_flowspec_serviceclass $s $src $bt 0
	    }
	}
	3 {
	    $bt config -text "Experimental 1"
	    if {$change=="1"} {
		set flowspec($s,$src,type) 4
		choose_flowspec_serviceclass $s $src $bt 0
	    }
	}
	4 {
	    $bt config -text "Experimental 2"
	    if {$change=="1"} {
		set flowspec($s,$src,type) 5
		choose_flowspec_serviceclass $s $src $bt 0
	    }
	}
	5 {
	    $bt config -text "Experimental 3"
	    if {$change=="1"} {
		set flowspec($s,$src,type) 0
		choose_flowspec_serviceclass $s $src $bt 0
	    }
	}
    }
    if { $debug == "pmm" } {
	puts stdout "$s isc-->rsvp serviceclass $flowspec($s,$src,type)"
    }

}


# this function is the main entry point for the reservation stuff. 
# We start creating the needed sessions and build the appropriate windows
#
# RSVP deals with siplex flows, i.e. we have to register two sessions
# at the rsvp daemon
#
proc do_reservation {s {rate ""}} {
    global Session conf session tspec delay localhost rsvp

    # convert from bit to byte
    #set rate [expr $rate *1000 / 8]
    # just to inform the user 
    footer {"Starting RSVP support, please wait..."}
    # this is our session destination
    set host $session($s,host)
    set localhost $session($s,local)
    # is it a multicast session ?
    if {[connection_is_multicast $host]} {
	set session($s,rsvp) [do_multicast_reservation $s $rate]
    } else {
	set session($s,rsvp) [do_unicast_reservation  $s $rate]
    }
    configure_reservation_win_$rsvp($session($s,rsvp),mode) $session($s,rsvp) $rsvp($session($s,rsvp),reserve_win)
}
 

#
# UNICAST reservation
# we have to register two different sessions, one for each direction
# The sessions will be called: reservation/REMOTEHOST:PORT and 
# reservation/LOCALHOST:PORT
# When session_create is called, pmm-rsvp will be informed about the 
# new RAPI_SESSIONs immediately
# so we have to handle all network events (PATH_EVENT, ...) and the GUI
# window stuff separately, otherwise
# there is a chance to loose information on the PMM bus
#
# the window variables will be handled using the reservation/REMOTEHOST:PORT
# as a session id
# 
proc do_unicast_reservation { s {rate ""}} {
    global Session conf session tspec delay localhost rsvp Rsvp

    set remotehost $session($s,host)

    # RSVP flowdirection: remotehost -> localhost
    set sreceiver reservation/$localhost:$session($s,port)
    # save some variable for later use
    set rsvp($sreceiver,mode) "unicast"
    # RSVP flowdirection: localhost -> remotehost
    set ssender reservation/[host2ip $session($s,host)]:$session($s,port)
    if {[lsearch -exact $conf(sessions) $ssender] <0 } {
	set Session(type) reservation
	set Session(host) $session($s,host)
	set Session(port) $session($s,port)
	set rsvp($ssender,flowsender) $localhost
	set rsvp($ssender,flowreceiver) $remotehost
	set rsvp($ssender,mode) unicast
	session_create
    }

    if { [lsearch -exact $conf(sessions) $ssender] < 0 } {
	footer "Could not Establish: $sreceiver"
	after $delay(display) "footer {}"
    } else {		    
	# store the default tspec
	#set tspec($ssender,r) $rate
	update_rsvp $ssender $rate

	# creating the rsvp window for the second flow
	set rsvp($ssender,path) $Rsvp(path)
	set rsvp($ssender,reserve) $Rsvp(reserve)
	# set the id of the senders session accompanaying the receiver 
	# session to ssender
	regsub -all {\.} $session($ssender,host) {_} host
	set port $session($ssender,port)     
	#set rsvp($ssender,win) [configure_reservation_win $ssender ]
	set rsvp($sreceiver,reserve_win) [configure_reservation_win $ssender ]
	# save the window name to the sedner session as well
	set rsvp($ssender,reserve_win) $rsvp($sreceiver,reserve_win)

	# initialize the WF and FF windows to ""
	set rsvp($ssender,WF) ""
	set rsvp($ssender,FF) ""
	set rsvp($sreceiver,WF) ""
	set rsvp($sreceiver,FF) ""
    }
    
    #RSVP flowdirection: remotehost -> localhost
    if {[lsearch -exact $conf(sessions) $sreceiver] <0 } {
	set Session(type) reservation
	set Session(host) $localhost
	set Session(port) $session($s,port)
	set rsvp($sreceiver,flowsender) $remotehost
	set rsvp($sreceiver,flowreceiver) $localhost
	session_create
    }


    if { [lsearch -exact $conf(sessions) $sreceiver] < 0 } {
	footer "Could not Establish: $sreceiver"
	after $delay(display) "footer {}"
    } else {		    
	# store the default tspec
	#set tspec($sreceiver,r) $rate
	update_rsvp $sreceiver $rate

	set rsvp($sreceiver,path) $Rsvp(path)
	set rsvp($sreceiver,reserve) $Rsvp(reserve)
	set rsvp($sreceiver,filtertype) $Rsvp(filtertype)
    }
    # to always be able to find the sending session when knowing the receiver
    # session and vice versa
    set rsvp($sreceiver,other) $ssender
    set rsvp($ssender,other) $sreceiver
    
    return $sreceiver
}


#
# MULTICAST reservation
# we have to register only one session for multicast
# The session will be called: reservation/mcaddr:PORT
# When session_create is called, pmm-rsvp will be informed about the new 
# RAPI_SESSION rsvpd will start sending PATH informations to pmm-rsvp 
# using the callback function.
# We do not have to 
# register a second session as needed for unicast !
# 
proc do_multicast_reservation {s {rate ""}} {
    global Session conf session tspec delay localhost rsvp Rsvp

    set remotehost $session($s,host)

    set smulticast reservation/$session($s,host):$session($s,port)
    if {[lsearch -exact $conf(sessions) $smulticast] <0 } {
	set Session(type) reservation
	set Session(host) $session($s,host)
	set Session(port) $session($s,port)
	set rsvp($smulticast,flowsender) $localhost
	set rsvp($smulticast,flowreceiver) $remotehost
	set session($smulticast,host) $session($s,host)
	set session($smulticast,loss_upper_threshold) 0
	session_create
    }


    if { [lsearch -exact $conf(sessions) $smulticast] < 0 } {
	footer "Could not Establish: $smulticast"
	after $delay(display) "footer {}"
    } else {		    
	set rsvp($smulticast,path) $Rsvp(path)
	set rsvp($smulticast,reserve) $Rsvp(reserve)
	set rsvp($smulticast,filtertype) $Rsvp(filtertype)
	regsub -all {\.} $session($smulticast,host) {_} host
	set port $session($smulticast,port) 
	
	#set tspec($smulticast,r) $rate
	#set the appropriate reservation values
	update_rsvp $smulticast $rate

    }  

    # create the reservation window
    set rsvp($smulticast,reserve_win) [configure_reservation_win $smulticast ]

    # save some variable for later use
    set rsvp($smulticast,mode) multicast
    # initialize the WF and FF windows to ""
    set rsvp($smulticast,WF) ""
    set rsvp($smulticast,FF) ""

    return $smulticast
    
    
}

#
# if a media session closed make sure to close any rsvp session
# attached to it
#
proc rsvp_attached_delete { s } {
    global rsvp session
    if { [info exists session($s,rsvp)] } {
	rsvp_session_delete $session($s,rsvp)
	unset session($s,rsvp)
	set w .reservation_[noperiod $s]
	if { [winfo exists $w] } { 
	    destroy  $w
	}
    }
}
   
#
# end an rsvp session (for multicast just delete the session,
# for unicast delete two sessions
proc rsvp_session_delete { s } {
    global rsvp session
    
    # check if a sender session is atached to this session as well
    # and delete it
    if { [info exists rsvp($s,other)] } {
	session_delete $rsvp($s,other)
    }
    session_delete $s
}


# 
# display the mesages sent by pmm-revp
#
proc rsvp_message { mesg w} {
    global delay
    if { ![winfo exists $w] } return
    set w [winfo toplevel $w]
    if {[string match *ERROR* $mesg]} {
	$w.message configure -fg red -text $mesg
    } else { 
	$w.message configure -fg darkgreen -text $mesg
    }
    after $delay(display) "rsvp_message {} $w"
}
#
# pack the actual contents of the reservation window 
# here only the parts handling the path stuff are created and thus
# we use the id of the remote host for unicast
# return the id of the reservation window
#
proc configure_reservation_win {s } {
    global conf  rsvp limit session Rsvp message Message
    
    set w .reservation_[noperiod $s]


    if [winfo exists $w] return
    toplevel $w
    set x [expr [winfo screenwidth .] /2 ]
    set y [expr [winfo screenheight .] /2 ]
    wm geometry $w +$x+$y
    wm title $w $s


    if {![winfo exists $w.path]} {
	set w2 [frame $w.path -relief groove  -borderwidth 1]
	tixLabelFrame $w2.path \
	    -label "Flow: $rsvp($s,flowsender) -> $rsvp($s,flowreceiver)" 
    } else {
	set w2 $w.path
    }
    set f_path [$w2.path subwidget frame]


    pack \
	[checkbutton $f_path.path_radio \
	     -text "send PATH to $rsvp($s,flowreceiver)" -relief flat \
	     -variable rsvp($s,path)\
	     -borderwidth 1 \
	     -command "update_tspec $s;config_parameters $s rsvp($s,path)" ] \
	-side left
    pack \
	[button $f_path.tspec_button \
	     -text "Tspec $rsvp($s,flowsender)" \
	     -command "config_tspec $s $w $rsvp($s,flowreceiver)" -relief groove]\
	-side right 

    
    pack $w2.path $w2  -side top -expand yes -fill both
    
    #
    # receiver part
    #

    if {![info exists message($s,rsvp_msg)]} {
	set message($s,rsvp_msg) $Message(rsvp_msg)
    }

    if {![winfo exists $w.reserve]} {
	set w2 [frame $w.reserve -relief groove  -borderwidth 1]
	tixLabelFrame $w2.reserve \
	    -label "Flow: $rsvp($s,flowreceiver) -> $rsvp($s,flowsender)" 
    } else {
	set w2 $w.reserve
    }
    set f_reserve [$w.reserve.reserve subwidget frame]
    set rsvp($s,reserve_win) $f_reserve
    pack $w2.reserve $w2 -side top -expand yes -fill both



    # general configuration for RSVP
    footer {}


    if {![winfo exists $w.message]} {
	pack [message $w.message -aspect 10000 -anchor w ] \
	    -expand yes -fill both
    }	
    

    if {![winfo exists $w.end]} {    
	pack [button $w.end -text "End RSVP Session" \
		 -command "rsvp_session_delete $s; destroy $w" -borderwidth 1]\
	    -expand yes -fill both 
    }	
    config_buttons_rsvp $w $s config_accept_rsvp  
    return $f_reserve
}

#
# pack the parts for the WF reservation
# this is only relevant for the multicast case
#
proc configure_reservation_win_multicast { s w } {
    global session rsvp
    
    if ![winfo exists $w] return

    if {![winfo exists $rsvp($s,WF)]} {
	# any path messages seen yet?
	if { [llength $rsvp($s,name)] != 0 } {
	    set f_reserve [frame $w.wf]
	    regsub -all {\.} $rsvp($s,flowreceiver) _ host_tmp
	    set rsvp($s,source$host_tmp) 0
	    pack \
		[button $f_reserve.flowspec_button  -text "Flowspec WF" \
		     -command "config_flowspec $s $w $host_tmp $rsvp($s,flowreceiver)" -relief groove]\
		-side right
	    pack \
		[checkbutton $f_reserve.reserve_button \
		     -text "reserve WF $rsvp($s,flowreceiver)" \
		     -relief flat -variable rsvp($s,source$host_tmp) \
		     -borderwidth 1 \
		     -command "set rsvp($s,filtertype) 0; \
                            config_parameters $s rsvp($s,filtertype); \
                            update_flowspec $s $host_tmp;\
                            config_parameters $s rsvp($s,source$host_tmp)" ] \
		-side left
	    
	    pack $f_reserve -side top -fill both -expand yes
	    set rsvp($s,WF) $f_reserve
	}
    }
    configure_reservation_win_unicast $s $w
}
#
# pack the parts for the FF reservation
#
proc configure_reservation_win_unicast {s w} {
    global rsvp session
    
    set source ""
    # FF Style buttons and reserve radiobutton
    if {![winfo exists $w.ff]} {
	set rsvp($s,FF) [frame $w.ff]
	pack $rsvp($s,FF) -side top -fill both -expand yes
    }
    foreach entry $rsvp($s,name) {
	regsub -all {\/} $entry \. source
	regsub -all {_} $entry \. source_tmp
	set rsvp($s,source$source) 0
	if {![winfo exists $rsvp($s,FF).$source]} {

	    frame $rsvp($s,FF).$source 
	    pack \
		[button $rsvp($s,FF).$source.flowspec_button \
		     -text "Flowspec $source_tmp"  \
		     -command "config_flowspec $s $w $source $source_tmp" \
		     -relief groove] \
		-side right
	    pack \
		[checkbutton $rsvp($s,FF).$source.reserve_button \
		     -text "reserve FF $source_tmp" \
		     -relief flat -borderwidth 1 \
		     -variable rsvp($s,source$source) \
		     -command "set rsvp($s,filtertype) 1;\
                              config_parameters $s rsvp($s,filtertype);\
                              update_flowspec $s $source;\
                              config_parameters $s rsvp($s,source$source)" ] \
		-side left
	    pack $rsvp($s,FF).$source -side top -expand yes -fill both
	    # save the window name so we can delete it easily
	    set rsvp($s,FF$source_tmp) $rsvp($s,FF).$source
	}
    }
}


proc update_tspec {{s ""}} {
    global Tspec rsvp tspec conf
    
    set cs $conf(id)/$s
    set rsvp($s,tspec) [concat $tspec($s,type) $tspec($s,b) \
			  $tspec($s,r) $tspec($s,p) $tspec($s,m) $tspec($s,M)]
    #set Session(tspec) $session($s,tspec)
    control_send "$cs tspec \{$rsvp($s,tspec)\}"
    control_send "$cs path \{$rsvp($s,path)\}"
}

proc update_flowspec {{s ""} {src ""}} {
    global  Flowspec rsvp flowspec conf
    set cs $conf(id)/$s    

    regsub -all \/ $src . source
    regsub -all _ $source . source
    
    set rsvp($s,flowspec$src) \
	[concat $flowspec($s,$src,type) $flowspec($s,$src,b) \
	     $flowspec($s,$src,r) $flowspec($s,$src,p) $flowspec($s,$src,m) \
	     $flowspec($s,$src,M) $flowspec($s,$src,R) $flowspec($s,$src,S)]
    set Session(flowspec) $rsvp($s,flowspec$src)
    #config_parameters $s rsvp($s,filtertype) 
    control_send "$cs flowspec $source \{$rsvp($s,flowspec$src)\} "

}



#
# stub rsvp configuration window
#
proc config_reservation {{s ""} {w .reservation}} {
    configure_reservation_win_sender $s $w
    configure_reservation_win_receiver $s $w
    
}

#
# clear the entries for the members who teared down their PATH messages
# for the case that all members cleared their sessions delete the WF entry as 
# well
proc configure_reservation_win_delete { s m} {
    global rsvp 
    
    if {[winfo exists $rsvp($s,FF$m)]} {
	destroy $rsvp($s,FF$m)
    }
    if { [llength $rsvp($s,name)] == 0 } {
	if { [info exists rsvp($s,WF)]} {
	    if {[winfo exists $rsvp($s,WF)]} {
		destroy $rsvp($s,WF)
	    }
	}
    }

}

#
# merge the values of the different received tspecs to generate a flowspec with
# the largest values for the WF filter reservation
#
proc merge_values {recv_value srclist e s} {
    global flowspec

    set value $recv_value
    set value2 $flowspec($s,[lindex $srclist 0],$e)
    foreach host $srclist {
	if { $flowspec($s,$host,$e) > $value2 } {
	    set $value2 $flowspec($s,$host,$e)
	}
    }

    if { $value > $value2 } {
	return $value 
    } else {
	return $value2 
    }
    
}
