# delete the session member list window
#
proc session_member_del { } {
    destroy .member_list
}

#
# list all member of a certain session
#
proc session_members {w s x y} {
    global member session
    set w $w.member_list
    if { [ winfo exists $w ] } return
    toplevel $w
    wm title $w  "Members"
    wm geometry $w +$x+$y
    set x 0
    foreach m $session($s,members) {
	label $w.$x -text $m;iscBalloon $w.$x "\
Cname:  $member($m,cname)
" iscBalloonMember $m
	pack $w.$x -fill both
	incr x
    }
    pack [button $w.dismiss -text Dismiss -command "destroy $w" -relief raised] -fill both
}

#
# create the entry itself
#
proc session_create_entry { win col {text ""}} {
    global colors Session

    # the header entry consists only of the color index and the label should
    # have no length
    if { $text == ""} {
	tixLabelEntry $win -label " " -options {
	    label.width 0
	    entry.width 1
	    label.anchor w
	}
    } else {
	# the entry consists of the color index and label s
	# all label should have the same length a
	tixLabelEntry $win -cursor gumby -label $text -options {
	    label.width 30
	    entry.width 1
	    label.anchor w
	}
	if { $Session(server) == "normal" } {
	    bind $win.label <ButtonRelease-1> "config_media $text"
	}
	bind $win.label <ButtonRelease-2> "session_members %W $text %X %Y"
    }

    $win subwidget entry configure \
	-bg [lindex $colors [expr $col % [llength $colors]]]\
	-relief flat -state disabled -width 1
    return $win
}

# 
# create a session entry in the members window consisting of the session name
# and an identification color
#
proc session_entry { s} {
    global conf Session
    set col [expr [llength $conf(entry)] -1 ]
    # add session to the session list
    pack [session_create_entry $Session(f_sessions).$col $col $s] \
	-anchor w -expand yes

    
}

#
# This function returns the dotted notation of another host
# by resolving the given name from a hello cmd
# e.g.: user@host.domain.net -> 192.168.0.1
#       user@192.168.0.1     -> 192.168.0.1
#
proc get_host_from_cname {user_host} {
    set  addrstr $user_host
    set  start  [string first "@" $addrstr]
    incr start
    set  end    [string length $addrstr]
    set  addr   [string trim [string range $addrstr $start $end]]

    set  dotted [host2ip $addr]

    return $dotted

}

#
# This function returns the user from a cname string
# received by a hello cmd
#
proc get_user_from_cname {user_host} {
    set  userstr $user_host
    set  end  [string first "@" $userstr]
    incr end -1
    set  user   [string trim [string range $userstr 0 $end]]

    return $user

}
     
#
# Session commands issued by media agents for session 's'.
# On error, delete session by calling 'close'.
# Some commands are also called locally.
#
proc session {s c args} {
    global conf session member Session rsvp flowspec Flowspec Rsvp \
	delay Ifloor message Message debug

    set ss [noperiod $s]
    set cs $conf(id)/$s
    if { $debug == "pmm" } {
	puts "incoming: $cs $c $args"
    }
    switch $c {
       hello {
            if { [regexp -nocase ifloor $s]} return
            set m [lindex $args 0]
            set mm [noperiod $m]
            set member($m,cname) $m
            member_add $s $m

        }
        closed {
            if {[llength [join [lindex $args 0]]] > 0} {
                tk_dialog .d {Session Closed} [join [lindex $args 0]] \
                    error 0 {Ok}
            }
	    set session($s,closed) 1
            session_close $s
        }
        error {
            if {[tk_dialog .d {Error} [join [lindex $args 0] ": "] \
                     error 0 {Abort} {Continue}] == 0} {
                if {$session($s,active)} {
                    set session($s,active) 0
                    control_send "$cs close"
                }
            }
        }
        active {
            set session($s,active) 1
        }  
	created {
            after cancel $session($s,timeout)
            set conf($s,waiting) 1
            foreach e [array names Session] {
                control_send "$cs $e [list $session($s,$e)]"
            }
            control_send "$cs local $session($s,local)"
            control_send "$cs open $session($s,host)"
            set session($s,created) 1
            set conf($s,waiting) 0
            if { [regexp -nocase ifloor $s]} {
#isc:
                if {[lindex [split $session($s,type) /] 0] == "ifloor"} {
                    if [info exist Ifloor] {
                        config_read_ifloor $s
                    }
                }
            }
            #check if there is already an entry for this call
            if { [lsearch -exact $conf(sessions) $s] != -1 } {
                return
            }
            # pack the sessions entries
            set session($s,active) 1
            ladd conf(sessions) $s
            # keep track of the column of the entries without considering all
            # the sorting and removing of entries that conf(session) requires
            # note that as we set the members textual description to column 100
            # we might get into trouble if we joined and left more than 100
            # sessions during the lifetime of isc
            lappend conf(entry) $s
 
            #create the session entries
            if { [regexp -nocase reservation $s]} return
            if { ![regexp -nocase ifloor $s]} {
                session_entry $s
            }
            set M $Session(f_control).leave.menu
 
            if {![winfo exists $M.$ss]} {
                $M add command -label "$s" \
                    -command "session_delete $s"
            }
#isc:
            if {[lindex [split $session($s,type) /] 0] == "ifloor"} {
                config_read_ifloor $s
            }
 
            set session($s,timeout) \
                [after $delay(regestration) "set conf($s,waiting) -1"]
            tkwait variable conf($s,waiting)
            after cancel session($s,timeout)
            if { $conf($s,waiting) < 0 } {
                footer "Could not open Session:$s"
                after $delay(display) "footer {}"
            }
            set conf($s,waiting) 1
	}
        opened {
            set conf($s,waiting) 1
        }
        receiving {
            set conf(rx) [lindex $args 0]
        }
        send {
            regsub -all "{" $args "" args
            regsub -all "}" $args "" args
	    set m $session($s,cname)
	    set member($m,$s,rx) [lindex $args 0]
	    set session($s,send) [lindex $args 0]
 
        }
	talk { # expect values in the form of {1/0 {talker moderator}}
            regsub -all "{" $args "" args
            regsub -all "}" $args "" args
	    set value [lindex $args 0] 
	    for {set i 1} {$i<  [llength $args ]} {incr i} {
		set talker [lappend talker [lindex $args $i]]
	    }
	    # close all connections in case of changing the talker
	    if { $value== 1 } {
		set me $session($s,cname)
		foreach e $session($s,members) {
		    if {![string match $me $e]} {
			set m $member($e,cname)
			set member($m,$s,rx) 0
			set session($s,send) 0
			    control_send "$conf(id)/$s/$member($m,cname) receive 0" 
		    }
		}
	    }
 	    foreach m $talker {
		set member($m,$s,rx) 1
		control_send "$conf(id)/$s/$m receive $value " 	
	    }
	    if [info exists talker] {unset talker}
	}	        
	receive { 
	    # to allow ifloor to close all incoming connections
	    # ifloor sends receive -1
	    set value 1
	    if {[lindex $args 0] <=0 } {
		set value 0
	    }
	    foreach e $session($s,members) {
                set me $session($s,cname)
		foreach e $session($s,members) {
		    if {![string match $me $e]} {
			set m $member($e,cname)
			set member($m,$s,rx) $value
			set session($s,send) $value
			control_send "$conf(id)/$s/$member($m,cname) receive $value " 
		    }
                }
            }
        }
        flowspec {
            global flowspec
	    regsub -all {\{|\}} $args " " new_flow

	    if { [llength $new_flow] == 6 } {
		regsub -all {\.} "[lindex $new_flow 0]" _ src
		set flowspec($s,$src,r) [lindex $new_flow 1]
		set flowspec($s,$src,b) [lindex $new_flow 2]
		set flowspec($s,$src,p) [lindex $new_flow 3]
		set flowspec($s,$src,m) [lindex $new_flow 4]
		set flowspec($s,$src,M) [lindex $new_flow 5]
	    }

	    set ss_ $s
	    set ss [string trim [string range $ss_ [string first / $ss_] \
				     [string first . $ss_]] "/."]
	    set sss [string trim [string range $ss_ [string first / $ss_] \
				      [string first : $ss_]] "/:"]
	    if {[connection_is_multicast $sss]} { # Multicast ??
		set src $sss
		if { [llength $new_flow] == 6 } {
		    if {![info exists flowspec($s,$src,r)]} {
			set flowspec($s,$src,r) [lindex $new_flow 1]
			set flowspec($s,$src,b) [lindex $new_flow 2]
			set flowspec($s,$src,p) [lindex $new_flow 3]
			set flowspec($s,$src,m) [lindex $new_flow 4]
			set flowspec($s,$src,M) [lindex $new_flow 5]
			regsub -all {\.} "[lindex $new_flow 0]" \/ src2

		    } else {

			regsub -all {\.} "[lindex $new_flow 0]" \/ src2

			set r_new [lindex $new_flow 1]
			set flowspec($s,$src,r) \
			    [ merge_values $r_new $rsvp($s,name) "r" $s]
			set b_new [lindex $new_flow 2]
			set flowspec($s,$src,b) \
			    [ merge_values $b_new $rsvp($s,name) "b" $s]
			set p_new [lindex $new_flow 3]
			set flowspec($s,$src,p)\
			    [ merge_values $p_new $rsvp($s,name) "p" $s]
			set m_new [lindex $new_flow 4]
			set flowspec($s,$src,m) \
			    [ merge_values $m_new $rsvp($s,name) "m" $s]
			set M_new [lindex $new_flow 5]
			set flowspec($s,$src,M)\
			    [ merge_values $M_new $rsvp($s,name) "M" $s]

		    }
		}
		
	    }
	}
	message {
            set host $session($s,host)
            set port $session($s,port)
	    rsvp_message $args $rsvp($s,reserve_win)
        }
	
	# add src to Source list
	add_src {
	    set args_list [split $args " {}"]
	    set src [lindex $args_list 1]
	    set src_tmp $src
	    set host [host2ip [lindex $args_list 2]]
	    set host_tmp $host
	    set port [lindex $args_list 3]
	    regsub -all {\.} $host _ host
	    regsub -all {\.} $src _ src
	    if {[lsearch $rsvp($s,name) $src] == -1} {
		lappend rsvp($s,name) $src
	        foreach e [array names Flowspec] {
		    set flowspec($s,$src,$e) $Flowspec($e)
		}
	    }
	    set rsvp($s,flowspec$src) $Rsvp(flowspec)
	    set session($s,flowspec$src) $rsvp($s,flowspec$src)
	    configure_reservation_win_$rsvp($s,mode) $s $rsvp($s,reserve_win)
	}
        rm_src {
            set args_list [split $args " {}"]
            set src [lindex $args_list 1]
            set src_tmp $src
            set host [host2ip [lindex $args_list 2]]
            set port [lindex $args_list 3]
            regsub -all {\.} $host _ host
            regsub -all {\.} $src _ src
            set res [lsearch $rsvp($s,name) $src]
            if {$res != -1} {
                set rsvp($s,name) [lreplace $rsvp($s,name) $res $res]
            }
	    configure_reservation_win_delete $s [lindex $args_list 1]
        }
	netghost_path {
	    set session($s,netghost_path) $args
	}
	netghost_host {
	    set session($s,netghost_host) $args
	}
 
	default {
	}

    }
}



#
# Media type changes.
#
proc session_type {name element op} {
  global Session

  set M .add.config.menu
  .add.config.menu 
} 
#
# check if soem left overs from a previous session exist and delete them
# then create the session
#
proc session_create {{counter 0}} {
    global conf session Session 
    # Create unique session name from pid and multicast address/port.
    incr conf(count)
    set s $Session(type)/$Session(host):$Session(port)

    set cs $conf(id)/$s
    # Create handler for session.
    if {[info procs $cs] != ""} {
	isc_message "Previous session seems to have 
not been terminated properly.
Deleting that first"
	session_delete $s
    }
     session_create_actively $counter
    return $s
}    

#
# Given session information in Session(), open it.
# Return session name.
#
proc session_create_actively {{counter 0}} {
    global conf session Session Flowspec Tspec flowspec tspec rsvp Rsvp delay connect_env LAN_Default WAN_Default LAN_User WAN_User

    # Create unique session name from pid and multicast address/port.
    incr conf(count)
    set s $Session(type)/$Session(host):$Session(port)

    set cs $conf(id)/$s

    # Copy configuration to new session.
    foreach e [array names Session] {
	set session($s,$e) $Session($e) 
    }

    
    if {[string match reservation $Session(type)]} {
	regsub {(reservation\/)([0-9\.]+)(:.+)} $s {\2} host
	regsub {(reservation\/)([0-9\.]+:)(.+)} $s {\3} port
	set host [host2ip $host]
	regsub -all {\.} $host _ host
	set ses .$Session(type)_$host:$port
	foreach e [array names Tspec] {
	    set tspec($s,$e) $Tspec($e)
	}
	foreach e [array names Flowspec] {
	    set flowspec($s,$host,$e) $Flowspec($e)
	}
	set Rsvp(reserve) ""
	set rsvp($s,name) $rsvp(name)

    }
    
    # Create handler for session.
    eval "proc {$cs} {c args} {session $s \$c \$args}"
    
    # Tell pmm about new session and all potential members.
    control_send "+ $cs"
    control_send "+ ${cs}/*"
    
    # Ask media agent to create this session.
    control_send "session $cs"
    
    # Start timer that starts application if no "created" answer.
    # Calls session_create recursively.
    set session($s,created) 0
    set conf($s,waiting) 0
    set session($s,members) {}
    set session($s,timeout) [after $delay(local) "set conf($s,waiting) -2"]
    if { $counter == 0} {
	tkwait variable conf($s,waiting)
	after cancel $session($s,timeout)
	if { $conf($s,waiting) < 0 } {
	    session_noanswer $s
	}
    } else {
	tkwait variable conf($s,waiting)
	after cancel $session($s,timeout)
	if { $conf($s,waiting) < 0 } {
	    footer "Could create proberly a media agent for session:$s"
	    after $delay(display) "footer {}"
	}
	#set conf(waiting) 1
	
    }
    return $s
}

#
# Invoked if session has not been created. If application can be
# found, try session_create again.
# Called only if no timer, to avoid endless recursions.
#
proc session_noanswer {s} {
    global conf session delay
    
    footer "No answer to create request for $s"
    # Set up MIME string.
    set p "$session($s,type)/[lindex $session($s,media) 0]; \
          host=$session($s,host);port=$session($s,port);ttl=$session($s,ttl); \
          confid=$session($s,confid); conf=\"$conf(name)\"; \
          name=\"$session($s,name)\""
    
    # Try to find and start application.
    if [catch {set aid [application $p]} msg] {
	tk_dialog .d {Session Closed} $msg error 0 {Ok}
	session_delete $s
    } else {
	# no id value, thus not a pmm application; nothing further to do
	if {$aid == ""} {
	    footer "Started non-pmm application for $p"     
	} else {
	    set conf(waitfor) $aid
	    # give application time to wake up
	    footer "Waiting for $conf(waitfor)..."
	    set conf(started) 0
	    # Make sure we escape eventually - doesn't work.
	    set id [after $delay(regestration) "set conf(started) 0"]
	    tkwait variable conf(started)
	    after cancel $id
	    if {$conf(started) == 0} {
		if [info exists conf(waitfor)] {
		    unset conf(waitfor)
		}
		session_delete $s
		#error "Application did not register"
		footer "Application did not register"
		after $delay(regestration) "set conf(started) 0"
		return
	    } else {
		footer "Started: $conf(started)"
		if [info exists conf(waitfor)] {
		    unset conf(waitfor) 
		}
		after $delay(display) "footer {}"
		# create the session but with counter set to 1 to avoid 
		# endless loops
		session_create_actively 1
		return
	    }
	}
    }   
}

#
# Invoked when application hasn't started.
#
proc session_timeout {} {
  global conf
  set conf(started) 0
}

# session_close sends a `close',
# media agent (hopefully) sends a `closed' leads to session_delete
# if the agent does not send a closed we do the session_delete after 
# delay(local)
proc session_delete {s} {
    global conf session Session debug conf delay

    if { $debug == "pmm" } {
	puts stderr "deleleting session $s"
    }
    
    if { [info exists conf(id)] && $conf(id) != "" } {
	    # send a close message 
	control_send "$conf(id)/$s close"
    }

    # is there such a session?
    if { [lsearch -exact $conf(sessions) $s] <0 } return


    # wait until a closed message arrives -if none arrives the after should do
    # the job
    set session($s,closed) 0
    after $delay(local) "set session($s,closed) 1; session_close $s"
    tkwait variable session($s,closed)
}

#
# after closing the session, delete all members, close the windows
# and do some cleaning up
#
proc session_close { s } {
    global conf session Session debug conf delay
    
    # is there such a session?
    if { [lsearch -exact $conf(sessions) $s] <0 } return
    # Remove all session members.
    if [info exists session($s,members)] {
	foreach m $session($s,members) {
	    member_delete $s $m
	}
    }

    # check if there is an RSVP session attached to a media session and close 
    # it
    rsvp_attached_delete $s

    # check if there is an ifloor session attached to a media session and close
    # it
    ifloor_close $s



    #determine the column of the session state and delete its header
    if { [lsearch -exact $conf(entry) $s] >-1 } {
	set col [lsearch -exact $conf(entry) $s]
    } else {
	return
    }

    # configuration window
    regsub -all {\.} $s {_} w_child
    regsub -all {/} $w_child {_} w_child
    catch { .configure.all  delete $w_child}

    set ss [noperiod $s]
    set M $Session(f_control).leave.menu
    
    catch {$M delete [$M index $s] }

    #determine the column of the session state and delete its header
    catch {destroy $Session(f_sessions).$col}

    #replace conf(entry) index for this session with a pseudo entry to keep
    # the right ordering
    set conf(entry) [lreplace $conf(entry) [lsearch -exact $conf(entry) $s] \
			 [lsearch -exact $conf(entry) $s] xxxx]
    
    # Remove all members of that session.
    lremove conf(sessions) $s
    # Unblock any tkwaits waiting for session creation, e.g., 
    # when creating several applications from commandline arguments.
    set session($s,created) 0
     
    # close all windows the might be related to this session
    set w .$session($s,type)_[join [split [lindex [split $s "/"] 1] "."]  "_"] 
    if [winfo exists $w] { 
	catch {destroy $w }
    }
    # close all windows the might be related to this session
    set w .$session($s,type)_[noperiod $s]
    if [winfo exists $w] { 
	catch {destroy $w }
    }
    foreach e [array names session] {
	if {[string match $s,* $e]} {
	    unset session($e)
	}
    }

    #remove the sessions name from the proc list
    rename $conf(id)/$s {}
}


#
# Invoked when new agent starts up. Sets variable conf(started) to
# agent's identifier.
#
proc agent {args} {
  global conf
  # check whether it's the right application
  if {[info exists conf(waitfor)]} {
    if {[string match $conf(waitfor) [lindex $args 0]]} {
      set conf(started) [lindex $args 0]
    }
  }
}


