#
# sip_phonebook.tcl
# simple editor to modify ~/.sip/phonebook 
#



proc sip_phonebook {} {
  global image env selected phonebook invite

  # create phonebook window
  set w .phonebook
  if [winfo exists $w] return
  toplevel $w
  set title "SIP Phonebook"
  wm title $w $title
  frame $w.title
  pack \
    [label $w.title.im -image $image(phonebook) -anchor center] \
    [label $w.title.title -text $title -font *-*-*-bold-i-*-18-*-*-*-*-*-*-*] \
    -side left -anchor w

  frame $w.top -relief ridge -borderwidth 4

  foreach s {left right} {
  frame $w.top.$s
  frame $w.top.$s.box
  frame $w.top.$s.buttons
  pack \
    [listbox $w.top.$s.box.lb -yscrollcommand "$w.top.$s.box.scroll set"] \
    [scrollbar $w.top.$s.box.scroll -command "$w.top.$s.box.lb yview"] \
    -side left -fill y
  }

  pack \
    [button $w.top.left.buttons.add -text Add -font *-*-*-*-*-*-10-*-*-*-*-*-*-* -command add_alias] \
    [button $w.top.left.buttons.remove -text Remove -font *-*-*-*-*-*-10-*-*-*-*-*-*-* -command remove_alias] \
      [button $w.top.left.buttons.invite -text Invite -font *-*-*-*-*-*-10-*-*-*-*-*-*-* -command {append invite(guest) "[string trim $selected(alias)]"}] \
    -side left -expand 1

  pack \
    [label $w.top.left.l -text Aliases] \
    $w.top.left.box \
    [entry $w.top.left.entry -textvariable selected(alias)] \
    $w.top.left.buttons \
    -side top -fill x -anchor nw

  pack \
    [button $w.top.right.buttons.add -text Add -font *-*-*-*-*-*-10-*-*-*-*-*-*-* -command add_addr] \
    [button $w.top.right.buttons.remove -text Remove -font *-*-*-*-*-*-10-*-*-*-*-*-*-* -command remove_addr] \
    -side left -expand 1

  pack \
    [label $w.top.right.l -text Addresses] \
    $w.top.right.box \
    [entry $w.top.right.entry -textvariable selected(addr)] \
    $w.top.right.buttons \
    -side top -fill x

  pack \
    $w.top.left \
    $w.top.right \
    -side left -padx 2m -pady 1m

  frame $w.buttons

  pack \
    [button $w.buttons.save   -text "Save" -width 15 -command {save_phonebook}] \
    [button $w.buttons.dismiss -text "Dismiss" -width 15 -command "save_check phonebook"] \
    -side left

  pack $w.title 
  pack $w.top -padx 1m -pady 1m -fill x 
  pack $w.buttons

  # open and read phonebook file
  set phonebook(entries) [read_file .sipphonebook]
  set phonebook(modified) 0

  # get aliases
  foreach line $phonebook(entries) {
    $w.top.left.box.lb insert end [lindex [split $line ":"] 0]
  }

  bind $w.top.left.box.lb <ButtonRelease-1> {
    set selected(alias) [%W get [%W nearest %y]]
    alias_selection
  }

  bind $w.top.right.box.lb <ButtonRelease-1> {
    set selected(addr) [%W get [%W nearest %y]]
  }
} ;# sip_phonebook



# invoked if an alias was selected
proc alias_selection {} {
  global selected phonebook

  .phonebook.top.right.box.lb delete 0 end
  # search selected alias and display aliased addresses
  set line [lindex $phonebook(entries) [lsearch  $phonebook(entries) "$selected(alias):*"]]
  set addresses [lindex [split $line ":"] 1]

  foreach addr $addresses {
   .phonebook.top.right.box.lb insert end $addr
  }
} ;# alias_selection
    


# add new alias to phonebook
proc add_alias {} {
  global selected phonebook

  if {($selected(alias) == "") || ([lsearch $phonebook(entries) "$selected(alias):*"] >= 0)} return

  set phonebook(modified) 1

  lappend phonebook(entries) "$selected(alias):"
  .phonebook.top.left.box.lb insert end $selected(alias)
  .phonebook.top.right.box.lb delete 0 end
} ;# add_alias



# remove alias from phonebook
proc remove_alias {} {
  global selected phonebook

  set index [lsearch $phonebook(entries) "$selected(alias):*"]
  if {$index < 0} return
    
  set phonebook(modified) 1
    
  set phonebook(entries) [lreplace $phonebook(entries) $index $index]
  .phonebook.top.left.box.lb delete $index $index
  .phonebook.top.right.box.lb delete 0 end
} ;# remove_alias



# add new address to alias
proc add_addr {} {
  global selected phonebook

  if {($selected(addr) == "") || ($selected(alias) == "")} return
    
  set phonebook(modified) 1
    
  set index [lsearch $phonebook(entries) "$selected(alias):*"]
  if {$index < 0} {
    add_alias
  } else {
    set phonebook(entries) [lreplace $phonebook(entries) $index $index "[lindex $phonebook(entries) $index] $selected(addr)"]
    .phonebook.top.right.box.lb insert end $selected(addr)
  }
} ;# add_addr



# remove address from alias
proc remove_addr {} {
  global selected phonebook

  set phonebook(modified) 1

  set alias_index [lsearch  $phonebook(entries) "$selected(alias):*"]
  set line [lindex $phonebook(entries) $alias_index]
  set addresses [lindex [split $line ":"] 1]
  set index [lsearch  $addresses "$selected(addr)"]
  if {$index < 0} return

  set addresses [lreplace $addresses $index $index]
  set phonebook(entries) [lreplace $phonebook(entries) $alias_index $alias_index "$selected(alias):$addresses"]

  .phonebook.top.right.box.lb delete $index $index
} ;# remove_addr



proc save_phonebook {} {
  global phonebook

  set phonebook(modified) 0
  save_file .sipphonebook $phonebook(entries)
  destroy .phonebook
} ;# save_phonebook



