#
# Create statistics pop-up window for member 'm' of session 's'.
# 
# $Id: statistics.tcl,v 1.5 1998/09/04 15:20:56 dor Exp $
#


# Create row with labels.
proc l_row {w f y text} {
  upvar y r
  grid  configure [label $w.${f}_ -text $text] -row $r -column 0 -sticky w
  grid  configure [label $w.$f] -row $r -column 1 -sticky w -columnspan 4
  incr r
}

# Create row with frame.
proc f_row {w f y text} {
  upvar y r
  grid configure  [label $w.${f}_ -text $text] -row $r -column 0 -sticky w
  grid configure [frame $w.$f] -row $r -column 1 -sticky w -columnspan 4
  incr r
}

# Create row with two values.
proc p_row {w f y text {text3 " "} {text4 " "}} {
  upvar y r
  grid configure \
    [label $w.${f}0 -text $text] -row $r -column 0 -sticky  w 
  grid configure \
    [label $w.${f}3 -text $text3 -width 8 -anchor e] -row $r -column 3 -sticky  w 
  grid configure \
    [label $w.${f}4 -text $text4 -width 8 -anchor e] -row $r -column 4 -sticky w
  incr r
}

# Create row with four values and graphics checkbox.
proc g_row {m s w f y text} {
  upvar y r
  grid configure\
    [checkbutton $w.$f -text $text -relief flat \
       -variable member($m,$s,G,$f) \
       -command "stats_graph $s $m $f y 30"] -row $r -column 0 -sticky w 

  grid configure\
    [label $w.${f}1 -width 6 -anchor e] -row $r -column 1 -sticky w 

  grid configure\
    [label $w.${f}2 -width 6 -anchor e] -row $r -column 2 -sticky w 

  grid configure\
    [label $w.${f}3 -width 8 -anchor e] -row $r -column 3 -sticky  w 

  grid configure\
    [label $w.${f}4 -width 8 -anchor e] -row $r -column 4 -sticky  w
  incr r
}

proc statistics {s m stats} {
    global conf member top statis
    # if stats aren't enabled, we return immediately; this is the
    # only way to escape the periodic statistics query
    if {$member($m,$s,stats) == 0} return
    # window name for statistics
    set w .[noperiod $m],[noperiod $s]
    if {[winfo exists $w] == 0} {
	option add $top${w}.*font fixed
	toplevel $w
	wm title $w $member($m,cname) 
	set w_all $w
	# create a frame for the upper part of the window
	set w_upper [frame $w_all.upper -relief groove -borderwidth 2]

	# create a frame for the sdes part
	set w_upper_left [frame $w_upper.left]
	tixLabelFrame  $w_upper_left.sdes -label "Source Description"
	set w_sdes [$w_upper_left.sdes subwidget frame]
	set w $w_sdes
	set y 0
	l_row $w session  y "Session"
	set statis($s,$m,session) $w_sdes.session
	l_row $w name     y "User name"
	set statis($s,$m,name) $w_sdes.name
	l_row $w created  y "Created"
	set statis($s,$m,created) $w_sdes.created

	$w.session config -text $s
	$w.name    config -textvariable member($m,name)
	
	#create frame for the packet information
	tixLabelFrame $w_upper.packet -label "Packet Information"
	set w_packet [$w_upper.packet subwidget frame]
	set w $w_packet
	l_row $w last     y "Last packet"
	set statis($s,$m,last) $w_packet.last
	l_row $w media    y "Media"
	set statis($s,$m,media) $w_packet.media
	p_row $w heading  y "Parameter" "new" "total"
	set statis($s,$m,heading) $w_packet.heading
	p_row $w control  y "Control packets"
	set statis($s,$m,control) $w_packet.control
	p_row $w expected y "Expected"
	set statis($s,$m,expected) $w_packet.expected
	p_row $w received y "Received"
	set statis($s,$m,received) $w_packet.received
	p_row $w good     y "Good packets"
	set statis($s,$m,good) $w_packet.good
	p_row $w marker   y "Marked packets"
	set statis($s,$m,marker) $w_packet.marker
	p_row $w bad_hdr  y "Bad header"
	set statis($s,$m,bad_hdr) $w_packet.bad_hdr
	p_row $w bad_fmt  y "Bad Format"
	set statis($s,$m,bad_fmt) $w_packet.bad_fmt
	p_row $w samples  y "Media samples"
	set statis($s,$m,samples) $w_packet.samples
	l_row $w reserved  y "Reserved"
	set statis($s,$m,reserved) $w_packet.reserved

	pack $w_upper_left  $w_upper.packet -side left -expand yes -fill both

	# create a frame for the lower part of the window
	set w_lower [frame $w_all.lower -relief groove -borderwidth 2]

	# create a frame for the qos part
	tixLabelFrame  $w_upper_left.qos -label "QoS Description"
	set w_qos [$w_upper_left.qos subwidget frame]
	set w $w_qos	
	grid  configure\
	    [checkbutton $w.jitter -text "Jitter" -relief flat \
		 -variable member($m,$s,G,jitter) \
		 -command "stats_graph $s $m jitter y2 30"] \
	    -row $y -column 0 -sticky w
	set statis($s,$m,jitter) $w_qos.jitter

	grid  configure\
	    [label $w.jitter4 -width 8 -anchor e] \
	    -row $y -column 4 -sticky w
	set statis($s,$m,jitter)4 $w_qos.jitter4
	incr y
	
	grid configure\
	    [checkbutton $w.rr_jitter -text "RR_Jitter" -relief flat \
		 -variable member($m,$s,G,rr_jitter) \
		 -command "stats_graph $s $m rr_jitter y2 30"] \
	    -row $y -column 0 -sticky w 
	set statis($s,$m,rr_jitter) $w_qos.rr_jitter
	
	grid  configure\
	    [label $w.rr_jitter4 -width 8 -anchor e] -row $y \
	    -column 4 -sticky  w
	set statis($s,$m,rr_jitter)4 $w_qos.rr_jitter4
	incr y
	
	grid configure\
	    [checkbutton $w.delay -text "Playout Delay" -relief flat \
		 -variable member($m,$s,G,delay) \
		 -command "stats_graph $s $m delay y2 30"] \
	    -row $y -column 0 -sticky w 
	set statis($s,$m,delay) $w_qos.delay
	
	grid  configure\
	    [label $w.delay4 -width 8 -anchor e] -row $y \
	    -column 4 -sticky  w
	set statis($s,$m,delay)4 $w_qos.delay4
	incr y
	grid  configure\
      [checkbutton $w.actual_framerate -text "Frame Rate" -relief flat \
	   -variable member($m,$s,G,actual_framerate) \
	   -command "stats_graph $s $m actual_framerate y2 30"] \
	    -row $y -column 0 -sticky  w 
	set statis($s,$m,actual_framerate) $w_qos.actual_framerate
	
	grid  configure\
	    [label $w.actual_framerate4 -width 15 -anchor e]   \
	    -row $y -column 1 -sticky  w 
	set statis($s,$m,actual_framerate)4 $w_qos.actual_framerate4
	
	set statis($s,$m,framerate)4 $w_qos.framerate4
	incr y

	grid configure \
	    [checkbutton $w.actual_bandwidth -text "Bandwidth" -relief flat \
		 -variable member($m,$s,G,actual_bandwidth) \
		 -command "stats_graph $s $m actual_bandwidth y2 30"]\
	    -row $y -column 0 -sticky  w 
	set statis($s,$m,actual_bandwidth) $w_qos.actual_bandwidth
	
	grid  configure\
	    [label $w.actual_bandwidth4 -width 15 -anchor e]   \
	-row $y -column 1 -sticky  w 
	set statis($s,$m,actual_bandwidth)4 $w_qos.actual_bandwidth4
	
	
	pack $w_upper_left.sdes $w_upper_left.qos -side top -expand yes -fill both 

	# create a frame for the part describing loss of the window
	set w_lower1 [frame $w_all.lower1 -relief groove -borderwidth 2]

	# create a frame for the qos part
	tixLabelFrame  $w_lower1.loss -label "Loss Statistics"
	set w_loss [$w_lower1.loss subwidget frame]
	set w $w_loss
	grid  configure\
	    [label $w.pheading0 -text "Parameter"] -row $y -column 0 -sticky w 
	set statis($s,$m,pheading)0 $w_loss.pheading0
	
	grid  configure\
	   [label $w.pheading1 -text "%" -width 6 -anchor e] -row $y -column 1\
	    -sticky w 
	set statis($s,$m,pheading)1 $w_loss.pheading1
	
	grid  configure\
	    [label $w.pheading2 -text "total %" -width 8 -anchor e] -row $y \
	    -column 2 -sticky w 
	set statis($s,$m,pheading)2 $w_loss.pheading2
	
	grid  configure\
	    [label $w.pheading3 -text "count" -width 8 -anchor e] -row $y \
	    -column 3 -sticky w 
	set statis($s,$m,pheading)3 $w_loss.pheading3
	
	grid  configure\
	    [label $w.pheading4 -text "total" -width 8 -anchor e] -row $y \
	    -column 4 -sticky w 
	set statis($s,$m,pheading)4 $w_loss.pheading4
	
	incr y

	g_row $m $s $w lost      y "Lost"
	set statis($s,$m,lost) $w_loss.lost
	g_row $m $s $w rr_lost   y "RR_Lost"
	set statis($s,$m,rr_lost) $w_loss.rr_lost
	g_row $m $s $w late      y "Late"
	set statis($s,$m,late) $w_loss.late
	g_row $m $s $w dup       y "Duplicate"
	set statis($s,$m,dup) $w_loss.dup
	g_row $m $s $w order     y "Out of order"
	set statis($s,$m,order) $w_loss.order

	pack $w_lower1.loss -side left -expand yes -fill both 



	pack $w_upper     $w_lower1\
	    [button $w_all.dismiss -text "Dismiss" \
		 -command "set member($m,$s,stats) 0; \
                 trace vdelete member($m,img) w img_trace;  \
                 destroy $w_all; destroy_graph $s $m"] \
	    -side top -expand yes -fill both

	foreach p \
	    {bad_fmt bad_hdr control dup expected good late rr_lost \
		 rr_fraction lost marker order received samples} {
		set member($m,$s,$p) 0
	    }
	foreach p {dup late rr_lost rr_fraction lost order delay \
		       jitter rr_jitter actual_bandwidth average_bandwidth \
		       actual_framerate } {
	    set member($m,$s,G,$p) 0
	}
	set member($m,$s,graphs) 0
	
	# trace images
	trace variable member($m,$s,img) w img_trace
	# initial display of images
	if {[info exist member($m,$s,img)]} {
	    img_trace member $m,$s,img w
	}
    }
 
    # update statistics, except for last
    
    foreach e $stats {
 	set p [lindex $e 0]
	set v [lindex $e 1]
 	switch $p {
	    created {
		$statis($s,$m,created) configure -text [strftime $v]
	    }
	    reserved {
		$statis($s,$m,reserved) configure -text $v
	    }
	    last {
		set f {}
		foreach e $v {
		   if {$e == 0} {lappend f "none"} {lappend f [strftime $e %T]}
		}
		$statis($s,$m,last) configure -text [join $f]
	    }
	    bad_fmt -
	    bad_hdr -
	    control -
	    expected -
	    good -
	    received - 
	    marker -
	    samples {
		set d [expr $v - $member($m,$s,$p)]
		if {$p == "received" } {
		    global newly_received 
		    set newly_received [expr $v - $member($m,$s,received)]
		}
		$statis($s,$m,${p})3 configure -text $d
		$statis($s,$m,${p})4 configure -text $v
		set member($m,$s,$p) $v
	    }
	    dup -
	    late -
	    lost -
	    order {
		global newly_received 
		set d [expr $v - $member($m,$s,$p)]
		set member($m,$s,$p) $v
		if {[info exists newly_received]} {
		    if {$newly_received >0} {
			set f [format "%0.3g" [expr ($d * 100.0) / \
						   ($d + $newly_received)]]
			$statis($s,$m,${p})1 configure -text $f
			if {$member($m,$s,G,$p)} {stats_draw $s $m $p $f}
		    }
		} else {
		    $statis($s,$m,${p})1 configure -text 0
		}
		if {$member($m,$s,received) > 0} {
		    set f [format "%0.3g" [expr ($v * 100.0)/ \
					       ($v + $member($m,$s,received))]]
		    $statis($s,$m,${p})2 configure -text $f
		} else {
		    
		    $statis($s,$m,${p})2 configure -text 0
		}
		$statis($s,$m,${p})3 configure -text $d
		$statis($s,$m,${p})4 configure -text $v
	    }
	    rr_fraction {
		global fraction
		set fraction $v 
	    }
	    rr_lost {
		global fraction
		set d [expr $v - $member($m,$s,$p)]
		set member($m,$s,$p) $v
		if {[info exists fraction]} {
		    set f [format "%2.2f" [expr ($fraction * 100.0) ] ]
		    $statis($s,$m,${p})1 configure -text $f
		    if {$member($m,$s,G,$p)} {stats_draw $s $m $p $f}
		} else {
		    $statis($s,$m,${p})1 configure -text 0
		}
		$statis($s,$m,${p})3 configure -text $d
		$statis($s,$m,${p})4 configure -text $v
	    }
	    jitter {
		$statis($s,$m,${p})4 configure -text $v
		if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
	    }
	    rr_jitter {
		$statis($s,$m,${p})4 configure -text $v
		if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
	    }
	    delay {
		$statis($s,$m,${p})4 configure -text $v
		if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
	    }
	    actual_bandwidth {
		$statis($s,$m,${p})4 configure -text $v
		if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
	    }
	    actual_framerate {
		$statis($s,$m,${p})4 configure -text $v
		if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
	    }
	    default {
		catch {$w.${p} configure -text $v}
	    }
	}
    }
  after [expr $conf(stats_period)*1000] stats_request $s $m
}
proc statistics_1 {s m stats} {
    global conf member top
    # if stats aren't enabled, we return immediately; this is the
    # only way to escape the periodic statistics query
    if {$member($m,$s,stats) == 0} return
    # window name for statistics
    set w .[noperiod $m],[noperiod $s]
    if {[winfo exists $w] == 1} {
	set member($m,$s,stats) 0; 
	trace vdelete member($m,img) w img_trace;  
	destroy $w; 
	destroy_graph $s $m
    }
    
    if {[winfo exists $w] == 0} {
	option add $top${w}.*font fixed
	toplevel $w
	wm title $w $member($m,cname) 
	
	# creare a frame for the upper part of the window
	set w_upper [frame $w.upper -relief groove -bandwidth 2]

	# create a frame for the sdes part
	tixLabelFrame $w.upper.sdes -label "Source Description"
	set w_sdes [$w.upper.sdes subwidget frame]
	set w $w_sdes
	set y 0
	l_row $w session  y "Session"
	l_row $w cname    y "Canonical name"
	l_row $w name     y "User name"
	l_row $w phone    y "Phone"
	l_row $w email    y "Email"
	l_row $w source   y "source"
	l_row $w tool    y "Tool"
	f_row $w img      y "Image"
	l_row $w note     y "Message"
	l_row $w from     y "Network source"
	l_row $w created  y "Created"

	
	tixLabelFrame $w.upper.packet -label "Packet Information"
	set w_packet [$w.upper.packet subwidget frame]
	set w $w_packet
	l_row $w last     y "Last packet"
	l_row $w media    y "Media"
	p_row $w heading  y "Parameter" "new" "total"
	p_row $w control  y "Control packets"
	p_row $w expected y "Expected"
	p_row $w received y "Received"
	p_row $w good     y "Good packets"
	p_row $w marker   y "Marked packets"
	p_row $w bad_hdr  y "Bad header"
	p_row $w bad_fmt  y "Bad Format"
	p_row $w samples  y "Media samples"
	l_row $w reserved  y "Reserved"

    grid  configure\
      [checkbutton $w.jitter -text "Jitter" -relief flat \
       -variable member($m,$s,G,jitter) \
       -command "stats_graph $s $m jitter y2 30"] -row $y -column 0 -sticky w
 
    grid  configure\
      [label $w.jitter4 -width 8 -anchor e] -row $y -column 4 -sticky w
    incr y

    grid configure\
      [checkbutton $w.rr_jitter -text "RR_Jitter" -relief flat \
       -variable member($m,$s,G,rr_jitter) \
       -command "stats_graph $s $m rr_jitter y2 30"] \
	-row $y -column 0 -sticky w 

    grid  configure\
      [label $w.rr_jitter4 -width 8 -anchor e] -row $y -column 4 -sticky  w

    incr y
puts $y
    grid  configure\
      [checkbutton $w.delay -text "Playout delay" -relief flat \
       -variable member($m,$s,G,delay) \
       -command "stats_graph $s $m delay y2 30"] -row $y -column 0 -sticky w 
    incr y

    grid  configure\
      [checkbutton $w.actual_framerate -text "Frame Rate" -relief flat \
       -variable member($m,$s,G,actual_framerate) \
       -command "stats_graph $s $m actual_framerate y2 30"] \
	-row $y -column 0 -sticky  w 

    grid  configure\
      [label $w.actual_framerate4 -width 15 -anchor e]   \
	-row $y -column 1 -sticky  w 

    grid  configure\
      [checkbutton $w.framerate -text "Mean Frame Rate" -relief flat \
       -variable member($m,$s,G,framerate) \
	   -command "stats_graph $s $m  framerate y2 30"] \
	-row $y -column 2 -sticky  w 

    grid  configure\
      [label $w.framerate4 -width 15 -anchor e]   -row $y -column 4 -sticky  w 
    incr y
    grid configure \
      [checkbutton $w.actual_bandwidth -text "Bandwidth" -relief flat \
       -variable member($m,$s,G,actual_bandwidth) \
	   -command "stats_graph $s $m actual_bandwidth y2 30"]\
	-row $y -column 0 -sticky  w 

    grid  configure\
      [label $w.actual_bandwidth4 -width 15 -anchor e]   \
	-row $y -column 1 -sticky  w 

    grid  configure\
      [checkbutton $w.bandwidth -text "Mean Bandwidth" -relief flat \
       -variable member($m,$s,G,bandwidth) \
       -command "stats_graph $s $m  bandwidth y2 30"] \
	-row $y -column 2 -sticky  w 

    grid  configure\
      [label $w.bandwidth4 -width 15 -anchor e]   -row $y -column 4 -sticky  w 

    incr y

    grid  configure\
      [label $w.delay4 -width 8 -anchor e] -row $y -column 4 -sticky w
    incr y


    grid  configure\
      [label $w.pheading0 -text "Parameter"] -row $y -column 0 -sticky w 

    grid  configure\
      [label $w.pheading1 -text "%" -width 6 -anchor e] -row $y -column 1 \
	-sticky w 

    grid  configure\
      [label $w.pheading2 -text "total %" -width 8 -anchor e] -row $y \
	-column 2 -sticky w 

    grid  configure\
      [label $w.pheading3 -text "count" -width 8 -anchor e] -row $y \
	-column 3 -sticky w 

    grid  configure\
      [label $w.pheading4 -text "total" -width 8 -anchor e] -row $y \
	-column 4 -sticky w 

    incr y

    g_row $m $s $w lost      y "Lost"
    g_row $m $s $w rr_lost   y "RR_Lost"
    g_row $m $s $w late      y "Late"
    g_row $m $s $w dup       y "Duplicate"
    g_row $m $s $w order     y "Out of order"

    grid  configure\
      [button $w.dismiss -text "Dismiss" \
       -command "set member($m,$s,stats) 0; trace vdelete member($m,img) w img_trace;  destroy $w; destroy_graph $s $m"] \
       -column $y -row 0 -columnspan 5 -sticky w 
    incr y

    $w.session config -text $s
    $w.cname   config -textvariable member($m,cname)
    $w.name    config -textvariable member($m,name)
    $w.tool   config -textvariable member($m,tool)
    $w.email   config -textvariable member($m,email)
    $w.source  config -textvariable member($m,$s,source)
    $w.note    config -textvariable member($m,$s,note)
    $w.from    config -textvariable member($m,$s,from)

    foreach p \
    {bad_fmt bad_hdr control dup expected good late rr_lost rr_fraction lost marker order received samples} {
      set member($m,$s,$p) 0
    }
    foreach p {dup late rr_lost rr_fraction lost order delay jitter rr_jitter actual_bandwidth average_bandwidth actual_framerate } {
      set member($m,$s,G,$p) 0
    }
    set member($m,$s,graphs) 0

    # trace images
    trace variable member($m,$s,img) w img_trace
    # initial display of images
    if {[info exist member($m,$s,img)]} {
      img_trace member $m,$s,img w
    }
  }
  # update statistics, except for "last"
  foreach e $stats {
    set p [lindex $e 0]
    set v [lindex $e 1]
    switch $p {
      created {
        $w.created configure -text [strftime $v]
      }
      reserved {
        $w.reserved configure -text $v
      }
      last {
        set f {}
        foreach e $v {
          if {$e == 0} {lappend f "none"} {lappend f [strftime $e %T]}
        }
        $w.last configure -text [join $f]
      }
      bad_fmt -
      bad_hdr -
      control -
      expected -
      good -
      received - 
      marker -
      samples {
        set d [expr $v - $member($m,$s,$p)]
        if {$p == "received"} {
          global newly_received 
          set newly_received [expr $v - $member($m,$s,received)]
        }
        $w.${p}3 configure -text $d
        $w.${p}4 configure -text $v
        set member($m,$s,$p) $v
      }
      dup -
      late -
      lost -
      order {
       	global newly_received 
	set d [expr $v - $member($m,$s,$p)]
	set member($m,$s,$p) $v
	if {[info exists newly_received]} {
	  if {$newly_received >0} {
	    set f [format "%0.3g" [expr ($d * 100.0) / ($d + $newly_received)]]
		  $w.${p}1 configure -text $f
		  if {$member($m,$s,G,$p)} {stats_draw $s $m $p $f}
	    }
	} else {
	      $w.${p}1 configure -text 0
	}
	if {$member($m,$s,received) > 0} {
          set f [format "%0.3g" [expr ($v * 100.0)/ ($v + $member($m,$s,received))]]
          $w.${p}2 configure -text $f
        } else {

          $w.${p}2 configure -text 0
        }
        $w.${p}3 configure -text $d
        $w.${p}4 configure -text $v
      }
      rr_fraction {
	  global fraction
          set fraction $v 
      }
      rr_lost {
	global fraction
        set d [expr $v - $member($m,$s,$p)]
        set member($m,$s,$p) $v
	  if {[info exists fraction]} {
	      set f [format "%2.2f" [expr ($fraction * 100.0) ] ]
	      $w.${p}1 configure -text $f
        } else {
          $w.${p}1 configure -text 0
	}
        $w.${p}3 configure -text $d
        $w.${p}4 configure -text $v
      }
      jitter {
        $w.${p}4 configure -text $v
        if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
      }
      rr_jitter {
        $w.${p}4 configure -text $v
        if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
      }
      delay {
        $w.${p}4 configure -text $v
        if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
      }
      actual_bandwidth {
        $w.${p}4 configure -text $v
        if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
      }
      actual_framerate {
        $w.${p}4 configure -text $v
        if {$member($m,$s,G,$p)} {stats_draw $s $m $p $v}
      }
      default {
        catch {$w.${p} configure -text $v}
      }
    }
  } 

  after [expr $conf(stats_period)*1000] stats_request $s $m
}


#
# If not dismissed by now, request statistics.
# Invoked periodically.
#
proc stats_request {s m} {
  global conf member session
    if {[info exists member($m,$s,alive)] && [info exists session($s,active)] } {
    control_send "$conf(id)/$s/$member($m,cname) statistics?"
  }
}

proc statistics_delete { s m } {
    set mm [noperiod $m]
    set ss [noperiod $s]

    catch {destroy .$mm,$ss}
     destroy_graph $s $m
}

