/*  @(#)results.c	1.5  97/12/23  */

/*
 * Copyright (c) Communications Research Centre, Government of Canada 1997. 
 * All rights reserved.
 * 
 * License is granted to copy and use this software, for research and
 * evaluation purposes, provided that the Communications Research Centre
 * is acknowledged in all documentation pertaining to any such copy or
 * use. The Communications Research Centre grants no other licenses
 * expressed or implied. Title, ownership rights, and intellectual
 * property rights in the software shall remain with the Communications
 * Research Centre.  The Centre's name and the Government of Canada
 * should not be used in any advertising without written permission.
 * 
 * THE COMMUNICATIONS RESEARCH CENTRE MAKES NO REPRESENTATIONS CONCERNING
 * EITHER THE MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF
 * THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided
 * "as is" without express or implied warranty of any kind.
 * 
 * These notices must be retained in any copies of any part of this
 * software.
 */


#include "mpoll.h"

int countlist[MAX_RESPONSES+1];
int slice, bar, space, width;
int xorigin = 20; /* where x axis starts */
time_t now, ntpnow;
long int tdiff;
Tcl_TimerToken results_timer_token;
Tcl_TimerToken member_timer_token;

/* this only handles simple bargraph presentations */


/* -------------------------------------------------------------- */
/* create the results GUI for histograms */

showresults(int thistopic)
   {
   int x, i, j;
   int localnum;
   char *txtmp = NULL;
   int totalpix = 270; 
      /* how much of the X axis in the canvas we can work with */


   /* get the label information for this question */
   strcpy(current_summary, topiclist[thistopic].summary);
   strcpy(current_question, topiclist[thistopic].question);
   strcpy(current_author, topiclist[thistopic].email);
   strcpy(current_option, topiclist[thistopic].option);
   
   Tcl_DeleteTimerHandler(results_timer_token);
   
   /* if a results window exists, get rid of it */
   if (ET_INT(winfo exists .rmain) != 0)
      {
      ET(destroy .rmain);
      }
      
   if (ET_INT(winfo exists .textmain) != 0)
      {
      ET(destroy .textmain);
      }
      
   if (strcmp(current_option, "text") == 0)
      {
      ET_INCLUDE( string-results.tcl );
      
      update_string_results(thistopic);      
      return TCL_OK;
      }
   
         
   /* create the GUI */
   ET_INCLUDE( results.tcl );
   
   /* assume that we have all possible responses until 
      we know otherwise */
   localnum = MAX_RESPONSES;
      
   /* now find out how many responses really for this data set */
   for (j = 1; j <= MAX_RESPONSES; j++)
      {
      if (topiclist[thistopic].response[j] == NULL)
         {
         localnum = j - 1;
         break;
         }
      }	
   
   /* calculate the parameters we need */
   slice = totalpix / localnum;
   bar = slice * .67;
   space = slice - bar;
   width = bar / 2;
   
/*
 *    printf("Using these parms: num bars=%d, slice=%d, bar=%d, space=%d, width=%d\n",
 *       localnum, slice, bar, space, width); 
 */
   
   /* label the X axis */
   for (i = 1; i <= localnum; i++)
      {
      x = xorigin + space + width + ((i - 1) * (slice));
      if ((topiclist[thistopic].response[i]) != NULL)
         {
         txtmp = strdup(topiclist[thistopic].response[i]);
         }
      if (! txtmp)
         {
         return(0);
         }
      ET(
         .rmain.canvasframe.area create text %d(x) 179  \
            -text  "%q(txtmp)" \
            -font "%q(default_font)" \
            -anchor n \
            -width %d(slice) \
            -justify left
         .rmain.canvasframe.area create line %d(x) 175 %d(x) 172 -width 1
      );
      free(txtmp);
      }
   
   /* draw the data bars */
   drawdata(thistopic);
   
   /* update the Tk display */
   ET(update idletasks);
   
   /* arrange to update the results periodically */
   update_results(thistopic);
   
   return(0);     
   }
   
   
/* -------------------------------------------------------------- */
/* update the results display periodically */

update_results(int thistopic)
   {
   int status;
   int interval;
   
   
   /* check that we still have a results window */
   status = ET_INT(winfo exists .rmain);
   if (! status)
      {
      Tcl_DeleteTimerHandler(results_timer_token);
      return(0);
      }
      
   /* get rid of the old data */
   ET(.rmain.canvasframe.area delete update);
   
   /* and put in the new data */
   drawdata(thistopic);
   
   /* and update the Tk display */
   ET(update idletasks);
   
   /* handler uses msecs so ... */
   interval = UPDATE_TIME * 1000; 
   
   /* use faster updates if this is a simulation */
   if (simulate)
      {
      interval = 1000;
      }
   
   /* infinite loop to this same function */
   results_timer_token =  Tcl_CreateTimerHandler(interval, 
     (Tk_TimerProc*)update_results, (ClientData)thistopic);
   
   }

/* -------------------------------------------------------------- */
/* update the display of the text results */
update_string_results(int thistopic)
   {
   int i;
   
   ET(
      .textmain.resultsframe.textwin delete 1.0 end
      );
      
   /* populate window with responses */
   /* .textmain.resultsframe.textwin */
   for (i = 1; i <= numresponses; i++)
      {
      /* match the ID */
      if (strcmp(topiclist[thistopic].qid, responselist[i].id) != 0)
         {
         continue;
         }

      /* check for decay */

      if (topiclist[thistopic].decay_time != 0)
         {
         time(&now);
         ntpnow = unixtontp(now);
         tdiff = ntpnow - responselist[i].time;

         /* life time expressed in minutes so multiply by 60 */
         if (tdiff > (topiclist[thistopic].decay_time * DECAY_FACTOR))
            {
            continue;
            }
         }

      ET(
         .textmain.resultsframe.textwin insert end "%q(responselist[i].response[1])\n"
         .textmain.resultsframe.textwin insert end "---\n"
         );
      }
            
   
   }   
/* -------------------------------------------------------------- */
/* counts up the results for a topic on the fly */
int *count_data(int tnum)
   {
   int i, j, k;
   char *thisresponse = NULL;
   static int simcountlist[MAX_TOPICS + 1][MAX_RESPONSES + 1];
   int simincrease;
   
   /* reset the counter to zeros */
   for (j = 0; j <= MAX_RESPONSES; j++)
      {
      countlist[j] = 0;
      }

   for (i = 1; i <= numresponses; i++)
      {
      /* match the ID */
      if (strcmp(topiclist[tnum].qid, responselist[i].id) != 0)
         {
         continue;
         }
      
      /* check for decay */
      
      if (topiclist[tnum].decay_time != 0)
         {
         time(&now);
         ntpnow = unixtontp(now);
         tdiff = ntpnow - responselist[i].time;

         /* life time expressed in minutes so multiply by 60 */
         if (tdiff > (topiclist[tnum].decay_time * DECAY_FACTOR))
            {
            continue;
            }
         }
      
      
      /* add up the count */
      
      for (j = 1; j <= MAX_RESPONSES; j++)
         {
         if ((responselist[i].response[j]) != NULL)
            {
            thisresponse = strdup(responselist[i].response[j]);
            for (k = 1; k <= MAX_RESPONSES; k++)
               {
               if (topiclist[tnum].response[k] != NULL)
                  {
                  if (strcmp(thisresponse, topiclist[tnum].response[k]) == 0)
                     {
                     /* found out where to put the data */
                     countlist[k]++;
                     }
                  }
               } /* k loop */
            free(thisresponse);
            }
         } /* j loop */
         
      countlist[0]++; /* count of number of responders */

      } /* end of response loop */
      
      
   /* added hack to simulate data for demo purposes */
   if (simulate)
      {
      for (k = 1; k <= MAX_RESPONSES; k++)
         {
         if (topiclist[tnum].response[k] != NULL)
            {
            /* create a bias in the simulated results */
            simincrease = rand() / (RAND_MAX / MAX_RESPONSES - k + 1);
            if (k < 3)
               {
               simincrease *= 2;
               }
            simcountlist[tnum][k] += simincrease;
            countlist[k] += simcountlist[tnum][k];
            simcountlist[tnum][0] += simincrease;
            }
         }
      countlist[0] += simcountlist[tnum][0];
      }
      
   /* return data values */
   return countlist;
      
   }
   

/* -------------------------------------------------------------- */
/* draw the data bar on the results screen */

drawdata(int thistopic)
   {
   int grandtotal = 0;
   int x;
   int i, j;
   int lnum = MAX_RESPONSES;
   char *utime = NULL;
   int *cdata;
   
   /* printf("Drawing data for topic %d\n", thistopic); */
   
   /* find out how many responses for this data set */
   for (j = 1; j <= MAX_RESPONSES; j++)
      {
      if (topiclist[thistopic].response[j] == NULL)
         {
         lnum = j - 1;
         break;
         }
      }	
   
   cdata = count_data(thistopic);
      
   /* calculate the total number of responses */
   grandtotal = cdata[0];  
   

   /* now plot each data bar */
   for (i = 1; i <= lnum; i++)
      {
      x = xorigin + space + width + ((i - 1) * (slice));
      databar(x, cdata[i], grandtotal); 
      }

   /*  get the date  */
   time(&now);
   utime = strdup(ctime(&now));
   linetostr(utime); /* get rid of \n */
   
/* get the status information on the screen */   
/*
 *    
 *    ET(
 *       .rmain.canvasframe.area create text \
 *         55 7 \
 *          -text "Updated: %s(utime)" \
 *          -font "%q(default_font)" \
 *          -tags "update"
 * 
 *       .rmain.canvasframe.area create text \
 *          150 7 \
 *          -text "Total Responses: %d(grandtotal)" \
 *          -font "%q(default_font)" \
 *          -tags "update"
 * 
 *       .rmain.canvasframe.area create text \
 *          250 7 \
 *          -text "Response Life Time: %d(topiclist[thistopic].decay_time)" \
 *          -font "%q(default_font)" \
 *          -tags "update"
 *     );
 */
 
 /* sparse version of status information */
   ET(
      .rmain.canvasframe.area create text \
         150 7 \
         -text "Total Responses: %d(grandtotal)" \
         -font "%q(default_font)" \
         -tags "update"
      );
 
   free(utime);
   }
   
/* -------------------------------------------------------------- */
/* do the actual data bars in the bar graph */

databar(int xpos, int amt, int total)
   {
   float percent;
   int ypos = 175;
   int topy, texty, leftx, rightx;
   
   if (total == 0)
      {
      /* can't divide by zero so bail out */
      return(0);
      }
   
   /* calculate the percent of total for this bar */
   percent = ((float)amt / (float)total)*100;
   
   /* translate into coordinates */
   topy = ypos - (percent * 1.5);
   texty = topy - 5;
   
   leftx = xpos - width;
   rightx = xpos + width;

   
   /* draw the bars */
   ET(
      .rmain.canvasframe.area create rectangle \
      %d(leftx) %d(ypos) \
      %d(rightx) %d(topy) \
      -fill red \
      -tags update
      
      .rmain.canvasframe.area create text \
      %d(xpos) %d(texty) \
      -text %d(amt) \
      -font "%q(default_font)" \
      -tags update
   );
   }
   
   
/* ----------------------------------------------------------- */
/* monitor data traffic flowing in and out */
traffic(char *message)
   {
   char *yview = NULL;
   float first, last;
   
   /* check that the monitor window is open */
   if (ET_INT(winfo exists .mmain) == 0)
      {
      return(0);
      }
   ET(
      .mmain.top.textwin insert end "%q(message)\n"
      );
      
      
   /* autoscroll if near the bottom of the window */
   
   yview = strdup(ET_STR(.mmain.top.textwin yview));
   
   sscanf(yview, "%f%f" , &first, &last);
   
   if (last > 0.90)
      {
      ET(.mmain.top.textwin yview moveto 1);
      }
   free(yview);
   }


/* ------------------------------------------------------------------ */
/* create the member list GUI */
ET_PROC( viewmembers )
   {
   
   Tcl_DeleteTimerHandler(member_timer_token);
      
    /* if a members window exists, get rid of it */
    
   if (ET_INT(winfo exists .members) != 0)
      {
      ET(destroy .members);
      }
      
   /* create the GUI */
   ET_INCLUDE( members.tcl );
   
   update_members();
   
   return ET_OK;
   }


/* ------------------------------------------------------------------ */

update_members()
   {
   int status;
   int interval;
   char tmps[TMPSTRLEN];
   int i;
   int keep;
   
   
   /* check that we still have a results window */
   status = ET_INT(winfo exists .members);
   if (! status)
      {
      Tcl_DeleteTimerHandler(member_timer_token);
      return(0);
      }
      
   /* get rid of the old data */
   ET(.members.top.list delete 1.0 end);
   
   time(&now);
   
   keep = ET_INT(set keep);
   ET(.members.top.list tag configure gray -foreground gray50);
   
   for(i = 1; i <= nummembers; i++)
      {
      /* don't show stale members */
      tdiff = now - memberlist[i].time;
      
      /* should never be more than 2 * MIN, but add some slack
         for missed packets */
         
      if (tdiff > MIN_INTERVAL_MEMBER * 4)
         {
         if (keep)
            {
            /* set text color grey */
            sprintf(tmps, "%s@%s\n", memberlist[i].user, memberlist[i].host);
            ET(.members.top.list insert end "%s(tmps)" gray);
            continue;
            }
         else
            {
            continue;
            }
         }
      sprintf(tmps, "%s@%s\n", memberlist[i].user, memberlist[i].host);
      ET(.members.top.list insert end "%s(tmps)");
      }
   
   /* and update the Tk display */
   ET(update idletasks);
   
   /* handler uses msecs so ... */
   interval = UPDATE_TIME * 1000; 
   
   /* infinite loop to this same function */
   member_timer_token =  Tcl_CreateTimerHandler(interval, 
     (Tk_TimerProc*)update_members, (ClientData)NULL);
   }

   
