#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "notify.h"
#include "multimer.h"
#include <sys/time.h>
#include <string.h>
#include "udp.h"
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
 
#define PMM_MULTICAST "225.2.0.1"
#define PMM_PORT  5010

  
char message[1000]; /* buffer space to contain the description message*/
char id[1000];       /* buffer space for the session id */
FILE *ptr=NULL;

void usage()
{
  fprintf(stderr, 
         "Usage: mint [-f coding] [-h] [-C SessionName] host/port/ttl\n ");
  if(ptr!=NULL) pclose(ptr);
  exit (-1);
}

/* adjust the session name to suite the TCL parsing mechanisnsm -i.e., no spaces
 * points or any non numbar of letter chars */
char* parse_name(char* session )
{
  int i;
  
  for(i=0;i<strlen(session);i++) {
    if(!isalnum(session[i]))
      session[i]='_';
  }
  return session;
}
/*
 * parse the online parameters and invoke the appropriate actions
 */

char* parse(int argc, char *argv[],char* ptr)
{
  extern char *optarg;
  extern int optind;
  char tmp[100];
  char c;
  optind=1;
  while ((c = getopt(argc, argv, "sb:hf:R:d:C:p:P:m:a:I:q:t:v:r")) != EOF) {
    switch(c) {
    case 'h':
      usage();
      break;
    case 'I':
      sprintf(tmp,"chan %s ",optarg);
      break;
    case 'P':
      sprintf(tmp,"transport %s ",optarg);
      break;
    case 'm':
      sprintf(tmp,"media %s ",optarg);
      break;
    case 'f':
      sprintf(tmp,"coding %s ",optarg);
      break;
    case 'r':
      sprintf(tmp,"transport RTP ");
      break;
    case 't':
      sprintf(tmp,"ttl %s ",optarg);
      break;
    case 'C':
      {
	char* name=parse_name(optarg);
	sprintf(tmp,"session %s ",name);
	sprintf(id,"%s",name);
	break;
      }
    }
    memcpy(ptr,tmp,strlen(tmp));
    ptr+=strlen(tmp);
  }
  return ptr;
}


/* 
 * if isc still does not answer after MAX_TIMEOUT end mint with 
 * an error message
 */
int mint_failed (Notify_client client)
{
  fprintf(stderr," failed to start isc");
  if(ptr!=NULL) pclose(ptr);
  exit(1);
}

/*
 *  invoked if isc did not respond after a timeout
 *
 *  in:         client: call_id
 */

int mint_timeout (Notify_client client)
{
  char command[1000];
  struct timeval        time;
  
  sprintf(command,"isc -C %s >& 0",id);
  
  if ((ptr = popen(command, "rw")) != NULL) {
    time.tv_sec = 20;  /* allow for 20 seconds to pass before panicking */
    time.tv_usec = 0;
    timer_set( &time,mint_failed, (Notify_client)client,1);
  } 
  else
    fprintf(stderr," could not start mint \n");

  return 0;
} /* mint_timeout */
 


/* Analyze the received messages and see if they are ment for us */
void pmm_analyze(Notify_client client, int sock)
{
  char *buf;
  char* id;
  char total_id[100],start_id[100];
  int i;

  for(i=0;i<sizeof(total_id);i++){
    total_id[i]=0;
    start_id[i]=0;
  }

  id=(char *)client;
  sprintf(total_id,"*/mint/ %s",id);
  buf=UDP_receive(sock);
  /* received the right message after isc received the message with 
     the parameters*/
  if(strstr(buf,total_id)!=NULL) {
    notify_set_input_func(0,NOTIFY_FUNC_INPUT_NULL,sock);
    notify_stop();
   if(ptr!=NULL) pclose(ptr);
   exit (1);
  }
  sprintf(start_id,"+ %s",id);
  if(strstr(buf,start_id)!=NULL) {/* received the initialization message 
					   from isc */
    /* send the parameters needed for the session to isc */
    UDP_write(sock,message,sizeof(message)," ",1); 
    
    /* stop the timer */
    notify_set_input_func(0,NOTIFY_FUNC_INPUT_NULL,sock);
        /* end mint */
    notify_stop();
    if(ptr!=NULL) pclose(ptr);
  exit (1);
  }
}

/*
 * Main function
 */
int main(int argc, char *argv[])
{
  extern char *optarg;
  int port,len;
  int i, sock,ttl = 16;
  char *s=NULL;
  struct timeval        time;
  char *ptr=NULL;
  char tmp[100];
  
  for(i=0;i<sizeof(message);i++){
    message[i]=0;
    id[0]=0;
  }

  ptr=message;
  sprintf(tmp,"*/mint/ ");
  memcpy(ptr,tmp,strlen(tmp));
  ptr+=strlen(tmp);
  /* parses host description */
  ptr=parse(argc,argv,ptr);

  if(argv[optind]==NULL){
    usage();
  }
  s = strchr(argv[optind], '/');
  if(!s)
    {
      usage();
    }
  else 
    {
      *s = '\0';
      port =(atoi(s+1));
      sprintf(tmp,"port %i ",port);
      memcpy(ptr,tmp,strlen(tmp));
      ptr+=strlen(tmp);
      
      s = strchr(s+1, '/');
      if (s) {
	ttl = atoi(s+1);
      }
      sprintf(ptr,"address %s ",argv[optind]);
    }
  sock=UDP_connect("225.2.0.1",PMM_PORT,0);
  notify_set_input_func(id, (Notify_func_input)pmm_analyze, sock);

  /* start timer for receiving a "Hello" answer form isc */
  time.tv_sec = 1;
  time.tv_usec = 0;
  timer_set( &time,(Notify_func)mint_timeout, (Notify_client)sock,1);

  len=UDP_write(sock,message,sizeof(message)," ",1);
  notify_start();
}

