/*
 * callbacks.c -- X11 callbacks for ghostview.
 * Copyright (C) 1992  Timothy O. Theisen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   Author: Tim Theisen           Systems Programmer
 * Internet: tim@cs.wisc.edu       Department of Computer Sciences
 *     UUCP: uwvax!tim             University of Wisconsin-Madison
 *    Phone: (608)262-0438         1210 West Dayton Street
 *      FAX: (608)262-9777         Madison, WI   53706
 */

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#ifndef BUFSIZ
#define BUFSIZ 1024
#endif

#ifdef VMS
#define getenv _getenv
#endif


extern char *getenv();

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>

#include "Ghostview.h"
#include "gv.h"
#include "ps.h"

static char sfilename[128];
static char shostname[128];
static struct statuspacket_t current_state_packet={-1, -1, -1, PORTRAIT, 0xFF, -1, -1, False,-1};
static String sessionname=NULL;
static appopen=0;						   
/* Start application folding up by Destroying the top level widget. */
/* The application exits when the last interpreter is killed during */
/* a destroy callback from ghostview widgets. */
void
quit_ghostview(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
  if(appopen){
  /* Leave pmm multicast group and send the closed message */
  sprintf(pmmmessage,"%s/netghost/%s:%d closed\n",sessionname,
	  nwd.address,nwd.port);
  send_pmm(pmmmessage);
  }
  /* Destroy data socket */
  close(nwd.pmmsd);
  destroy_socket(&nwd);
  XtDestroyWidget(toplevel);
}

/* Popup a window. */
void
popup(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    positionpopup((Widget)client_data);
    XtPopup((Widget)client_data, XtGrabNone);
    XRaiseWindow(XtDisplay((Widget)client_data), XtWindow((Widget)client_data));
}
void
net_dialog(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    
    char portstring[6];
    char ttlstring[6];
    char connection_IDstring[8];
    char send_rec[10];
    String what;
    Widget toggle;
    mode = NET;
    toggle = XtNameToWidget(netdialog,"sender");
    XawToggleSetCurrent(toggle,nwd.mode);
    sprintf(ttlstring,"%d",nwd.ttl);
    sprintf(portstring,"%d",nwd.port);
    sprintf(connection_IDstring,"%d",nwd.connection_ID);
    SetAddressDialogPrompt(netdialog, app_res.address_prompt);
    SetPortDialogPrompt(netdialog, app_res.port_prompt);
    SetTtlDialogPrompt(netdialog, app_res.ttl_prompt);
    SetIdDialogPrompt(netdialog, app_res.connection_ID_prompt);
    SetAddressDialogResponse(netdialog, nwd.address);
    SetPortDialogResponse(netdialog, portstring);
    SetTtlDialogResponse(netdialog, ttlstring);
    SetIdDialogResponse(netdialog, connection_IDstring);
    popup(w, netdialogpopup, call_data);
    
}


/* Popup a dialog box. */
void
popup_dialog(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
#ifdef SELFILE
    Widget button;
    String okay, cancel;
    String name, init_path;
    Arg args[1];
    FILE* fp;
    struct stat sbuf;
    extern FILE *XsraSelFile();

    button = XtNameToWidget(dialog, "cancel");
    if (button) {
	XtSetArg(args[0], XtNlabel, &cancel);
	XtGetValues(button, args, ONE);
    }
    button = XtNameToWidget(dialog, "okay");
    if (button) {
	XtSetArg(args[0], XtNlabel, &okay);
	XtGetValues(button, args, ONE);
    }
#endif

    mode = (int) client_data;
    switch (mode) {
    case PRINT_WHOLE:
    case PRINT_MARKED:
	SetDialogPrompt(dialog, app_res.print_prompt);
	if (app_res.default_printer)
	    SetDialogResponse(dialog, app_res.default_printer);
	else
	    ClearDialogResponse(dialog);
	popup(w, (XtPointer)dialogpopup, call_data);
	break;
    case OPEN:
#ifdef SELFILE
	if (filename && strcmp(filename, "-")) init_path = filename;
	else init_path = NULL;
	fp = XsraSelFile(toplevel, app_res.open_prompt, okay, cancel,
			 app_res.open_fail, init_path, "r", NULL, &name);
	if (fp == NULL) break;
	if (oldfilename) XtFree(oldfilename);
	oldfilename = filename;
	filename = name;
	if (psfile) fclose(psfile);
	psfile = fp;
	stat(filename, &sbuf);
	mtime = sbuf.st_mtime;
	new_file(0);
	show_page(0);
#else
	SetDialogPrompt(dialog, app_res.open_prompt);
	if (filename && strcmp(filename, "-"))
	    SetDialogResponse(dialog, filename);
	else
	    ClearDialogResponse(dialog);
	popup(w, dialogpopup, call_data);
#endif
	break;
    case SAVE:
#ifdef SELFILE
	fp = XsraSelFile(toplevel, app_res.save_prompt, okay, cancel,
			 app_res.save_fail, "", "w", NULL, &name);
	if (fp == NULL) break;
	pscopydoc(fp);
	fclose(fp);
	XtFree(name);
#else
	SetDialogPrompt(dialog, app_res.save_prompt);
	ClearDialogResponse(dialog);
	popup(w, dialogpopup, call_data);
#endif
	break;
    }
}

/* Explicitly reopen the file. */
void
reopen_file(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    struct stat sbuf;
    int number = current_page;

    fclose(psfile);
    psfile = fopen(filename, "r");
    mtime = sbuf.st_mtime;
    if (oldfilename) XtFree(oldfilename);
    oldfilename = XtNewString(filename);
    new_file(number);
    show_page(number);

    /* hier fehlt noch das reopen
    type_id = NET_REOPEN;
    if (nwd.mode == SENDER) 
    */
}

/* Get the selection, if no selection, get the insertion point. */
/* If the new_page is different from the current page show it.  */
/* If not at the first page, show the previous page. */
void
prev_page(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    XawTextPosition pos, end;
    int new_page;

    XawTextGetSelectionPos(toc, &pos, &end);
    if (pos == end) {		/* Nothing selected */
	pos = XawTextGetInsertionPoint(toc);
    }
    if ((new_page = pos/toc_entry_length) == current_page) {
	new_page = current_page - 1;
    }
    if (new_page < 0) return;
    show_page(new_page);


    if (nwd.mode == SENDER) {
      current_state_packet.page = new_page;
      send_status();
    }
     
}

/* Get the selection, if no selection, get the insertion point. */
/* Show this page.  */
void
this_page(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    if (toc_text) {
	XawTextPosition pos, end;
	int new_page;

	XawTextGetSelectionPos(toc, &pos, &end);
	if (pos == end) {		/* Nothing selected */
	    pos = XawTextGetInsertionPoint(toc);
	}
	new_page = pos/toc_entry_length;
	show_page(new_page);
        if (nwd.mode == SENDER) {
         current_state_packet.page = new_page;
         send_status();
        }

    } else {
	GhostviewDisableInterpreter(page);
	show_page(0);
	if (nwd.mode == SENDER) {
	  current_state_packet.page = 0;
	  send_status();
	}
	
    }
    
    
}

/* Get the selection, if no selection, get the insertion point. */
/* If the new_page is different from the current page show it.  */
/* If not at the last page, show the next page. */
void
next_page(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    XawTextPosition pos, end;
    int new_page = 0;

    if (toc_text) {
	XawTextGetSelectionPos(toc, &pos, &end);
	if (pos == end) {		/* Nothing selected */
	    pos = XawTextGetInsertionPoint(toc);
	}
	if ((new_page = pos/toc_entry_length) == current_page) {
	    new_page = current_page + 1;
	}
	if (new_page >= doc->numpages) return;
    }
    show_page(new_page);
    if (nwd.mode == SENDER) {
      current_state_packet.page = new_page;
      send_status();
    }
    
    
}

/* Center the viewport over the page */
void
center_page(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    Arg args[2];
    Widget scroll;
    float top, shown;

    scroll = XtNameToWidget(pageview, "vertical");
    if (scroll) {
	XtSetArg(args[0], XtNshown, &shown);
	XtSetArg(args[1], XtNtopOfThumb, &top);
	XtGetValues(scroll, args, TWO);

	top = (1.0 - shown) / 2.0;
	XtCallCallbacks(scroll, XtNjumpProc, &top);
    }

    scroll = XtNameToWidget(pageview, "horizontal");
    if (scroll) {
	XtSetArg(args[0], XtNshown, &shown);
	XtSetArg(args[1], XtNtopOfThumb, &top);
	XtGetValues(scroll, args, TWO);

	top = (1.0 - shown) / 2.0;
	XtCallCallbacks(scroll, XtNjumpProc, &top);
    }
}

/* Get the selection, if no selection, get the insertion point. */
/* Mark all pages in range, and cause toc to update. */
void
mark_page(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    XawTextPosition begin, end;
    int i;

    XawTextGetSelectionPos(toc, &begin, &end);
    if (begin == end) {		/* Nothing selected */
	begin = end = XawTextGetInsertionPoint(toc);
    } else {
	end--;			/* Sometimes end spills onto next line */
    }
    for (i = begin/toc_entry_length; i <= end/toc_entry_length; i++) {
	toc_text[i*toc_entry_length] = '*';
	XawTextInvalidate(toc, i*toc_entry_length, i*toc_entry_length+1);
    }
}

/* Get the selection, if no selection, get the insertion point. */
/* Unmark all pages in range, and cause toc to update. */
void
unmark_page(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    XawTextPosition begin, end;
    int i;

    XawTextGetSelectionPos(toc, &begin, &end);
    if (begin == end) {		/* Nothing selected */
	begin = end = XawTextGetInsertionPoint(toc);
    } else {
	end--;			/* Sometimes end spills onto next line */
    }
    for (i = begin/toc_entry_length; i <= end/toc_entry_length; i++) {
	toc_text[i*toc_entry_length] = ' ';
	XawTextInvalidate(toc, i*toc_entry_length, i*toc_entry_length+1);
    }
}

/* Set new magstep.  Reshow the current page if magstep changed. */
void
set_magstep(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    app_res.magstep = (int)client_data;
    if (set_new_magstep()) {
	layout_ghostview();
	show_page(current_page);
    }


}

/* Set new orientation.  Reshow the current page if orientation changed. */
void
set_orientation(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    app_res.force_orientation = force_setting;
    app_res.orientation = (XtPageOrientation) client_data;
    if (set_new_orientation(current_page)) {
	layout_ghostview();
	show_page(current_page);
    }

    if (nwd.mode == SENDER) {
      current_state_packet.orientation = app_res.orientation;
      send_status();
    }

}

/* Swap the landscape labels and change the flag. */
void
swap_landscape(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    Arg args[1];
    String s1, s2;

    app_res.swap_landscape = !app_res.swap_landscape;

    XtSetArg(args[0], XtNlabel, &s1);
    XtGetValues(landscapebutton, args, ONE);
    s1 = XtNewString(s1);
    XtSetArg(args[0], XtNlabel, &s2);
    XtGetValues(seascapebutton, args, ONE);
    s2 = XtNewString(s2);
    XtSetArg(args[0], XtNlabel, s2);
    XtSetValues(landscapebutton, args, ONE);
    XtSetArg(args[0], XtNlabel, s1);
    XtSetValues(seascapebutton, args, ONE);
    XtFree(s1);
    XtFree(s2);

    if (set_new_orientation(current_page)) {
	layout_ghostview();
	show_page(current_page);
    }
    if (nwd.mode == SENDER) {
      current_state_packet.swap_landscape = app_res.swap_landscape;
      send_status();
    }
    

}

/* Set new page media.  If new page media is different, update app_resources */
/* and redisplay page. */
void
set_pagemedia(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    int new_pagemedia = (int) client_data;
    if (new_pagemedia >= base_papersize) {
	default_pagemedia = new_pagemedia;
	app_res.force_pagemedia = force_setting;
    } else {
	document_media = new_pagemedia;
	force_document_media = force_setting;
    }
    if (set_new_pagemedia(current_page)) {
	layout_ghostview();
	show_page(current_page);
    }
}

/* track mouse pointer and popup zoom window */
void
track_and_zoom(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    Arg args[20];
    Cardinal num_args;
    Dimension width, height;
    Widget zoom;
    Widget zoomform;
    Widget zoompage;
    Widget zoomdismiss;
    FILE *zoomfile;
    struct stat sbuf;
    GhostviewReturnStruct *p = (GhostviewReturnStruct *)call_data;
    int llx;
    int lly;
    int urx;
    int ury;
    int bottom_margin;
    int left_margin;
    int right_margin;
    int top_margin;
    int i;

    /* locator events have zero width and height */
    if ((p->width == 0) || (p->height == 0)) {
	char buf[32];
	if (!app_res.show_locator) return;
	sprintf(buf, "(%d, %d)", p->psx, p->psy);
	XtSetArg(args[0], XtNlabel, buf);
	XtSetValues(locator, args, ONE);
	return;
    }

    /* If no file, nothing to zoom. */
    if (!psfile) return;

    /* If the file changed, cannot zoom */
    stat(filename, &sbuf);
    if (mtime != sbuf.st_mtime) return;
    zoom = XtCreatePopupShell("zoom", topLevelShellWidgetClass,
			      toplevel, NULL, ZERO);

    zoomform = XtCreateManagedWidget("form", formWidgetClass,
				     zoom, NULL, ZERO);

    llx = p->psx - p->width/2;
    lly = p->psy - p->height/2;
    urx = p->psx + p->width/2;
    ury = p->psy + p->height/2;

    /* Make sure zoom window doesn't go off the edge of the page */
    if (llx < current_llx) {
	llx = current_llx;
	urx = llx + p->width;
    }
    if (lly < current_lly) {
	lly = current_lly;
	ury = lly + p->height;
    }
    if (urx > current_urx) {
	urx = current_urx;
	llx = urx - p->width;
    }
    if (ury > current_ury) {
	ury = current_ury;
	lly = ury - p->height;
    }
    if (llx < current_llx) {
	llx = current_llx;
    }
    if (lly < current_lly) {
	lly = current_lly;
    }
    bottom_margin = lly - current_lly;
    left_margin = llx - current_llx;
    right_margin = current_urx - urx;
    top_margin = current_ury - ury;

							num_args = 0;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainBottom);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainRight);	num_args++;
    XtSetArg(args[num_args], XtNorientation, current_orientation);
							num_args++;
    XtSetArg(args[num_args], XtNllx, llx);      	num_args++;
    XtSetArg(args[num_args], XtNlly, lly);      	num_args++;
    XtSetArg(args[num_args], XtNurx, urx);      	num_args++;
    XtSetArg(args[num_args], XtNury, ury);      	num_args++;
    XtSetArg(args[num_args], XtNbottomMargin, bottom_margin);
							num_args++;
    XtSetArg(args[num_args], XtNleftMargin, left_margin);
							num_args++;
    XtSetArg(args[num_args], XtNrightMargin, right_margin);
							num_args++;
    XtSetArg(args[num_args], XtNtopMargin, top_margin); num_args++;
    XtSetArg(args[num_args], XtNbottomMargin, bottom_margin);
							num_args++;
    XtSetFloatArg(args[num_args], XtNxdpi, p->xdpi);	num_args++;
    XtSetFloatArg(args[num_args], XtNydpi, p->ydpi);	num_args++;
    if (!toc_text) {
        XtSetArg(args[num_args], XtNfilename, filename);	num_args++;
    }
    zoompage = XtCreateManagedWidget("page", ghostviewWidgetClass,
				     zoomform, args, num_args);
    num_ghosts++;
    XtAddCallback(zoompage, XtNcallback, track_and_zoom, (XtPointer)0);
    XtAddCallback(zoompage, XtNmessageCallback, message, (XtPointer)zoompage);
    XtAddCallback(zoompage, XtNdestroyCallback, destroy_ghost,
		  (XtPointer)zoompage);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, zoompage);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainBottom);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainBottom);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainRight);	num_args++;
    zoomdismiss = XtCreateManagedWidget("dismiss", commandWidgetClass,
				       zoomform, args, num_args);
    XtAddCallback(zoomdismiss, XtNcallback, destroy, (XtPointer)zoom);

    XtSetArg(args[0], XtNwidth, &width);
    XtGetValues(zoompage, args, ONE);
    XtSetArg(args[0], XtNwidth, width);
    XtSetValues(zoomdismiss, args, ONE);

    XtRealizeWidget(zoom);
    positionpopup(zoom);

							num_args = 0;
    XtSetArg(args[num_args], XtNwidth, &width);		num_args++;
    XtSetArg(args[num_args], XtNheight, &height);	num_args++;
    XtGetValues(zoom, args, num_args);

							      	num_args = 0;
    XtSetArg(args[num_args], XtNminWidth, width);		num_args++;
    XtSetArg(args[num_args], XtNminHeight, height);		num_args++;
    XtSetArg(args[num_args], XtNmaxWidth, width);		num_args++;
    XtSetArg(args[num_args], XtNmaxHeight, height);		num_args++;
    XtSetValues(zoom, args, num_args);
    XSetWMProtocols(XtDisplay(zoom), XtWindow(zoom), &wm_delete_window, 1);
    XtPopup(zoom, XtGrabNone);

    if (toc_text) {
	zoomfile = fopen(filename, "r");
	if (zoomfile == NULL) return;
	GhostviewSendPS(zoompage, zoomfile, doc->beginprolog,
			doc->lenprolog, False);
	GhostviewSendPS(zoompage, zoomfile, doc->beginsetup,
			doc->lensetup, False);
	if (doc->pageorder == DESCEND)
	    i = (doc->numpages - 1) - current_page;
	else
	    i = current_page;
	GhostviewSendPS(zoompage, zoomfile, doc->pages[i].begin,
			doc->pages[i].len, True);
    }
}

/* Process messages from ghostscript */
/* Refresh occurs when window was resized unexpectedly */
void
message(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    int i;
    char *error;

    if (!strcmp((char *) call_data, "Failed")) {
	if ((Widget)client_data == page) {
	    error = "Error: PostScript interpreter failed in main window.\n\n";
	} else {
	    error = "Error: PostScript interpreter failed in zoom window.\n\n";
	}
	output(w, NULL, error);
    } else if (!strcmp((char *) call_data, "BadAlloc")) {
	if ((Widget)client_data == page) {
	    error = 
	       "Warning: Could not allocate backing pixmap in main window.\n\n";
	} else {
	    error = 
	       "Warning: Could not allocate backing pixmap in zoom window.\n\n";
	}
	output(w, NULL, error);
    } else if (!strcmp((char *) call_data, "Refresh")) {
	if (toc_text) {
	    GhostviewSendPS(w, psfile, doc->beginprolog,
			    doc->lenprolog, False);
	    GhostviewSendPS(w, psfile, doc->beginsetup,
			    doc->lensetup, False);
	    if (doc->pageorder == DESCEND)
		i = (doc->numpages - 1) - current_page;
	    else
		i = current_page;
	    GhostviewSendPS(w, psfile, doc->pages[i].begin,
			    doc->pages[i].len, False);
	}
    }
}

/* Take output from ghostscript and display it in the infotext popup window */
void
output(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    Arg args[2];
    XawTextBlock message_block;

    message_block.firstPos = 0;
    message_block.length = strlen(call_data);
    message_block.ptr = call_data;
    message_block.format = FMT8BIT;

    XawTextDisableRedisplay(infotext);

    XtSetArg(args[0], XtNeditType, XawtextAppend);
    XtSetValues(infotext, args, ONE);
    XawTextReplace(infotext, info_length, info_length, &message_block);
    info_length = info_length + message_block.length;

    XtSetArg(args[0], XtNeditType, XawtextRead);
    XtSetArg(args[1], XtNinsertPosition, info_length);
    XtSetValues(infotext, args, TWO);
    XawTextEnableRedisplay(infotext);
    if (!info_up) XtPopup(infopopup, XtGrabNone);
    info_up = True;
}

/* Dismiss popup dialog */
void
okay(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    String name, error, address, port, ttl, connection_ID;
    Widget dialog,toggle;
    Arg args[2];

    dialog = (Widget)client_data;
    switch (mode) {
    case PRINT_WHOLE:
    case PRINT_MARKED:
	name = GetDialogResponse(dialog);
	if (error = print_file(name, (mode == PRINT_WHOLE))) {
	    char *buf = XtMalloc(strlen(error) +
				 strlen(app_res.print_prompt) + 2);
	    sprintf(buf, "%s\n%s", error, app_res.print_prompt);
	    SetDialogPrompt(dialog, buf);
	    XtFree(error);
	    XtFree(buf);
	} else {
	    XtPopdown(XtParent(dialog));
	}
	XtFree(name);
	break;
    case OPEN:
	name = GetDialogResponse(dialog);
	if (error = open_file(name)) {
	    char *buf = XtMalloc(strlen(error) +
				 strlen(app_res.open_prompt) + 2);
	    sprintf(buf, "%s\n%s", error, app_res.open_prompt);
	    SetDialogPrompt(dialog, buf);
	    XtFree(error);
	    XtFree(buf);
	} else {
	    XtPopdown(XtParent(dialog));
	}
	XtFree(name);
	break;
    case SAVE:
	name = GetDialogResponse(dialog);
	if (error = save_file(name)) {
	    char *buf = XtMalloc(strlen(error) +
				 strlen(app_res.save_prompt) + 2);
	    sprintf(buf, "%s\n%s", error, app_res.save_prompt);
	    SetDialogPrompt(dialog, buf);
	    XtFree(error);
	    XtFree(buf);
	} else {
	    XtPopdown(XtParent(dialog));
	}
	XtFree(name);
	break;
    case NET:
	address = GetAddressDialogResponse(dialog);
	port = GetPortDialogResponse(dialog);
	ttl =  GetTtlDialogResponse(dialog);
        connection_ID = GetIdDialogResponse(dialog);
	toggle = XtNameToWidget(dialog,"sender");
	nwd.mode = (int) XawToggleGetCurrent(toggle);
	XtFree(nwd.address);
	nwd.address=XtNewString(address);
        nwd.connection_ID = atoi(connection_ID);
	/*	if(nwd.connection_ID < 0){
	    char *buf = XtMalloc(strlen("Invalid connection ID") +
				 strlen(app_res.connection_ID_prompt) + 2);
	    sprintf(buf, "Invalid connection ID\n%s",  app_res.connection_ID_prompt);
	    SetIdDialogPrompt(dialog, buf);
	    XtFree(buf);
	}
	*/
	if((nwd.port=atoi(port))==0){
	  char *buf = XtMalloc(strlen("Invalid port Number") +
			       strlen(app_res.port_prompt) + 2);
	  sprintf(buf, "Invalid port Number\n%s",  app_res.port_prompt);
	  SetPortDialogPrompt(dialog, buf);
	  XtFree(buf);
	}else{
	  if ( !(nwd.ttl=atoi(ttl)) && !(*ttl=='0')){
	    char *buf = XtMalloc(strlen("Invalid TTL") +
				 strlen(app_res.ttl_prompt) + 2);
	    sprintf(buf, "Invalid TTL\n%s",  app_res.ttl_prompt);
	    SetTtlDialogPrompt(dialog, buf);
	    XtFree(buf);
	  }
	  else {
	    destroy_socket(&nwd);
	    switch(nwd.mode){ 
	    case SENDER:
	      if (error = setup_send_socket(&nwd)){
		char *buf = XtMalloc(strlen(error) +
				     strlen(app_res.address_prompt) + 2);
		sprintf(buf, "%s\n%s", error, app_res.address_prompt);
		SetAddressDialogPrompt(dialog, buf);
		XtFree(error);
		XtFree(buf);
		destroy_socket(&nwd);
	      } 
	      else {
		XtPopdown(XtParent(dialog));
		XtSetArg(args[0],XtNmode,True);
		XtSetValues(page,args,ONE);
		current_state_packet.visible=False;
		register_timer();
	      }
	      break;
	    case RECEIVER:
	      if (error = setup_receive_socket(&nwd)){
		char *buf = XtMalloc(strlen(error) +
				     strlen(app_res.address_prompt) + 2);
		sprintf(buf, "%s\n%s", error, app_res.address_prompt);
		SetAddressDialogPrompt(dialog, buf);
		XtFree(error);
		XtFree(buf);
		destroy_socket(&nwd);
	      }
	      else {
		XtPopdown(XtParent(dialog));
		XtSetArg(args[0],XtNmode,False);
		XtSetValues(page,args,ONE);
		current_state_packet.visible=False;
		net_id = XtAppAddInput(app_con, nwd.sd,  
				       (XtPointer)XtInputReadMask, 
				       receive_status, NULL);
	      }
	      break;
	    default:
	      XtPopdown(XtParent(dialog));
	      XtSetArg(args[0],XtNmode,True);
	      XtSetValues(page,args,ONE);
	      current_state_packet.visible=False;
	      break;
	    }
	  }
	}
	XtFree(address);
	XtFree(port);
	XtFree(ttl);
	/*   XtFree(connection_ID);*/
	break;
    }
}

/* Dismiss popup window */
void
dismiss(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    XtPopdown((Widget)client_data);
    if ((Widget)client_data == infopopup) info_up = False;
}

/* Destroy popup window */
void
destroy(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    XtDestroyWidget((Widget)client_data);
}

/* destroy callback for Ghostview widgets. */
/* The disable interpreter call ensures that ghostscript is killed. */
/* One the count goes to 0, we are sure that all forked processes have */
/* been killed and that we can safely exit. */
void
destroy_ghost(w, client_data, call_data)
    Widget w;
    XtPointer client_data, call_data;
{
    GhostviewDisableInterpreter((Widget) client_data);
    num_ghosts--;
    if (num_ghosts) return;
    if (dying) old_Xerror(XtDisplay(w), &bomb);
    XtDestroyApplicationContext(app_con);
    exit(0);
}

/* 
 * This function is called when the Pointer in the GhostviewWidget is changed
 */

void pointer_change(w, client_data, call_data)
     Widget w;
     XtPointer client_data, call_data;
{
  GhostviewReturnStruct *p = (GhostviewReturnStruct *)call_data;
  current_state_packet.x=p->psx;
  current_state_packet.y=p->psy;
  current_state_packet.visible=p->visible;
  send_status();
}

/*
 * This function is called periodically to inform the session controller
 * that the application ist still alive
 */

void Send_Active(client_data, id)
XtPointer client_data;
XtIntervalId *id;
{
 sprintf(pmmmessage,"%s/netghost/%s:%d active\n",sessionname,
	 nwd.address,nwd.port);
 send_pmm(pmmmessage);
 (void) XtAppAddTimeOut((XtAppContext)client_data,nwd.update_interval,
			Send_Active,client_data);
}

/* Timer to retransmit the status packet */
void update_timer(client_data, id)
XtPointer client_data;
XtIntervalId *id;
{
  if(nwd.mode==SENDER){
    send_status();
    nwd.interval_id=XtAppAddTimeOut((XtAppContext)client_data,nwd.interval,
				    update_timer,client_data);
  }
  else{
    nwd.interval_id=0;
  }
}


/* 
 * This function is called if a status packet arrives.
 * Depending on the input the appropriate Callback function is called
 */

void 
receive_status(client_data, fid, id)
XtPointer client_data;
int *fid;
XtInputId *id;
{

  int connection_ID;
  int nbytes;
  int addrlength=sizeof(struct sockaddr_in);
  struct statuspacket_t actual_packet;
  short int orientation;
  long int pagenr;  
  unsigned char sw_landscape;
  int x,y;
  Boolean visible;

  if ( (nbytes = recvfrom(*fid, &actual_packet, sizeof(struct statuspacket_t), 0, (struct sockaddr*)&(nwd.addr), &addrlength)) == -1) {
      perror("recvfrom");
  }

  connection_ID = ntohl(actual_packet.connection_ID);

  if (connection_ID == nwd.connection_ID) {
    
    orientation=ntohs(actual_packet.orientation);
    pagenr=ntohl(actual_packet.page);
    sw_landscape=actual_packet.swap_landscape;
    x=ntohl(actual_packet.x);
    y=ntohl(actual_packet.y);
    visible=actual_packet.visible;

    if (current_state_packet.orientation != orientation)  {
      set_orientation(NULL, (XtPointer)orientation, NULL);
      current_state_packet.orientation = orientation;
    }
    
    if (current_state_packet.page != pagenr)  {
      show_page(pagenr,NULL);
      current_state_packet.page =pagenr;
    }
    
    if (current_state_packet.swap_landscape != sw_landscape)  {
      swap_landscape(NULL, (XtPointer)&sw_landscape, NULL);
      current_state_packet.swap_landscape = sw_landscape;
    }
    if(visible)
      {
	if(current_state_packet.visible)
	  {
	    if(x==current_state_packet.x && y==current_state_packet.y)
	      {
		/* Do nothing it is a repetition of the previos state */
	      }else{
		/* 
		 * probaly we lost a packet undraw the old pointer and
		 * draw a new one at given location. Update the current
		 * state
		 */
		GhostviewUnDrawPointer(page);
		GhostviewDrawPointer(x,y,page);
		current_state_packet.x=x;
		current_state_packet.y=y;
	      }
	  }
	else
	  {
	    /* We got a packet with a valid draw call */
	    GhostviewDrawPointer(x,y,page);
	    current_state_packet.x=x;
	    current_state_packet.y=y;
	    current_state_packet.visible=visible;
	  }
      }
    else /* undraw call */
      {
	if(current_state_packet.visible)
	  {
	    /* undraw the pointer position doesn't matter */
	    GhostviewUnDrawPointer(page);
	  }
	/* update the current state */
	    current_state_packet.x=x;
	    current_state_packet.y=y;
	    current_state_packet.visible=visible;

      }

    if (nwd.mode == RECEIVER) {
      memcpy(shostname, actual_packet.host, 128);
      memcpy(sfilename, actual_packet.file, 128);
      sprintf(pmmmessage,"%s/netghost/%s:%d netghost_host %s\n",
	      sessionname,nwd.address,nwd.port,shostname);
      send_pmm(pmmmessage);
      sprintf(pmmmessage,"%s/netghost/%s:%d netghost_path %s\n",
	      sessionname,nwd.address,nwd.port,sfilename);
      send_pmm(pmmmessage);

    }

  }
}


/* 
 * This function is called if some state has changed or the timer interval
 * expired.
 */

void 
send_status()
{
  struct statuspacket_t actual_packet;

  actual_packet.connection_ID=htonl(nwd.connection_ID);
  actual_packet.page=htonl(current_state_packet.page);
  actual_packet.orientation=htons(current_state_packet.orientation);
  actual_packet.swap_landscape=current_state_packet.swap_landscape;
  actual_packet.x=htonl(current_state_packet.x);
  actual_packet.y=htonl(current_state_packet.y);
  actual_packet.visible=current_state_packet.visible;
  memcpy(actual_packet.host, shostname, 128); 
  memcpy(actual_packet.file, sfilename, 128); 

  sendto(nwd.sd, &actual_packet, sizeof(struct statuspacket_t),
	 0,(struct sockaddr*)&(nwd.addr) ,sizeof(struct sockaddr));
}

void receive_pmm(client_data, fid, id)
XtPointer client_data;
int *fid;
XtInputId *id;
{
  static char buf[300];
  static String cname=NULL,s3=NULL;
  static int status=0; /* bit 0 -> filename received, bit 1 -> hostname received */
  char *bufptr,*tok;
  int readbyte;
  Arg args[2];

  bzero(buf,300);
  if(read(nwd.pmmsd,buf,300)<0){
    perror("Read PMM Socket");
  }
  bufptr=buf;
  if(strstr(bufptr,"netghost")){ /* pmm message for ghostview */ 
    /* picture=netghost */
      if(*bufptr=='+'){ /* isc message */
	  bufptr+=2; /* remove first two characters */
	  if(tok=strtok(bufptr,"/")){ /*should be  the sessionname */
	      if(!sessionname)sessionname=XtNewString(tok);
	      if(tok=strtok(NULL,"/")){/* should be the identity */
		if(tok=strtok(NULL,":")){/* inetaddress */
		  if(nwd.address){
		    XtFree(nwd.address);
		  }
		  nwd.address=XtNewString(tok);
		}
		if(tok=strtok(NULL," ")){ /* port */
		  nwd.port=atoi(tok); 
		}
	      }
	  }
	  return;
      }
      
      if(strncmp(bufptr,"session",7)==0){
	sprintf(pmmmessage,"%s/netghost/%s:%d created",
		sessionname,nwd.address,nwd.port);
	send_pmm(pmmmessage);
	return;
      }


      if(strchr(bufptr,'/')&&(tok=strtok(bufptr,"/"))){ /*should be  the sessionname */
	if(!sessionname)sessionname=XtNewString(tok);
	if(tok=strtok(NULL,"/")){/* should be the identity */
	  if(tok=strtok(NULL,":")){/* inetaddress */
	    if(nwd.address){
	      XtFree(nwd.address);
	    }  
	    nwd.address=XtNewString(tok);
	  }
	  if(tok=strtok(NULL," ")){ /* port */
	    nwd.port=atoi(tok);
	  }else{
	    return;
	  }

	  if(tok=strtok(NULL," ")){ /* commands */
	    if(strncmp(tok,"opened",6)&&!strncmp(tok,"open",4)){ 
	      /* open command open socket and anounce it */
	      tok+=5;
	      if(nwd.address){
		XtFree(nwd.address);
	      }
	      nwd.address=XtNewString(tok);
	      appopen=TRUE;
	      if (nwd.mode == RECEIVER) {
		s3=setup_receive_socket(&nwd);
		if(s3!=NULL){
		  fprintf(stderr,"cannot connect to network resuming normal operation \n%s",s3);
		  nwd.mode = NORMAL;
		  XtSetArg(args[0],XtNmode,True);
		  XtSetValues(page,args,ONE);
		  current_state_packet.visible=False;
		  XtFree(s3);
		  destroy_socket(&nwd);
		}else{
		  net_id = XtAppAddInput(app_con,nwd.sd,
					 (XtPointer)XtInputReadMask,
					 receive_status,NULL);  
		}
	      }
    
	      if (nwd.mode == SENDER) {
		s3=setup_send_socket(&nwd);
		if(s3!=NULL){
		  fprintf(stderr,"cannot connect to network resuming normal operation \n%s",s3);
		  nwd.mode = NORMAL;
		  XtFree(s3);
		  destroy_socket(&nwd);
		}else{
		  register_timer();
		}
	      }

	      sprintf(pmmmessage,"%s/netghost/%s:%d hello %s\n",sessionname,
		      nwd.address,nwd.port,cname);
	      send_pmm(pmmmessage);	    
	      sprintf(pmmmessage,"%s/netghost/%s:%d opened %s\n",
		       sessionname,nwd.address,nwd.port,nwd.address);
	      send_pmm(pmmmessage);
	    }
	      
	    if(strncmp(tok,"closed",6)&&!strncmp(tok,"close",5)){ 
	      /*close command exit application */
	      
	      quit_ghostview(NULL,NULL,NULL);
	    } /*endif strcmp(tok,"close") */

	    if(!strncmp(tok,"update_interval",15)){
	      tok=strtok(NULL," ");
	      nwd.update_interval=atoi(tok);
	      XtAppAddTimeOut(app_con,nwd.update_interval,
			      Send_Active,(XtPointer)app_con);
	      return;
	    }

	    if(!strncmp(tok,"ttl",3)){
	      tok=strtok(NULL," ");
	      nwd.ttl=atoi(tok);
	      if(!appopen)return;
	      if(nwd.mode=SENDER&&(nwd.sd!=-1))
		setsockopt(nwd.sd, IPPROTO_IP, IP_MULTICAST_TTL,&nwd.ttl , 
		       sizeof(unsigned char));
	      return;
	    }
	     
	    if(!strncmp(tok,"netghost_id",11)){
	      tok=strtok(NULL," ");
	      nwd.connection_ID=atoi(tok);
	      return;
	    }

	    if(!strncmp(tok,"mode",4)){
	      tok=strtok(NULL," ");
	      if(!strcmp(tok,"receiver")){ /* switch to receiving mode */
		if(!appopen){
		  nwd.mode=RECEIVER;
		  XtSetArg(args[0],XtNmode,False);
		  XtSetValues(page,args,ONE);
		  current_state_packet.visible=False;
		  return;
		}
		if(nwd.sd!=-1){
		  destroy_socket(&nwd);
		  s3=setup_receive_socket(&nwd);
		  if(s3!=NULL){
		    fprintf(stderr,"cannot connect to network resuming normal operation \n%s",s3);
		    nwd.mode = NORMAL;
		    XtSetArg(args[0],XtNmode,True);
		    XtSetValues(page,args,ONE);
		    current_state_packet.visible=False;
		    XtFree(s3);
		    destroy_socket(&nwd);
		  }else{
		    net_id = XtAppAddInput(app_con,nwd.sd,
					   (XtPointer)XtInputReadMask,
					   receive_status,NULL);
		    nwd.mode=RECEIVER;
		    XtSetArg(args[0],XtNmode,False);
		    XtSetValues(page,args,ONE);
		    current_state_packet.visible=False;
		  }
		}
	      }else{ /*switch to sending mode */
		if(!appopen){
		  nwd.mode=SENDER;
		  XtSetArg(args[0],XtNmode,True);
		  XtSetValues(page,args,ONE);
		  current_state_packet.visible=False;
		  return;
		}
		if(nwd.sd!=-1){
		  destroy_socket(&nwd);
		  s3=setup_send_socket(&nwd);
		  if(s3!=NULL){
		    fprintf(stderr,"cannot connect to network resuming normal operation \n%s",s3);
		    nwd.mode = NORMAL;
		    XtSetArg(args[0],XtNmode,True);
		    XtSetValues(page,args,ONE);
		    current_state_packet.visible=False;
		    XtFree(s3);
		    destroy_socket(&nwd);
		  }else{
		    nwd.mode=SENDER;
		    XtSetArg(args[0],XtNmode,True);
		    XtSetValues(page,args,ONE);
		    current_state_packet.visible=False;
		    register_timer();
		  }
		}
	      }
	    }

	    /* If we receive a netghost_file command, the file has been 
	       downloaded and put on the local harddisc -> just open it */
	    if(!strncmp(tok,"netghost_path",12)){
	      if((tok=strtok(NULL," "))){
		/* set bit 0 in status register -> we received a filename from ISC */
		status |= 0x01;
		strcpy(sfilename, tok);
	      }
	    }

	    if(!strncmp(tok,"netghost_file",12)){
	      if((tok=strtok(NULL," "))){
		open_file(tok);
	      }
	    }

	    if(!strncmp(tok,"netghost_host",12)){
	      if((tok=strtok(NULL," "))) {
		/* set bit 1 in status register -> we received a hostname from ISC */		  
		status |= 0x02; 
		strcpy(shostname, tok);
		//fprintf(stderr," hostname -> %s %d\n",shostname,strlen(tok));

	      }
	    }

	    if (!strncmp(tok,"cname",5)&&!cname) {
	      char host[100],user[50],cn[150];
	      struct utsname name;     /* uname value */
	      struct hostent *hep;  
	      struct in_addr in;
	      struct passwd *pw;
	      uname(&name);
	      if ((hep = gethostbyname(name.nodename)) && 
		  hep->h_addrtype == AF_INET){
		in.s_addr = *(u_long *)(*hep->h_addr_list);
		strcpy((char*)host, (char*)inet_ntoa(in));
	      }
	      pw = getpwuid(getuid());
	      if (pw) {
		strcpy(user,pw->pw_name);
	      }
	      else{
		cuserid(user);
	      }
	      sprintf(cn,"%s@%s",user,host);
	      cname=XtNewString(cn);
	      return;
	    } /* endif strncmp(tok,"cname",5) */
	  } /*endif strtok(NULL,"/") commands*/
	}

	/* If we are sender and the user has specified a new file, just post
	   it to all other receivers. */
	/*if ((status & 0x01) && (status & 0x02) && (nwd.mode == SENDER)) {
	  fprintf(stderr,"posting file location to other receivers -> %s %s\n",shostname,sfilename);
	  }*/



	return;
      }
  } 
      /* canonical name of user 
	 message format is "cname canonical_name"
	 */
}
