/*
 * dialogs.c -- Dialog box for ghostview.
 * Copyright (C) 1992  Timothy O. Theisen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   Author: Tim Theisen           Systems Programmer
 * Internet: tim@cs.wisc.edu       Department of Computer Sciences
 *     UUCP: uwvax!tim             University of Wisconsin-Madison
 *    Phone: (608)262-0438         1210 West Dayton Street
 *      FAX: (608)262-9777         Madison, WI   53706
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xos.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include "gv.h"

static String okay_accelerators =
    "#override\n\
     <Key>Return: set() notify() unset()\n";

/* Create a dialog widget */
/* It is just a form widget with
 *   a label prompt
 *   a text response
 *   an oky button
 *   a cancel button */
Widget
CreateNetDialog(parent, name, okay_callback, cancel_callback)
    Widget parent;
    String name;
    XtCallbackProc okay_callback;
    XtCallbackProc cancel_callback;
{
    Widget netform, addressprompt, addressresponse, portprompt, portresponse, 
	ttlprompt, ttlresponse, connection_IDprompt, connection_IDresponse, 
        sender, receiver, okay, cancel;
    Arg args[20];
    Cardinal num_args;

    netform = XtCreateManagedWidget(name, formWidgetClass, parent, NULL, ZERO);

							num_args = 0;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNborderWidth, 0);	num_args++;
    addressprompt = XtCreateManagedWidget("addressprompt", labelWidgetClass,
				   netform, args, num_args);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromHoriz, addressprompt);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNeditType, XawtextEdit);	num_args++;
    XtSetArg(args[num_args], XtNresize, XawtextResizeWidth);	num_args++;
    XtSetArg(args[num_args], XtNstring, "");		num_args++;
    addressresponse = XtCreateManagedWidget("addressresponse", asciiTextWidgetClass,
				     netform, args, num_args);


							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, addressprompt);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNborderWidth, 0);	num_args++;
    portprompt = XtCreateManagedWidget("portprompt", labelWidgetClass,
				   netform, args, num_args);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, addressprompt);	num_args++;
    XtSetArg(args[num_args], XtNfromHoriz, portprompt);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNeditType, XawtextEdit);	num_args++;
    XtSetArg(args[num_args], XtNresize, XawtextResizeWidth);	num_args++;
    XtSetArg(args[num_args], XtNstring, "");		num_args++;
    portresponse = XtCreateManagedWidget("portresponse", asciiTextWidgetClass,
				     netform, args, num_args);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, portprompt);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNborderWidth, 0);	num_args++;
    ttlprompt = XtCreateManagedWidget("ttlprompt", labelWidgetClass,
				   netform, args, num_args);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, portprompt);	num_args++;
    XtSetArg(args[num_args], XtNfromHoriz, ttlprompt);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNeditType, XawtextEdit);	num_args++;
    XtSetArg(args[num_args], XtNresize, XawtextResizeWidth);	num_args++;
    XtSetArg(args[num_args], XtNstring, "");		num_args++;
    ttlresponse = XtCreateManagedWidget("ttlresponse", asciiTextWidgetClass,
				     netform, args, num_args);

						num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, ttlprompt);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNborderWidth, 0);	num_args++;
    connection_IDprompt = XtCreateManagedWidget("connection_IDprompt", labelWidgetClass,
				   netform, args, num_args);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, ttlprompt);	num_args++;
    XtSetArg(args[num_args], XtNfromHoriz, connection_IDprompt);   num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNeditType, XawtextEdit);	num_args++;
    XtSetArg(args[num_args], XtNresize, XawtextResizeWidth);	num_args++;
    XtSetArg(args[num_args], XtNstring, "");		num_args++;
    connection_IDresponse = XtCreateManagedWidget("connection_IDresponse", asciiTextWidgetClass,
				     netform, args, num_args);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, connection_IDprompt);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNlabel,"Sender");        num_args++;
    XtSetArg(args[num_args], XtNradioData,SENDER);      num_args++;
    sender = XtCreateManagedWidget("sender", toggleWidgetClass, netform,
				 args, num_args);

    XtSetArg(args[num_args], XtNfromVert, connection_IDprompt);	num_args++;
    XtSetArg(args[num_args], XtNfromHoriz, sender);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNradioGroup, sender);	num_args++;
    XtSetArg(args[num_args], XtNlabel,"Receiver");      num_args++;
    XtSetArg(args[num_args], XtNradioData,RECEIVER);    num_args++;
    receiver = XtCreateManagedWidget("receiver", toggleWidgetClass, netform,
				 args, num_args);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, sender);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    XtSetArg(args[num_args], XtNaccelerators,
	     XtParseAcceleratorTable(okay_accelerators));	num_args++;
    okay = XtCreateManagedWidget("okay", commandWidgetClass,
				 netform, args, num_args);
    XtAddCallback(okay, XtNcallback, okay_callback, netform);

							num_args = 0;
    XtSetArg(args[num_args], XtNfromVert, sender);	num_args++;
    XtSetArg(args[num_args], XtNfromHoriz, okay);	num_args++;
    XtSetArg(args[num_args], XtNtop, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNbottom, XtChainTop);	num_args++;
    XtSetArg(args[num_args], XtNleft, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNright, XtChainLeft);	num_args++;
    XtSetArg(args[num_args], XtNresizable, True);	num_args++;
    cancel = XtCreateManagedWidget("cancel", commandWidgetClass,
				   netform, args, num_args);
    XtAddCallback(cancel, XtNcallback, cancel_callback, parent);

    XtInstallAccelerators(addressresponse, okay);
    XtInstallAccelerators(portresponse, okay);
  
    return netform;
}

/* get the prompt from the dialog box.  Used a startup time to
 * save away the initial prompt */
String
GetAddressDialogPrompt(w)
    Widget w;
{
    Arg args[1];
    Widget addresslabel;
    String s;

    addresslabel = XtNameToWidget(w, "addressprompt");
    XtSetArg(args[0], XtNlabel, &s);
    XtGetValues(addresslabel, args, ONE);
    return XtNewString(s);
}    

String
GetPortDialogPrompt(w)
    Widget w;
{
    Arg args[1];
    Widget portlabel;
    String s;

    portlabel = XtNameToWidget(w, "portprompt");
    XtSetArg(args[0], XtNlabel, &s);
    XtGetValues(portlabel, args, ONE);
    return XtNewString(s);
}    

String
GetTtlDialogPrompt(w)
    Widget w;
{
    Arg args[1];
    Widget ttllabel;
    String s;

    ttllabel = XtNameToWidget(w, "ttlprompt");
    XtSetArg(args[0], XtNlabel, &s);
    XtGetValues(ttllabel, args, ONE);
    return XtNewString(s);
}    

String
GetIdDialogPrompt(w)
    Widget w;
{
    Arg args[1];
    Widget idlabel;
    String s;

    idlabel = XtNameToWidget(w, "connection_IDprompt");
    XtSetArg(args[0], XtNlabel, &s);
    XtGetValues(idlabel, args, ONE);
    return XtNewString(s);
}    


/* set the prompt.  This is used to put error information in the prompt */
void
SetAddressDialogPrompt(w, newprompt)
    Widget w;
    String newprompt;
{
    Arg args[1];
    Widget addresslabel;

    addresslabel = XtNameToWidget(w, "addressprompt");
    XtSetArg(args[0], XtNlabel, newprompt);
    XtSetValues(addresslabel, args, ONE);
}

void
SetPortDialogPrompt(w, newprompt)
    Widget w;
    String newprompt;
{
    Arg args[1];
    Widget portlabel;

    portlabel = XtNameToWidget(w, "portprompt");
    XtSetArg(args[0], XtNlabel, newprompt);
    XtSetValues(portlabel, args, ONE);
}

void
SetTtlDialogPrompt(w, newprompt)
    Widget w;
    String newprompt;
{
    Arg args[1];
    Widget ttllabel;

    ttllabel = XtNameToWidget(w, "ttlprompt");
    XtSetArg(args[0], XtNlabel, newprompt);
    XtSetValues(ttllabel, args, ONE);
}

void
SetIdDialogPrompt(w, newprompt)
    Widget w;
    String newprompt;
{
    Arg args[1];
    Widget idlabel;

    idlabel = XtNameToWidget(w, "connection_IDprompt");
    XtSetArg(args[0], XtNlabel, newprompt);
    XtSetValues(idlabel, args, ONE);
}

/* get what the user typed */
String
GetAddressDialogResponse(w)
    Widget w;
{
    Arg args[1];
    Widget addressresponse;
    String s;

    addressresponse = XtNameToWidget(w, "addressresponse");
    XtSetArg(args[0], XtNstring, &s);
    XtGetValues(addressresponse, args, ONE);
    return XtNewString(s);
}

String
GetPortDialogResponse(w)
    Widget w;
{
    Arg args[1];
    Widget portresponse;
    String s;

    portresponse = XtNameToWidget(w, "portresponse");
    XtSetArg(args[0], XtNstring, &s);
    XtGetValues(portresponse, args, ONE);
    return XtNewString(s);
}

String
GetTtlDialogResponse(w)
    Widget w;
{
    Arg args[1];
    Widget ttlresponse;
    String s;

    ttlresponse = XtNameToWidget(w, "ttlresponse");
    XtSetArg(args[0], XtNstring, &s);
    XtGetValues(ttlresponse, args, ONE);
    return XtNewString(s);
}

String
GetIdDialogResponse(w)
    Widget w;
{
    Arg args[1];
    Widget connection_IDresponse;
    String s;

    connection_IDresponse = XtNameToWidget(w, "connection_IDresponse");
    XtSetArg(args[0], XtNstring, &s);
    XtGetValues(connection_IDresponse, args, ONE);
    return XtNewString(s);
}

/* set the default reponse */
void
SetAddressDialogResponse(w, s)
    Widget w;
    String s;
{
    Arg args[3];
    Widget addressresponse;
    XFontStruct *font;
    Dimension width, leftMargin, rightMargin;
    addressresponse = XtNameToWidget(w, "addressresponse");
    XtSetArg(args[0], XtNfont, &font);
    XtSetArg(args[1], XtNleftMargin, &leftMargin);
    XtSetArg(args[2], XtNrightMargin, &rightMargin);
    XtGetValues(addressresponse, args, THREE);
    width = font->max_bounds.width * strlen(s) + leftMargin + rightMargin;

    XtSetArg(args[0], XtNstring, s);
    XtSetArg(args[1], XtNwidth, width);
    XtSetValues(addressresponse, args, TWO);
    XawTextSetInsertionPoint(addressresponse, strlen(s));

}

void
SetPortDialogResponse(w, s)
    Widget w;
    String s;
{
    Arg args[3];
    Widget portresponse;
    XFontStruct *font;
    Dimension width, leftMargin, rightMargin;
    portresponse = XtNameToWidget(w, "portresponse");
    XtSetArg(args[0], XtNfont, &font);
    XtSetArg(args[1], XtNleftMargin, &leftMargin);
    XtSetArg(args[2], XtNrightMargin, &rightMargin);
    XtGetValues(portresponse, args, THREE);
    width = font->max_bounds.width * strlen(s) + leftMargin + rightMargin;

    XtSetArg(args[0], XtNstring, s);
    XtSetArg(args[1], XtNwidth, width);
    XtSetValues(portresponse, args, TWO);
    XawTextSetInsertionPoint(portresponse, strlen(s));

}

void
SetTtlDialogResponse(w, s)
    Widget w;
    String s;
{
    Arg args[3];
    Widget ttlresponse;
    XFontStruct *font;
    Dimension width, leftMargin, rightMargin;

    ttlresponse = XtNameToWidget(w, "ttlresponse");
    XtSetArg(args[0], XtNfont, &font);
    XtSetArg(args[1], XtNleftMargin, &leftMargin);
    XtSetArg(args[2], XtNrightMargin, &rightMargin);
    XtGetValues(ttlresponse, args, THREE);
    width = font->max_bounds.width * strlen(s) + leftMargin + rightMargin;

    XtSetArg(args[0], XtNstring, s);
    XtSetArg(args[1], XtNwidth, width);
    XtSetValues(ttlresponse, args, TWO);
    XawTextSetInsertionPoint(ttlresponse, strlen(s));

}

void
SetIdDialogResponse(w, s)
    Widget w;
    String s;
{
    Arg args[3];
    Widget connection_IDresponse;
    XFontStruct *font;
    Dimension width, leftMargin, rightMargin;

    connection_IDresponse = XtNameToWidget(w, "connection_IDresponse");
    XtSetArg(args[0], XtNfont, &font);
    XtSetArg(args[1], XtNleftMargin, &leftMargin);
    XtSetArg(args[2], XtNrightMargin, &rightMargin);
    XtGetValues(connection_IDresponse, args, THREE);
    width = font->max_bounds.width * strlen(s) + leftMargin + rightMargin;

    XtSetArg(args[0], XtNstring, s);
    XtSetArg(args[1], XtNwidth, width);
    XtSetValues(connection_IDresponse, args, TWO);
    XawTextSetInsertionPoint(connection_IDresponse, strlen(s));

}

/* clear the response */
void
ClearAddressDialogResponse(w)
    Widget w;
{
    Arg args[2];
    Widget addressresponse;

    addressresponse = XtNameToWidget(w, "addressresponse");
    XtSetArg(args[0], XtNstring, "");
    XtSetArg(args[1], XtNwidth, 100);
    XtSetValues(addressresponse, args, TWO);
}

void
ClearPortDialogResponse(w)
    Widget w;
{
    Arg args[2];
    Widget portresponse;

    portresponse = XtNameToWidget(w, "portresponse");
    XtSetArg(args[0], XtNstring, "");
    XtSetArg(args[1], XtNwidth, 100);
    XtSetValues(portresponse, args, TWO);
}

void
ClearTtlDialogResponse(w)
    Widget w;
{
    Arg args[2];
    Widget ttlresponse;

    ttlresponse = XtNameToWidget(w, "ttlresponse");
    XtSetArg(args[0], XtNstring, "");
    XtSetArg(args[1], XtNwidth, 100);
    XtSetValues(ttlresponse, args, TWO);
}

void
ClearIdDialogResponse(w)
    Widget w;
{
    Arg args[2];
    Widget connection_IDresponse;

    connection_IDresponse = XtNameToWidget(w, "connection_IDresponse");
    XtSetArg(args[0], XtNstring, "");
    XtSetArg(args[1], XtNwidth, 100);
    XtSetValues(connection_IDresponse, args, TWO);
}
