/*
nevot.h
-------

Copyright 1992-1995 by Henning Schulzrinne
University of Massachusetts, AT&T Bell Laboratories, GMD Fokus
All rights reserved.  

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that: (1) source code
distributions retain the above copyright notice and this paragraph in
its entirety, (2) distributions including binary code include the
above copyright notice and this paragraph in its entirety in the
documentation or other materials provided with the distribution, and
(3) all advertising materials mentioning features or use of this
software display the following acknowledgment: ``This product includes
software developed by the University of Massachusetts, AT&T Bell
Laboratories and its contributors.'' Neither the name of the
University nor that of AT&T nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission. 

THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  

$Id: nevot.h,v 1.1.1.1 1997/11/12 17:16:33 dor Exp $
*/
#include "types.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>      /* STD??_FILENO */
#include <sys/time.h>    /* timeval */
#include <tcl.h>         /* Tcl_HashTable */
#include <tk.h>          /* TK event handlers */
#include "audio_descr.h"
#include "mbuf.h"
#include "sync.h"
#include "tsap.h"
#include "notify.h"      /* notify_set_input_func() */
#include "stats.h"
#include "rate.h"

/****  DO NOT CHANGE ANYTHING BELOW THIS LINE ***/
/* maximum pre-talkspurt cache */
#define MAX_CACHE 10
/* maximum packetization interval, in sec */
#define MAX_PACKETIZATION 0.2

/* packet type and commands; values to be used for mbuf types must 
   be less than 16 */
#define BM(x,b) x = (1<<b) 
typedef enum {
  BM(P_bot,               1),     /* first packet in talkspurt */
  BM(P_eot,               2),     /* last block in talkspurt */
  /* control packets */
  BM(P_bye,               5),
  BM(P_bye_all,           6),
  BM(P_sdes,              8),
  BM(P_sdes_all,          9),
  BM(P_csrc,              10),
  BM(P_traffic,           11)
} packet_t;

typedef enum {
  BM(DEBUG_RX,    1),
  BM(DEBUG_TX,    2),
  BM(DEBUG_AUDIO, 3), 
  BM(DEBUG_SD,    4),
  BM(DEBUG_TS,    5),
  BM(DEBUG_CNTL,  6)
} debug_t;

typedef u_int32 src_id_t;
#define SDES_MAX 15

/*
 * status of one conference (possibly with several members);
 * creation time: -> local site; 
 * NBO: network byte order
 */
#include "ring.h"
#include "Audio_sp.h"
typedef struct sdes {
  int type;              /* type of information */
  char *value;           /* SDES value */ 
  int len;               /* length without \0 */
  struct sdes *next;     /* link to next item in linked list */
} sdes_t;

typedef struct Audio_session_t {
  /* linkage to other sessions */
  struct Audio_session_t *next,*prev;   /* chain of session descriptors */
  ring_t *ring;                         /* ring descriptor */

  /* network I/O: audio and control */
  struct {
    void *data, *control;                     /* protocol structures */
    int (*read)(int sock, void *buf,          /* read from network */
      int size, tsap_t *tsap);
    int (*write)(void *, mbuf_t *, tsap_t *); /* write to network */
    int (*set_ttl)(void *, int ttl);          /* ttl modification */
    int (*close)(void *);                     /* close network for member */
    int ttl;                                  /* multicast time-to-live */
    struct sockaddr_in serv_addr;             /* address of the opened 
					       * data socket to which we will
					       * send data or receive from */
    
    struct sockaddr_in my_addr;                /* the interface address out
						* of which to send and receive
						* the data -only important
						* for multihomed hosts */
  } net;

  /* real-time protocol */
  struct rt_t {
    mbuf_t *(*write_data)(struct Audio_session_t *s,
      ring_t *ring, int type, mbuf_t *m);
    mbuf_t *(*write_control)(struct Audio_session_t *s,
      int type, char *reason);
    u_int32 (*packets_exp)(struct Audio_member_t *s);
    int (*change_audio)(struct rt_t *rt, audio_descr_t *);
    int audio_header;                   /* opaque audio header encoding */
  } rt;

  char *name;                           /* session name */
  Audio_sp_t sp;                        /* session connection descriptor */
  sdes_t *sdes;                         /* list of SDES items (any order) */
  boolean no_more;                      /* accept no more members */
  boolean talk;
  
  int connected;           /* 1= connection established 
			    * 2= path sending required but no connection yet
			    * 0= no connection or path requirements */
  /*
   * Listen to session; if disabled, incoming packets are not
   * decoded and talkers are not indicated to avoid confusion.
   * The audio device may be released.
   */
  boolean listen;

  /* encryption */
  boolean encrypt;                      /* encrypt outgoing/incoming */
  void *key;                            /* pointer to key (depends on lib.) */

  /* duplex mode */
  struct {
    int net_priority:1;
    int local_priority:1;
  } m;

  /* outgoing packet */
  audio_descr_t audio;                  /* audio format */
  int local_audio_channel;              /* play on this channel */
  double packetization;                 /* packetization interval (sec) */
  int bpm;                              /* blocks per message */
  int block;                            /* block index in current message */
  mbuf_t *b_head, *b_tail;              /* block pointer */
  u_int32 initial_ts;                   /* initial timestamp */
  u_int32 seq;                          /* current data sequence number */
  u_int32 packets[2];                   /* number of data/control packets sent */
  u_int32 samples;                      /* samples sent */
  u_int32 bytes;                        /* bytes sent */
  int senders;                          /* # of active senders */
  struct {
    double bw;                          /* bandwidth, bytes/sec */
    int avg;                            /* smoothed size */
    int last;                           /* last size */
    double period;                      /* nominal period in seconds */
  } control;                            /* (RTCP) control packets */
  rate_t rate_state;                    /* state for rate conversion */
  int force_ts;                         /* force a talkspurt */

  Tcl_HashTable member_table;           /* hash table for members */

  /*
   * One descriptor for each site. s->local points to local site.
   */
  int members;                          /* # of members (-1: only allocated) */
  struct Audio_member_t {
    /* linkage to session and other members */
    struct Audio_member_t *next,*prev;  /* chain of site descriptors */
    struct Audio_session_t *s;          /* parent session */

    struct sockaddr_in rmt_addr; /* address of the member*/
    /* information for data and control information */
    struct timeval last[3];             /* last arrival (data, control, SR) */
    tsap_t nsrc[2];                     /* network source address */
    tsap_t ndst[2];                     /* network destination address */ 
    src_id_t src;                       /* incoming source [NBO!] */
    src_id_t bridge;                    /* bridge identifier (SSRC) [NBO!] */
    char *id;                           /* printable identifier */
    sdes_t *sdes;                       /* SDES items */

    /* activity timestamps */
    time_t created;                     /* site created */
    u_int32 lsr_sec;                    /* last sender report */
    u_int32 lsr_frac;                   /* last sender report */

    /*** state ***/
    /*
     * The last audio sample received from this session member (end of
     * audio packet); measured in samples.
     */
    u_int32 last_sample;
    /*
     * Listen to this member; if false, we still need to update
     * the talker display. For s->local, this flag indicates us
     * talking instead.
     */
    boolean listen;
    /*
     * This session member is sending media data. Reset when
     * sending control messages.
     */
    boolean active;
    /*
     * This session member has been marked as talking.
     */
    boolean talking;
    /*
     * A soft time-out has occurred for this session member.
     */
    boolean soft_to;

    /*
     * Member statistics.
     */
    stats_t stats;


    /* audio */
    sync_t sync;                        /* synchronization status */
    audio_descr_t audio;                /* audio coding from net */
    short audio_header;                 /* opaque audio encoding in header */
    short covers;                       /* samples covered by last packet */
    boolean talkspurt;                  /* new talkspurt received */
    void *decoder_state;                /* pointer to state area */
  } *first,
    *local,                             /* we ourselves (local site) */
    *cache;                             /* last arrival */
   /* RSVP Structure */
    struct {
      int sid;   /* sid for path session */
      int rsid;  /* sid for reservation session*/
      int rate;
      char* Sender_Template;
      char* Sender_Tspec;
      char* Sender_Policy;
      int   ttl;
      int   fd;
      Tcl_File  tcl_fd;
      int   path;  /*1 if path packets were received for this rsvp session */
      Tk_TimerToken RSVP_TimerToken;   /* Timer token for retrying 
         			  	  reservations */
    } rsvp;

} *Audio_session_t;

typedef struct Audio_member_t *Audio_member_t;
typedef struct rt_t rt_t;

/* needs Audio_session_t, Audio_member_t */
#define AUDIO_INTERNAL
#include "Audio.h"

typedef struct {
  /* parameters that are externally setable */
  enum {R_end,R_gateway} role; /* role: gateway, end node */
  int audio_priority;          /* 0: on demand; 1: idle; 2: idle+demand; 3: never */
  
  int tsap_count;              /* valid addresses */
  tsap_t tsap[3];              /* own addresses */
  struct {
    boolean waiting;           /* somebody's waiting for her audio turn */
    boolean talk;              /* enable talking (summary of st->talk) */
    int talking;               /* we're talking */
    /*
     * Basic audio clock tick, in seconds. Audio is read and played
     * back with this period.
     */
    double tick;               /* seconds per chunk */
  } audio;

  struct {
    audio_descr_t ad;          /* current in/out audio format */
    int samples;               /* chunk to read/write (samples) */
    int bytes;                 /* same, but bytes */
  } tx, rx;

  Audio_session_t first;       /* linked list of session */
  Audio_handler_t event;       /* callback for events */
  int maxlen;                  /* buffer length */
} nevot_global_t;

extern nevot_global_t nevot;

/* traverse circular lists */
#define ALL_MEMBERS(s) \
  for (m = (s)->first; m; m = m->next)
#define ALL_SESSIONS(s) \
  for ((s) = nevot.first; (s); (s) = (s)->next) 

/* external functions */

/* audio.c */
extern void audio_vu(int i, float value);
extern void audio_vu_max(int i, float value);
extern Audio_errno_t audio_open(void);
extern void audio_close(void);
extern Audio_errno_t audio_change(int pr, audio_descr_t *ad);
extern void audio_out(Audio_member_t m, audio_descr_t *ad);
extern void audio_chunk(void);

/* convert.c */
extern void *convert_reset(audio_encoding_t c, void *state, double vocoder);
extern int bytes2samples(audio_descr_t *adp, int bytes);

/* member.c */
extern void member_sdes(Audio_member_t m, const SDES_t type,
  const char *data, int len);
extern Audio_member_t member_find(Audio_session_t s, int control, 
  tsap_t *nsrc, src_id_t src);

/* ring_mix.c */
extern Audio_errno_t ring_mix(Audio_member_t *m, int sources, 
  audio_descr_t *in_ad, char *b, int in_size, ring_t *ring, u_int32 p);

/* rtp.c */
extern void rtp_init(rt_t *rt);
extern Audio_errno_t rtp_read_data(Audio_session_t s, mbuf_t *mb,
  tsap_t *nsrc, int *cc, Audio_member_t *slist, Audio_member_t *m);
extern Audio_errno_t rtp_read_control(Audio_session_t s, mbuf_t *mb,
  tsap_t *nsrc, Audio_member_t *m);

/* cmd.c -the RSVP part */
extern void rsvp_close_all(Audio_session_t s);
extern void rsvp_close( int * sid );
extern void rsvp_send(Audio_session_t session);
extern void rsvp_open_rec(Audio_session_t session, int state);
extern int rsvp_rec(Audio_member_t m, Audio_session_t session);


/* rx.c */
extern void play_local(void *sidetone, audio_descr_t *ad, int sidetone_length,
  int monitor, int mute);
extern void play_gateway(void);
extern Notify_value rx(Notify_client client, int sock);
extern Audio_errno_t rx_file(int fd, char *info, char *fn);
extern Audio_errno_t rx_file_stop(void);
extern Audio_errno_t rx_file_pause(int flag); 
extern void rx_init(void);
extern void rx_periodic(Audio_session_t s);

/* session.c */
extern void session_enc(Audio_session_t s);
extern Audio_errno_t session_member_open(Audio_member_t, tsap_t *);
extern int session_ismulticast(Audio_session_t s);

/* talking.c */
extern void talking_clear(Audio_session_t s);
extern void talking_delete(Audio_member_t m);
extern void talking_set(ring_t *r, Audio_member_t m);
extern int talking_net(ring_t *r);
extern void talking_check(ring_t *r);

/* tx.c */
extern Notify_value tx(Notify_client client, int fd);
#ifdef AUDIO_UNDEF 
extern Audio_errno_t tx_file(int fd, audio_file_t *af);
#endif
extern Audio_errno_t tx_file_stop(void);
extern Audio_errno_t tx_file_pause(int flag);
extern Audio_errno_t tx_file_monitor(int flag);
extern Audio_errno_t tx_file_loop(int flag);
extern void tx_start(void);
extern void tx_init(void);
extern u_int32 tx_samples(void);
extern Notify_value tx_periodic(Notify_client client);

/* tx_packet.c */
extern void tx_packet_init(void);
extern int  tx_packet_session(Audio_session_t s, ring_t *ring_skip,
  int packet_type);
extern void tx_packet(mbuf_t *mb, void *ring_skip);
extern void tx_packet_close(void);

/* vat.c */
extern void vat_init(rt_t *rt);
extern Audio_errno_t vat_read_data(Audio_session_t s, mbuf_t *mb,
  tsap_t *nsrc, int *cc, Audio_member_t *mlist, Audio_member_t *m);
extern Audio_errno_t vat_read_control(Audio_session_t s, mbuf_t *mb,
  tsap_t *nsrc, Audio_member_t *m);
