/*
cache  --  speech cache, used for keeping the most recent buffers.

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include "mbuf.h"
#include <stdio.h>



static int cache_occupancy;   /* current number of mbufs in cache */
static mbuf_t *c_head, *c_tail;

/*
* Add packet to existing cache. If the cache size exceeds defined
* length 'limit' (in packets), throw away oldest entry. Set 'discard'
* if packet was thrown away; leave as is if not.
*/
void cache_write(mbuf_t *m, int limit, int *discarded)
{
  m->m_next = 0;
  if (c_head) {
    c_tail = c_tail->m_next = m;
    cache_occupancy++;
  }
  else {
    c_head = c_tail = m;
    cache_occupancy = 1;
  }
  if (cache_occupancy > limit) {
    m = c_head;
    c_head = c_head->m_next;
    m_free(m, "cw");
    *discarded = 1;
    cache_occupancy--;
  }
} /* cache_write */


/*
* Get first (oldest) entry from cache.
*/
mbuf_t *cache_read(void)
{
  mbuf_t *m;

  m = c_head;
  if (c_head) {
    c_head = c_head->m_next;
    cache_occupancy--;
  }
  return m;
} /* cache_read */
