/*
debug  --  display debug messages to stderr

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include <stdio.h>    /* fprintf() */
#include <string.h>   /* strdup() */
#include <stdlib.h>   /* malloc() */
#include <stdarg.h>   /* ... */
#include <sys/time.h> /* gettimeofday() */
#include "debug.h"
#include <ansi.h>



static struct {
  FILE *f;
  char *fn;
  int  changed;
  int  mask; 
} d = {stderr, 0, 0};

/*
* Set current debug output file. Open file only when debug
* mask changes. Empty name ("") denotes stderr.
*/
void debug_file(char *fn)
{
  d.changed = !d.fn || (strcmp(fn, d.fn) != 0);
  if (d.changed) {
    if (d.fn) free(d.fn);
    d.fn = strdup(fn);
  }
} /* debug_file */


/*
* Set current debug mask. Append to existing files.
* On failure, continue to stderr.
*/
void debug_mask(int mask)
{
  if (mask != d.mask) {
    if (d.changed) {
      if (d.f != stderr) fclose(d.f);
      if (*d.fn) {
        d.f = fopen(d.fn, "a");
        if (!d.f) d.f = stderr;
      }
      else d.f = stderr;
    }
    d.mask = mask;
  }
} /* debug_init */ 

/*
* Finish debug.
*/
void debug_close(void)
{
  if (d.f) {
    fclose(d.f);
    d.f = NULL;
  }
  if (d.fn) {
    free(d.fn);
    d.fn = NULL;
  }
}

/*
* Write debug message to current file if 'type' matches mask.
*/
void debug(int type, char *fmt, ...)
{
  va_list args;
  char buffer[256];
  struct timeval tv;

  if (type & d.mask) {
    va_start(args, fmt);
    gettimeofday(&tv, 0);
    fprintf(d.f, "%3ld.%06ld ", tv.tv_sec, tv.tv_usec);
    vsprintf(buffer, fmt, args);
    va_end(args);
    fputs(buffer, d.f); fputc('\n', d.f);
  }
} /* debug */
