/*
event  --  generic event handler (using notifier)

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include <sys/time.h>
#include "notify.h"



/*
* GUI-specific wrapper for input function handler. No action if file
* descriptor fd is < 0.
*/
void set_input_func(int fd, Notify_func func, void *client)
{
  if (fd >= 0)
    notify_set_input_func((Notify_client)client, func, fd);
} /* set_input_func */


Notify_func set_signal_func(int signal, Notify_func handler, void *client)
{
  return notify_set_signal_func((Notify_client)client, handler,
    signal, NOTIFY_ASYNC);
} /* set_signal_func */


/*
 * Periodic invocation function.  A 'interval' value of zero disables
 * the function.
*/
void set_periodic_func(void (*func)(), void *client, int interval)
{
#define ITIMER_NULL ((struct itimerval *)0)
  if (interval == 0) {
    notify_set_itimer_func((Notify_client)client,
           NOTIFY_FUNC_NULL, ITIMER_REAL, ITIMER_NULL, ITIMER_NULL);
  }
  else {
    struct itimerval timer;

    timer.it_interval.tv_usec = (interval % 1000) * 1000;
    timer.it_interval.tv_sec  = interval / 1000;
    timer.it_value = timer.it_interval;
    notify_set_itimer_func((Notify_client)client, (Notify_func)func,
      ITIMER_REAL, &timer, ITIMER_NULL);
  }
} /* set_periodic_func */
