/*
* Handle TCP messages framed like tcl-dp.
*/
#include "types.h"         /* u_int32 */
#include <unistd.h>
#include <netinet/in.h>    /* htonl() */
#include <errno.h>
#include <stdio.h>
#include <ansi.h>

#define PACKET_MAGIC        0x6feeddcc

extern int errno;

/*
* Read (TCP) message, framed like tcl-dp (32-bit MAGIC, 32-bit length,
* including header). No trailing zero. Return zero on EOF.
*/
int frame_read(int fd, char *buf, int nbyte)
{
  char *b;     /* buffer pointer */
  long rcount; /* bytes read */
  int n;       /* bytes left */
  long flen;   /* frame length */
  long magic;  /* magic number */

  /* read magic number first */
  for (b = (char *)&magic, n = 4; n > 0; n -= rcount, b += rcount) {
    if ((rcount = read(fd, b, n)) < n) goto error;
  }
  magic = ntohl(magic);
  if (magic != PACKET_MAGIC) return -1;

  /* read framing length */
  for (b = (char *)&flen, n = 4; n > 0; n -= rcount, b += rcount) {
    if ((rcount = read(fd, b, n)) < n) goto error;
  }
  flen = ntohl(flen) - 8;
  if (flen > nbyte) return -1;

  /* read data part */
  for (b = buf, n = flen; n > 0; n -= rcount, b += rcount) {
    if ((rcount = read(fd, b, n)) < n) goto error;
  }
  return flen;

error:
  if (rcount < 0 && errno == ECONNRESET) return 0;
  return rcount;
} /* frame_read */


/*
* Write (TCP) message, prefixed by magic number and length.
*/
int frame_write(int fd, char *buf, int nbyte)
{
  struct {
    u_int32 magic;
    u_int32 nbyte;
  } header;
  int sent;

  header.magic = htonl(PACKET_MAGIC);
  header.nbyte = htonl(nbyte + sizeof(header));
  sent = write(fd, (char *)&header, sizeof(header));
  if (sent < sizeof(header)) return sent;
  sent = write(fd, buf, nbyte);
  return sent;
} /* frame_write */
