/*
 * COPYRIGHT BY OERTEL & ZAHL SOFTWAREENTWICKLUNG GbR
 *	file:		ibm_audio.c
 *	creation date:	06-Mar-1995
 *	last update:	<obsolete>
 *	written by:	Frank Oertel, Christian Zahl
 *	description:	Device dependent part of AUDIO support for the
 *			IBM ULTIMEDIA Adapter
 *
 * $Id: ibm_audio.c,v 1.1.1.1 1997/11/12 17:16:28 dor Exp $
 *
 * $Log: ibm_audio.c,v $
 * Revision 1.1.1.1  1997/11/12 17:16:28  dor
 *
 * Revision 1.1  1995/08/17  13:58:26  hgs
 * Initial revision
 *
 * Revision 1.18  1995/06/30  16:17:18  czahl
 * - bug: while searching for an audio device filename, the comparison
 *   was wrong! Result: no audio :-(
 *
 * Revision 1.17  1995/06/28  11:40:27  czahl
 * - minor changes so that it can be compiled in other environments than
 *   mine.
 *
 * Revision 1.16  1995/06/27  15:21:51  czahl
 * - After a long time of changes, trail and so on, we are now in the
 *   happily position to make this release to an alpha version!!!
 *
 * Revision 1.15  1995/06/25  16:19:06  czahl
 * - alpha 0. Before we begin to support the audioctl tcl extension
 *   via ahw_set, we wonna save this version.
 *
 * Revision 1.14  1995/06/22  19:52:12  czahl
 * *** empty log message ***
 *
 * Revision 1.13  1995/06/19  12:23:27  czahl
 * - now after we got some very helpfull hints from Paul Wadehra, we
 *   implemented some new sets of commands for setting various audio
 *   parameters. This was done for testing purposes only because it was
 *   easier to do this rather than changing ahw_set.
 * - Because we now we know how to deal with the audio device, we will
 *   switch back to use ahw_set for changing the audio parameters.
 *   Just for commenting, we save this nice version...
 *
 * Revision 1.12  1995/06/12  10:42:44  czahl
 * - after a long period of testing and understanding the behavior of
 *   the UMS audio adapter, we decided to support just a relatively
 *   poor interface. Before doing so, we want to save this bloody
 *   version first...
 *
 * Revision 1.11  1995/05/10  15:13:10  czahl
 * - some minor changes
 *
 * Revision 1.10  1995/04/21  12:48:01  czahl
 * - minor changes
 *
 * Revision 1.9  1995/04/20  17:43:45  czahl
 * - the logging facillity has been extracted to log.c. Currently this
 *   will be included but will be a stand-alone module in the near
 *   future
 * - A stripping mechanism has been installed for the case that the
 *   write buffer becommes to big which results in an increasing delay.
 *   This is done by a two stage mechanism. First every 16th sample
 *   will be discarded, which resulst in a faster pitch and decreased the
 *   buffer. Second the whole frame will be thrown away above a special
 *   threashold.
 *
 * Revision 1.8  1995/04/20  10:39:46  czahl
 * - all the UMS related stuff have been removed
 * - audio I/O will now be done with read/write/ioctl
 * - some kind of logging debugging informations has been installed.
 *   It will be used for tuning the audio part but can also be
 *   used for other purposes
 * - direct select support for the audio device has been used.
 *   Unfortunately this seems NOT to work with the old M-ACPA, just
 *   with the Ultimedia Audio Adapter (6302)
 *
 * Revision 1.7  1995/04/18  17:14:40  czahl
 * - this is the first working release by using the UMS audio
 *   stuff.
 *
 * Revision 1.6  1995/04/17  11:46:44  czahl
 * - first real tries with record and play
 *
 * Revision 1.5  1995/03/16  22:09:10  czahl
 * - minor changes
 *
 * Revision 1.4  1995/03/16  22:02:37  czahl
 * - bug in ahw_close; missing flag
 * - better ahw_read simulation for continous data stream
 *
 * Revision 1.3  1995/03/16  19:15:23  czahl
 * - using MU_LAW instead of PCM for testing
 *
 *update history: <obsolete>
 * date      who	why
 * 06-Mar-95 Chris	kommentieren der Funktionen, um UMS einbauen zu
 *			knnen...
 * 10-Mar-95 Chris	weiter gehts...
 *			nun ein Dummy zu debuggen
 * 16-Mar-95 Chris	problem mit "boolean", wird in UMS und include/types.h
 *			definiert; zum Glueck gleich...
 *			+ RCS
 */
static	char rcsid[] = "$Id: ibm_audio.c,v 1.1.1.1 1997/11/12 17:16:28 dor Exp $";
/*--------------------------------------------------------------------*/

/*** standard includefiles ***/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/audio.h>		/* this is the new one from Paul!!! */
#include <sys/acpa.h>
#ifndef AUDIO_MODIFY_LIMITS
#  include "./ibm_audio.h"
#endif

/*** user includefiles ***/
#define	NO_BOOLEAN
#include "audio_generic.h"
#include "ahw.h"
#include "util.h"
#include "config.h"
#ifdef CZAHL
#  define WITH_TRACE		0
#  include "/u/include/debug.h"
#else
#  define TRACE(xxx)
#  define FATAL_ERROR(xxx) {printf xxx; printf ("\n"); exit (1);}
#endif

/*** defines ***/

/*** type definitions ***/
typedef struct {
	int			f;				/* filedesc. */
	int			Mode;				/* O_RDONLY, O_WRONLY */
	char			*Fn;				/* filename for the device */
	int			Operation;			/* RECORD or PLAY */
	int			Output;				/* OUTPUT_1, ... */
	int			Input;				/* MIC, ... */
	int			Rate;				/* sample rate */
	int			Fmt;				/* audio format */
	int			Channels;			/* # of channels (1 / 2 */
	int			BitsPerSample;			/* 8 or 16 */
	unsigned int		Flags;				/* for signed / unsigned */
	int			BlockSize;			/* for the DMA page */
	int			Monitor;			/* monitoring on / off */
	int			MonitorVolume;			/* monitoring Volume 0..255 */
	int			Volume;				/* playback volume 0..255 */
	int			Balance;			/* balance while playback -100..100 */
	int			LeftGain;			/* left gain level 0..255 */
	int			RightGain;			/* right gain level 0..255 */
	int			IntSpeaker;			/* on or off */
	int			Port;				/* currently selected ports */
	int			HighWater;			/* when we start to reduce the buffer */
} t_AudioDev;

/*** local functions ***/
static	int			AhwSetRecord (audio_set_t What, int Val);
static	int			AhwSetPlayback (audio_set_t What, int Val);
static	int			AhwSetMonitor (audio_set_t What, int Val);
static	int			AudioDevClose (t_AudioDev *Dev);
static	int			AudioDevOpen (t_AudioDev *Dev);
static	void			AudioInitDev ();
	/*** for IBM audio hardware ***/
static	int			IbmAudioInitAudioChange (audio_change *AudioChange);
static	int			IbmAudioInitAudioControl (audio_control *AudioControl);
static	int			IbmAudioSelectInput (int f, int Input);
static	int			IbmAudioSelectOutput (int f, int Output);
static	int			IbmAudioSetBalance (int f, int Balance);
static	int			IbmAudioSetGain (int f, int LeftGain, int RightGain);
static	int			IbmAudioSetIntSpeaker (int f, int Enabled);
static	int			IbmAudioSetMonitor (int f, int Monitor);
static	int			IbmAudioSetVolume (int f, int Volume);
static	int			IbmAudioStart (int f);
static	int			IbmAudioStop (int f);

/*** global functions ***/

/*** local variables ***/
static	t_AudioDev		_PlayDev;			/* the playback device */
static	t_AudioDev		_RecDev;			/* the recording device */
static	t_AudioDev		*PlayDev = &_PlayDev;		/* for better handling */
static	t_AudioDev		*RecDev = &_RecDev;		/* for better handling */

static	char			*RecDevFn[] = {
					"/dev/baud0/1",
					"/dev/baud1/1",
					"/dev/paud0/1",
					"/dev/paud1/1",
					NULL};
static	char			*PlayDevFn[] = {
					"/dev/baud0/2",
					"/dev/baud1/2",
					"/dev/paud0/2",
					"/dev/paud1/2",
					NULL};

/*** global variables ***/
	int			ahw_changed;			/* flag: audio device has changed */
	int			ahw_errno;			/* error indication */


/*--------------------------------------------------------------------*/
int ahw_open (int ndelay)
{
	TRACE (("ahw_open (ndelay = %d)\n", ndelay));
	AudioInitDev ();
/*** open and start the record part ***/
	AudioDevOpen (RecDev);
	AudioDevInit (RecDev);
	IbmAudioSelectInput (RecDev->f, RecDev->Input);
	IbmAudioSetGain (RecDev->f, RecDev->LeftGain, RecDev->RightGain);
	IbmAudioSetMonitor (RecDev->f, RecDev->Monitor);
/*	IbmAudioSetMonitorVolume (RecDev->f, RecDev->MonitorVol); */
	IbmAudioSetCaptureBuffer (RecDev->f, 2);
	IbmAudioStart (RecDev->f);
/*** open and start the playback part ***/
	AudioDevOpen (PlayDev);
	AudioDevInit (PlayDev);
	IbmAudioSelectOutput (PlayDev->f, PlayDev->Output);
	IbmAudioSetVolume (PlayDev->f, PlayDev->Volume);
	IbmAudioSetBalance (PlayDev->f, PlayDev->Balance);
	IbmAudioSetIntSpeaker (PlayDev->f, PlayDev->IntSpeaker);
	IbmAudioStart (PlayDev->f);
	return RecDev->f;					/* return fd for select call */
} /* ahw_open */
/*--------------------------------------------------------------------*/
int ahw_close (void)
{
	int			OldFd;

	TRACE (("ahw_close ()"));
/*** close the audio device ***/
	OldFd = RecDev->f;					/* we need it (see below) */
	IbmAudioStop (RecDev->f);				/* stop recording */
	IbmAudioStop (PlayDev->f);				/* and playback now */
	AudioDevClose (RecDev);					/* close the recording */
	AudioDevClose (PlayDev);				/* and playback track */
/*** ***/
	return OldFd;						/* return the old input fd */
} /* ahw_close */
/*--------------------------------------------------------------------*/
static int AhwSetRecord (audio_set_t What, int Val)
/*
 * Does all the settings for the recording part.
 */
{
TRACE (("AhwSetRecord (%d, %d)", What, Val));
	switch (What) {
/*** select input source ? ***/
	  case AS_port:
	    if (Val == AP_in_line_1)				/* LINE-IN (external) */
	      RecDev->Input = LINE_1;
	    else if (Val == AP_in_line_2)			/* AUX-1 (internal CD-ROM) */
	      RecDev->Input = AUX1;
	    else if (Val == AP_in_microphone_high)		/* high gain mike */
	      RecDev->Input = HIGH_GAIN_MIKE;
	    else if (Val == AP_in_microphone_low)		/* low gain mike */
	      RecDev->Input = LOW_GAIN_MIKE;
	    else						/* anything else? */
	      return -1;					/* n/a */
	    IbmAudioSelectInput (RecDev->f, RecDev->Input);	/* select the port */
	    return 0;
/*** set the gain for left and right ***/
	  case AS_gain:						/* input gain for left=right */
	  case AS_gain_right:					/* input gain for right */
	  case AS_gain_left:					/* input gain for left */
	    if (Val < 0)
	      Val = 0;
	    else if (Val > 255)
	      Val = 255;
	    if ((What == AS_gain) || (What == AS_gain_left))
	      RecDev->LeftGain = Val;
	    if ((What == AS_gain) || (What == AS_gain_right))
	      RecDev->RightGain = Val;
	    IbmAudioSetGain (RecDev->f, RecDev->LeftGain, RecDev->RightGain);
	    return 0;
/*** set gain step ***/
	  case AS_gain_step:					/* ??? */
	    TRACE (("AS_gain_step is not supported"));
	    return 0;						/* n/a, ignore */
/*** set fillpoint ***/
	  case AS_fillpoint:					/* ??? */
	    TRACE (("AS_fillpoint is not supported"));
	    return 0;						/* currently not supported */
/*** bullshit? ***/
	  default:
	    TRACE (("\007unknown action: ahw_set (record, %d, %d)", What, Val));
	    return -1;
	  } /* switch */
} /* AhwSetRecord */
/*--------------------------------------------------------------------*/
static int AhwSetPlayback (audio_set_t What, int Val)
/*
 * Does all the settings for the playback part.
 */
{
TRACE (("AhwSetPlayback (%d, %d)", What, Val));
	switch (What) {
/*** select output ? ***/
	  case AS_port:
	    if (Val == AP_out_line_1) {				/* external output */
	      PlayDev->Port = AP_out_line_1;
	      PlayDev->Output = OUTPUT_1;
/*
 *!!! we don't HAVE TO select the output once again, because the internal speaker
 *!!! will be reset to the default (on) value
	      IbmAudioSelectOutput (PlayDev->f, PlayDev->Output);
 */
	      }
	    else if (Val == AP_out_speaker) {			/* internal speaker */
	      PlayDev->Port = AP_out_speaker;
	      }
	    else						/* anything else ? */
	      return -1;					/* n/a */
	    return 0;
/*** set the gain for left and right ***/
	  case AS_gain:						/* output volume for left=right */
	  case AS_gain_right:					/* output volume for right */
	  case AS_gain_left:					/* output volume for left */
TRACE (("AhwSetPlayback.gain (%d, %d)", What, Val));
	    if (Val < 0)
	      Val = 0;
	    else if (Val > 255)
	      Val = 255;
	    if (PlayDev->Port == AP_out_speaker) {
	      if (Val >= 128)
	        PlayDev->IntSpeaker = 1;
	      else
	        PlayDev->IntSpeaker = 0;
	      IbmAudioSetIntSpeaker (PlayDev->f, PlayDev->IntSpeaker);
	      }
	    else {
	      PlayDev->Volume = Val;				/* we don't distinguish between left and right */
	      IbmAudioSetVolume (PlayDev->f, PlayDev->Volume);
	      }
	    return 0;
/*** set balance ***/
	  case AS_balance:					/* output balance */
	    if (Val < -255)
	      Val = -255;
	    else if (Val > 255)
	      Val = 255;
	    PlayDev->Balance = Val;
	    IbmAudioSetBalance (PlayDev->f, PlayDev->Balance);
	    return 0;
/*** set gain step ***/
	  case AS_gain_step:					/* ??? */
	    TRACE (("AS_gain_step is not supported"));
	    return 0;						/* n/a, ignore */
/*** set fillpoint ***/
	  case AS_fillpoint:					/* ??? */
	    TRACE (("AS_fillpoint is not supported"));
	    return 0;						/* currently not supported */
/*** bullshit? ***/
	  default:
	    TRACE (("\007unknown action: ahw_set (playback, %d, %d)", What, Val));
	    return -1;
	  } /* switch */
} /* AhwSetPlayback */
/*--------------------------------------------------------------------*/
static int AhwSetMonitor (audio_set_t What, int Val)
/*
 * Does all the settings for the monitoring part.
 */
{
TRACE (("AhwSetMonitor (%d, %d)", What, Val));
	switch (What) {
/*** select output ? ***/
	  case AS_port:
	    return -1;						/* n/a */
	    return 0;
/*** set the gain for left and right ***/
	  case AS_gain:						/* output volume for left=right */
	  case AS_gain_right:					/* output volume for right */
	  case AS_gain_left:					/* output volume for left */
	    if (Val < 0)
	      Val = 0;
	    else if (Val > 255)
	      Val = 255;
	    return 0;
/*** set gain step ***/
	  case AS_gain_step:					/* ??? */
	    TRACE (("AS_gain_step is not supported"));
	    return 0;						/* n/a, ignore */
/*** set fillpoint ***/
	  case AS_fillpoint:					/* ??? */
	    TRACE (("AS_fillpoint is not supported"));
	    return 0;						/* currently not supported */
/*** bullshit? ***/
	  default:
	    TRACE (("\007unknown action: ahw_set (monitor, %d, %d)", What, Val));
	    return -1;
	  } /* switch */
} /* AhwSetMonitor */
/*--------------------------------------------------------------------*/
int ahw_set (int Device, audio_set_t What, int Val)
{

	AudioInitDev ();
	switch (Device) {
	  case AD_RECORD:
	    return AhwSetRecord (What, Val);
	  case AD_PLAYBACK:
	    return AhwSetPlayback (What, Val);
	  case AD_MONITOR:
	    return AhwSetMonitor (What, Val);
	  default:						/* oops!?$%& */
	    TRACE (("\007unknown action: ahw_set (%d, %d, %d)", Device, What, Val));
	    return -1;
	  } /* switch */
} /* ahw_set */
/*--------------------------------------------------------------------*/
int ahw_fmt(audio_descr_t *Rec, audio_descr_t *Play)
{
fprintf (stderr, "ahw_fmt (Rec=..., Play=...)\n");
fprintf (stderr, "\tRec->\n");
fprintf (stderr, "\t\tenc\t%d\n", Rec->encoding);
fprintf (stderr, "\t\trate\t%d\n", Rec->sample_rate);
fprintf (stderr, "\t\tchan\t%d\n", Rec->channels);
fprintf (stderr, "\tPlay->\n");
fprintf (stderr, "\t\tenc\t%d\n", Play->encoding);
fprintf (stderr, "\t\trate\t%d\n", Play->sample_rate);
fprintf (stderr, "\t\tchan\t%d\n", Play->channels);

	AudioInitDev ();
/*** check Rec format ***/
	RecDev->Flags = FIXED;
	switch (Rec->encoding) {
	  case AE_PCMU:
TRACE (("mu-law"));
	    RecDev->Fmt = MU_LAW;
	    RecDev->BitsPerSample = 8;
	    break;
	  case AE_PCMA:
TRACE (("a-law"));
	    RecDev->Fmt = A_LAW;
	    RecDev->BitsPerSample = 8;
	    break;
	  case AE_L8:
TRACE (("pcm"));
	    RecDev->Fmt = PCM;
	    RecDev->BitsPerSample = 8;
	    break;
	  case AE_L16:
	    RecDev->Fmt = PCM;
	    RecDev->BitsPerSample = 16;
	    RecDev->Flags = FIXED | SIGNED;
	    break;
	  default:
	    Rec->encoding = AE_PCMU;
	    RecDev->Fmt = MU_LAW;
	    RecDev->BitsPerSample = 8;
	    break;
	  } /* switch */
/*	audio_descr (Rec);					/* update some paramters */
	RecDev->Rate = Rec->sample_rate;
	RecDev->Channels = Rec->channels;
	RecDev->BlockSize = 256;			/* !!! depends on rate */
/*** check play format ***/
	PlayDev->Flags = FIXED;
	switch (Play->encoding) {
	  case AE_PCMU:
	    PlayDev->Fmt = MU_LAW;
	    PlayDev->BitsPerSample = 8;
	    break;
	  case AE_PCMA:
	    PlayDev->Fmt = A_LAW;
	    PlayDev->BitsPerSample = 8;
	    break;
	  case AE_L8:
	    PlayDev->Fmt = PCM;
	    PlayDev->BitsPerSample = 8;
	    break;
	  case AE_L16:
	    PlayDev->Fmt = PCM;
	    PlayDev->BitsPerSample = 16;
	    PlayDev->Flags = FIXED | SIGNED;
	    break;
	  default:
	    Play->encoding = AE_PCMU;
	    PlayDev->Fmt = MU_LAW;
	    PlayDev->BitsPerSample = 8;
	    break;
	  } /* switch */
/*	audio_descr (Play);					/* update some parameters */
	PlayDev->Rate = Play->sample_rate;
	PlayDev->Channels = Play->channels;
	PlayDev->BlockSize = 128*3;
#define ms *8
	PlayDev->HighWater = 130 ms;
	return 0;						/* ok */
} /* ahw_fmt */
/*--------------------------------------------------------------------*/
int ahw_get (audio_state_t *AudioState)
{
  TRACE (("ahw_get"));
  AudioInitDev ();
  return 0;
} /* ahw_get */
/*--------------------------------------------------------------------*/
int ahw_waiting (void)
{
	TRACE (("ahw_waiting ()"));
	return 0;
} /* ahw_waiting */
/*--------------------------------------------------------------------*/
void ahw_track (signal_func_t Handler)
/* Not supported or undocumented feature for the IBM RS/6000 family. */
{
	TRACE (("ahw_track: IS NOT SUPPORTED"));
} /* ahw_track */
/*--------------------------------------------------------------------*/
int ahw_write_queue (int *Underrun)
/*
 * Returns the number of bytes in the write queue. If an underrun
 * condition has been occured, we return a negative value and set
 * Underrun to TRUE, otherwise to FLASE.
 * NOTE: Because we handle audio underruns in a special manner, we
 * will allways lie and no not indicate any underun!
 */
{
	audio_buffer		AudioBuffer;

	if (ioctl (PlayDev->f, AUDIO_BUFFER, &AudioBuffer) == -1) {
	  TRACE (("ioctl"));
	  return -1;
	  }
	*Underrun = 0;
	return AudioBuffer.write_buf_size;
} /* ahw_write_queue */
/*--------------------------------------------------------------------*/
int ahw_read_queue(int *errors)
{
	TRACE (("ahw_read_queue (errors=%d)", *errors));
	return 0;
} /* ahw_read_queue */
/*--------------------------------------------------------------------*/
int ahw_write (char *Buf, int Bytes)
{
#define	SKIP_INTERVAL		(32)
static	int			SkipFlag = FALSE;
static	int			HighWaterReached = 0;
	audio_buffer		BufStat;
	int			n;
	int			x;
	int			i;

	if (PlayDev->f == -1) {					/* device not open? */
	  TRACE (("ahw_write: dev is not open"));
	  return Bytes;						/* we need to lie */
	  } /* if */
/*** get the buffer status ***/
	if (ioctl (PlayDev->f, AUDIO_BUFFER, &BufStat) == -1) {	/* get buffer state */
	  TRACE (("ahw_write: unable to get buffer status"));
	  return 0;
	  }
/* printf ("%d ", BufStat.write_buf_size); fflush (stdout); */
/*** check if we need buffer adjustment ***/
	if (BufStat.write_buf_size >= PlayDev->HighWater)	/* buffer too full, should shrink? */
	  SkipFlag = TRUE;					/* set flag */
	if (SkipFlag) {
	  HighWaterReached++;
	  if (HighWaterReached < 10)
	    SkipFlag = 0;
	  else
	    HighWaterReached = 0;
	  }
	else
	  HighWaterReached = 0;
/*** eventualy reduce the buffer by omitting some samples ***/
	if (SkipFlag) {						/* reduce the buffer? */
printf ("[%3d ms]", BufStat.write_buf_size /8);
fflush (stdout);
	  for (i=0, x=0; i<Bytes; i++)
	    if (! (i % SKIP_INTERVAL))
	      continue;
	    Buf[x++] = Buf[i];
	  }
	else							/* no reducing? */
	  x = Bytes;
/*** write the data to the device ***/
	n = write (PlayDev->f, Buf, x);				/* write to playback track */
	if (n == -1) {						/* error ? */
printf ("[W error %d]", errno); fflush (stdout);
	  return -1;
	  }
	return Bytes;						/* always lie; all has been written */
} /* ahw_write */
/*--------------------------------------------------------------------*/
int ahw_read (char *Buf, int Bytes)
{
	int			n;
	audio_buffer		AudioBuffer;

	if (RecDev->f == -1) {					/* track is not open? */
	  TRACE (("ahw_read: dev not open"));
	  return Bytes;						/* lie anyway */
	  } /* if */
ioctl (RecDev->f, AUDIO_BUFFER, &AudioBuffer);
if (AudioBuffer.flags != 0) {
  printf ("[R-%d]", AudioBuffer.flags);
  fflush (stdout);
}
	n = read (RecDev->f, Buf, Bytes);			/* read the Data */
	if (n == -1)						/* error occured? */
	  if (errno == EINTR)					/* interrupted syscall? */
	    return 0;						/* ok, try later again */
	  else {						/* real error ? */
	    TRACE (("read"));
	    return -1;
	    } /* else */
	return n;						/* return # bytes captured */
} /* ahw_read */
/*--------------------------------------------------------------------*/
static AudioDevOpen (t_AudioDev *Dev)
{
	TRACE (("AudioDevOpen (%p)", Dev));
	AudioInitDev ();						/* initialize internal control structures */
/*** check if we allready have it done ***/
	if (Dev->f != -1)					/* device is open ? */
	  return 0;						/* that's all folks */
/*** open the device ***/
	Dev->f = open (Dev->Fn, Dev->Mode);			/* open the right device part */
	if (Dev->f == -1) {					/* error ? */
	  TRACE (("unable to open"));
	  FATAL_ERROR (("open"));
	  return -1;						/* report error */
	  }
	return 0;
} /* AudioDevOpen */
/*--------------------------------------------------------------------*/
static AudioDevInit (t_AudioDev *Dev)
/*
 * Initializes on of the audio tracks. The track has to be opened and
 * it has to stoped. Otherwise onpredicted results occurs.
 */
{
	audio_init		AudioInit;

	TRACE (("AudioDevInit (%p)", Dev));
	AudioInitDev ();						/* initialize internal control structures */
	AudioInit.srate =		Dev->Rate;		/* sample rate */
	AudioInit.bits_per_sample =	Dev->BitsPerSample;	/* bits / sample */
	AudioInit.bsize =		Dev->BlockSize;		/* DMA block size */
	AudioInit.mode =		Dev->Fmt;		/* encoding format */
	AudioInit.channels =		Dev->Channels;		/* 1 or 2 */
	AudioInit.position_resolution =	AUDIO_IGNORE;		/* no interest now */
	AudioInit.flags =		Dev->Flags;		/* needed Flags */
	AudioInit.operation =		Dev->Operation;		/* record or playback */
	AudioInit.reserved =		NULL;			/* we are nice today... */
	if (ioctl (Dev->f, AUDIO_INIT, &AudioInit) == -1) {
	  TRACE (("AudioDevInit: ioctl"));
	  FATAL_ERROR (("AudioDevInit: ioctl"));
	  return -1;
	  }
printf ("AudioInit.bsize = %d\n", Dev->BlockSize);
} /* AudioDevInit */
/*--------------------------------------------------------------------*/
static AudioDevClose (t_AudioDev *Dev)
{
	TRACE (("AudioDevClose (%p)", Dev));
	if (Dev->f == -1)					/* already closed? */
	  return 0;						/* ok, job done */
	close (Dev->f);						/* close it */
	Dev->f = -1;						/* and mark it as closed */
	return 0;						/* return success */
} /* AudioDevClose */
/*--------------------------------------------------------------------*/
static void AudioInitDev ()
/*
 * Initializes the characteristics of the audio device, for the recording
 * and playback part.
 */
{
static	int			Init = 0;
	int			i;
	int			f;

	if (Init)
	  return;
	Init = 1;
/*** search audio device ***/
	for (i=0; RecDevFn[i] != NULL; i++)
	  if ((f = open (RecDevFn[i], O_RDONLY)) != -1)
	    break;
	  else
	    if (errno != ENOENT)
	      break;
	if (f != -1)
	  close (f);
/*** record ***/
	RecDev->f =		-1;
	RecDev->Fn =		RecDevFn[i];
	RecDev->Mode =		O_RDONLY | O_NDELAY;
	RecDev->Operation =	RECORD;
	RecDev->Input = 	AUDIO_IGNORE;
	RecDev->Rate = 		8000;
	RecDev->Fmt = 		PCM;
	RecDev->Channels = 	1;
	RecDev->BitsPerSample =	8;
	RecDev->Flags = 	FIXED;
	RecDev->BlockSize =	AUDIO_IGNORE;
	RecDev->Monitor = 	MONITOR_OFF;
	RecDev->LeftGain =	0;
	RecDev->RightGain =	0;
	RecDev->Volume =	AUDIO_IGNORE;
	RecDev->Balance = 	AUDIO_IGNORE;
	RecDev->Output = 	AUDIO_IGNORE;
	RecDev->IntSpeaker =	AUDIO_IGNORE;
/*** playback ***/
	PlayDev->f =		-1;
	PlayDev->Fn =		PlayDevFn[i];
	PlayDev->Mode =		O_WRONLY | O_NDELAY;
	PlayDev->Operation =	PLAY;
	PlayDev->Output = 	OUTPUT_1;
	PlayDev->Rate = 	8000;
	PlayDev->Fmt = 		PCM;
	PlayDev->Channels = 	1;
	PlayDev->BitsPerSample = 8;
	PlayDev->Flags = 	FIXED;
	PlayDev->BlockSize =	AUDIO_IGNORE;
	PlayDev->Volume =	0;
	PlayDev->Balance = 	0;
	PlayDev->IntSpeaker =	0;
	PlayDev->Input = 	AUDIO_IGNORE;
	PlayDev->Monitor = 	AUDIO_IGNORE;
	PlayDev->LeftGain =	AUDIO_IGNORE;
	PlayDev->RightGain =	AUDIO_IGNORE;
} /* AudioInitDev */
/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/
/*--------------------------------------------------------------------*/
/*
 * PRIVAT ROUTINES FOR SETTING THE AUDIO PARAMTERS
 * THESE ROUTINES CAN BE USED IN GENERAL BECAUSE THEY DON'T DEPEND
 * ON NEVOT.
 */

/*--------------------------------------------------------------------*/
static IbmAudioInitAudioChange (audio_change *AudioChange)
{
	AudioChange->dev_info =		NULL;
	AudioChange->input =		AUDIO_IGNORE;
	AudioChange->output =		AUDIO_IGNORE;
	AudioChange->monitor =		AUDIO_IGNORE;
	AudioChange->volume =		AUDIO_IGNORE;
	AudioChange->volume_delay =	AUDIO_IGNORE;
	AudioChange->balance =		AUDIO_IGNORE;
	AudioChange->balance_delay =	AUDIO_IGNORE;
	AudioChange->treble =		AUDIO_IGNORE;		/* not supported */
	AudioChange->bass =		AUDIO_IGNORE;		/* not supported */
	AudioChange->pitch =		AUDIO_IGNORE;		/* not supported */
} /* IbmAudioInitAudioChange */
/*--------------------------------------------------------------------*/
static IbmAudioInitAudioControl (audio_control *AudioControl)
{
	AudioControl->ioctl_request =	AUDIO_IGNORE;
	AudioControl->request_info =	NULL;
	AudioControl->position =	0;
	AudioControl->return_code =	0xDeadBeaf;
} /* IbmAudioInitAudioControl */
/*--------------------------------------------------------------------*/
static IbmAudioStart (int f)
/*
 * Starts the specified device.
 */
{
	audio_control		AudioControl;

TRACE (("IbmAudioStart (%d)", f));
	if (f == -1)
	  return 0;
	IbmAudioInitAudioControl (&AudioControl);
	AudioControl.ioctl_request = AUDIO_START;		/* start cmd */
	if (ioctl (f, AUDIO_CONTROL, &AudioControl) != 0) {
	  TRACE (("unable to issue AUDIO_START"));
	  perror ("AudioDevStart:ioctl");
	  return -1;						/* return failiure */
	  }
	return 0;						/* return success */
} /* IbmAudioStart */
/*--------------------------------------------------------------------*/
static IbmAudioStop (int f)
{
	audio_control		AudioControl;

TRACE (("IbmAudioStop (%d)", f));
	if (f == -1)
	  return 0;
	IbmAudioInitAudioControl (&AudioControl);
	AudioControl.ioctl_request = AUDIO_STOP;		/* stop cmd */
	if (ioctl (f, AUDIO_CONTROL, &AudioControl) != 0) {
	  TRACE (("unable to issue AUDIO_STOP"));
	  perror ("AudioDevStop:ioctl");
	  return -1;						/* return failiure */
	  } /* if */
	return 0;						/* return success */
} /* IbmAudioStop */
/*--------------------------------------------------------------------*/
static IbmAudioSetVolume (int f, int Volume)
/*
 * Volume:	0..255
 */
{
	audio_change		AudioChange;
	audio_control		AudioControl;

TRACE (("IbmAudioSet VOlume (%d, %d)", f, Volume));
	if (f == -1)
	  return 0;
/*** range check ***/
	if (Volume < 0)
	  Volume = 0;
	else if (Volume > 255)
	  Volume = 255;
/*** set it ***/
	IbmAudioInitAudioChange (&AudioChange);
	IbmAudioInitAudioControl (&AudioControl);
	AudioChange.volume = 0x7fff0000 / 255 * Volume;
	AudioChange.volume_delay = 0;
	AudioControl.request_info = &AudioChange;
	AudioControl.ioctl_request = AUDIO_CHANGE;
	if (ioctl (f, AUDIO_CONTROL, &AudioControl) == -1) {
	  TRACE (("AudioDevSetVolume"));
	  }
	return 0;
} /* IbmAudioSetVolume */
/*--------------------------------------------------------------------*/
static int IbmAudioSetBalance (int f, int Balance)
/*
 * Blance:	-255..0..255
 */
{
	audio_change		AudioChange;
	audio_control		AudioControl;

TRACE (("IbmAudioSetBalance (%d, %d)", f, Balance));
	if (f == -1)
	  return 0;
/*** range check ***/
	if (Balance < -255)
	  Balance = -255;
	else if (Balance > 255)
	  Balance = 255;
	PlayDev->Balance = Balance;
/*** scaling ***/
	Balance += 255;
	Balance = 0x7fff0000 / (511) * Balance;
/*** setting ***/
	IbmAudioInitAudioChange (&AudioChange);
	IbmAudioInitAudioControl (&AudioControl);
	AudioChange.balance = Balance;
	AudioChange.balance_delay = 0;
	AudioControl.request_info = &AudioChange;
	AudioControl.ioctl_request = AUDIO_CHANGE;
	if (ioctl (f, AUDIO_CONTROL, &AudioControl) == -1) {
	  TRACE (("AudioDevSetBalance"));
	  }
	return 0;
} /* IbmAudioSetBalance */
/*--------------------------------------------------------------------*/
static int IbmAudioSetGain (int f, int LeftGain, int RightGain)
/*
 * LeftGain:	0..255
 * RightGain:	0..255
 */
{
	audio_set_gain		AudioSetGain;

TRACE (("IbmAudioSetGain (%d, %d, %d)", f, LeftGain, RightGain));
	if (f == -1)
	  return 0;
/*** range checks ***/
	if (LeftGain < 0)
	  LeftGain = 0;
	else if (LeftGain > 255)
	  LeftGain = 255;
	if (RightGain < 0)
	  RightGain = 0;
	else if (RightGain > 255)
	  RightGain = 255;
/*** set it ***/
	AudioSetGain.left_gain = LeftGain * 100 / 255;
	AudioSetGain.right_gain = RightGain * 100 / 255;
	if (ioctl (f, AUDIO_SET_GAIN, &AudioSetGain) == -1) {
	  TRACE (("AudioDevSetGain"));
	  }
	return 0;
} /* IbmAudioSetGain */
/*--------------------------------------------------------------------*/
static int IbmAudioSelectInput (int f, int Input)
{
	audio_change		AudioChange;
	audio_control		AudioControl;

TRACE (("IbmAudioSelectInput (%d, %d)", f, Input));
	if (f == -1)
	  return 0;
	IbmAudioInitAudioChange (&AudioChange);
	IbmAudioInitAudioControl (&AudioControl);
	AudioChange.input = Input;
	AudioControl.request_info = &AudioChange;
	AudioControl.ioctl_request = AUDIO_CHANGE;
	if (ioctl (f, AUDIO_CONTROL, &AudioControl) == -1) {
	  TRACE (("ioctl"));
	  }
	return 0;
} /* IbmAudioSelectInput */
/*--------------------------------------------------------------------*/
static int IbmAudioSelectOutput (int f, int Output)
{
	audio_change		AudioChange;
	audio_control		AudioControl;

TRACE (("IbmAudioSelectOutput (%d, %d)", f, Output));
	if (f == -1)
	  return 0;
	IbmAudioInitAudioChange (&AudioChange);
	IbmAudioInitAudioControl (&AudioControl);
	AudioChange.output = Output;
	AudioControl.request_info = &AudioChange;
	AudioControl.ioctl_request = AUDIO_CHANGE;
	if (ioctl (f, AUDIO_CONTROL, &AudioControl) == -1) {
	  TRACE (("ioctl"));
	  }
	return 0;
} /* IbmAudioSelectOutput */
/*--------------------------------------------------------------------*/
IbmAudioSetIntSpeaker (int f, int Enabled)
{
	audio_change		AudioChange;
	audio_control		AudioControl;

TRACE (("IbmAudioSetIntSpeaker (%d, %d)", f, Enabled));
	if (f == -1)
	  return 0;
	IbmAudioInitAudioChange (&AudioChange);
	IbmAudioInitAudioControl (&AudioControl);
	if (Enabled)
	  AudioChange.output = INTERNAL_SPEAKER_ON;
	else
	  AudioChange.output = INTERNAL_SPEAKER_OFF;
	AudioControl.request_info = &AudioChange;
	AudioControl.ioctl_request = AUDIO_CHANGE;
	if (ioctl (f, AUDIO_CONTROL, &AudioControl) == -1) {
	  TRACE (("ioctl"));
	  }
	return 0;
} /* IbmAudioSetIntSpeaker */
/*--------------------------------------------------------------------*/
static int IbmAudioSetMonitor (int f, int Monitor)
{
	audio_change		AudioChange;
	audio_control		AudioControl;

TRACE (("IbmAudioSetMonitor (%d, %d)", f, Monitor));
	if (f == -1)
	  return 0;
	IbmAudioInitAudioChange (&AudioChange);
	IbmAudioInitAudioControl (&AudioControl);
	if (Monitor)
	  AudioChange.monitor = MONITOR_UNCOMPRESSED;
	else
	  AudioChange.monitor = MONITOR_OFF;
	AudioControl.request_info = &AudioChange;
	AudioControl.ioctl_request = AUDIO_CHANGE;
	if (ioctl (f, AUDIO_CONTROL, &AudioControl) == -1) {
	  TRACE (("ioctl"));
	  }
	return 0;
} /* IbmAudioSetMonitor */
/*--------------------------------------------------------------------*/
static int IbmAudioSetCaptureBuffer (int f, int Time)
{
	init_buf_req		BufReq;

	BufReq.buffer_size = 8000* Time;	/* has to be multiple of #bytes/s */
	BufReq.play_block_count = AUDIO_IGNORE;
	BufReq.play_lower_limit = AUDIO_IGNORE;
	BufReq.cap_upper_limit = AUDIO_IGNORE;
	BufReq.byte_count = AUDIO_IGNORE;
	BufReq.return_code = 0xDeadBeaf;
	if (ioctl (f, AUDIO_MODIFY_LIMITS, &BufReq) == -1) {
	  TRACE (("ioctl"));
	  }
	return 0;
} /* IbmAudioSetCaptureBuffer */
