/*
* Library functions for Solaris.
*/

#include "types.h"
#include "arch_sys.h"
#include <sys/systeminfo.h> /* sysinfo() */
#include <math.h>           /* HUGE_VAL, exp(), pow() */
#include <stdlib.h>         /* atoi() */
#include <signal.h>
#include <time.h>

double infinity(void)
{
  return HUGE_VAL;
} /* infinity */


double exp10(double x)
{
  return pow(10., x);
}


long gethostid(void)
{
  char buf[80];

  sysinfo(SI_HW_SERIAL, buf, sizeof(buf));
  return atoi(buf);
} /* gethostid */


typedef void Sigfunc(int);

/*
* Reliable versio of signal(), using POSIX sigaction().
* (From Stevens, p. 298).
*/
Sigfunc *signal(int signo, Sigfunc *func)
{
  struct sigaction act, oact;

  act.sa_handler = func;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if (signo == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;
#endif    
  }
  else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;
#endif
  }
  if (sigaction(signo, &act, &oact) < 0) return SIG_ERR;
  return oact.sa_handler;
} /* signal */
