/*
 * COPYRIGHT BY OERTEL & ZAHL SOFTWAREENTWICKLUNG GbR
 *	file:		audioctl.c
 *	written by:	Christian Zahl
 *	description:	Tcl/Tk extensions for audio control
 *
 * $Id: cmd_audioctl.c,v 1.1.1.1 1997/11/12 17:16:32 dor Exp $
 *
 * $Log: cmd_audioctl.c,v $
 * Revision 1.1.1.1  1997/11/12 17:16:32  dor
 *
 * Revision 1.2  1995/08/30  11:11:25  hgs
 * 3.32
 *
 * Revision 1.1  1995/08/17  13:54:58  hgs
 * Initial revision
 *
 * Revision 1.2  1995/06/27  13:17:07  czahl
 * - Privat routines for IBM substituted by ahw_set
 * - Added headspeaker support
 *
 * Revision 1.1  1995/06/27  12:47:44  czahl
 * Initial revision
 *
 */


/*** standard includefiles ***/
#include <stdlib.h>   /* atoi() */
#include <string.h>   /* strcmp() */
#include <tcl.h>

/*** user includefiles ***/
#include "ahw.h"

/*** local functions ***/
static	int		ScanBalance (Tcl_Interp *Interp, int argc, char *argv[]);
static	int		ScanGain (Tcl_Interp *Interp, int argc, char *argv[]);
static	int		ScanInput (Tcl_Interp *Interp, int argc, char *argv[]);
static	int		ScanIntSpeaker (Tcl_Interp *Interp, int argc, char *argv[]);
static	int		ScanLeftGain (Tcl_Interp *Interp, int argc, char *argv[]);
static	int		ScanOutput (Tcl_Interp *Interp, int argc, char *argv[]);
static	int		ScanRightGain (Tcl_Interp *Interp, int argc, char *argv[]);
static	int		ScanVolume (Tcl_Interp *Interp, int argc, char *argv[]);

/*-------------------------------------------------------------------------*/
static int ScanOption (Tcl_Interp *Interp, int argc, char *argv[])
{
	if (strcmp (*argv, "-input") == 0)
	  return ScanInput (Interp, argc, argv);
	if (strcmp (*argv, "-output") == 0)
	  return ScanOutput (Interp, argc, argv);
	if (strcmp (*argv, "-volume") == 0)
	  return ScanVolume (Interp, argc, argv);
	if (strcmp (*argv, "-balance") == 0)
	  return ScanBalance (Interp, argc, argv);
	if (strcmp (*argv, "-gain") == 0)
	  return ScanGain (Interp, argc, argv);
	if (strcmp (*argv, "-leftgain") == 0)
	  return ScanLeftGain (Interp, argc, argv);
	if (strcmp (*argv, "-rightgain") == 0)
	  return ScanRightGain (Interp, argc, argv);
	if (strcmp (*argv, "-intspeaker") == 0)
	  return ScanIntSpeaker (Interp, argc, argv);
	Interp->result = "bad option, should be -input, -output, -intspeaker, ",
		"-volume, -gain, -leftgain, -rightgain";
	return -1;
} /* ScanOption */
/*-------------------------------------------------------------------------*/
static int ScanInput (Tcl_Interp *Interp, int argc, char *argv[])
{
	int			Input = -1;

	if (argc <= 1) {
	  Interp->result = "Missing option to -input";
	  return -1;
	  }
	if (strcmp (argv[1], "lowgainmike") == 0)
	  Input = AP_in_microphone_low;
	else if (strcmp (argv[1], "highgainmike") == 0)
	  Input = AP_in_microphone_high;
	else if (strcmp (argv[1], "line1") == 0)
	  Input = AP_in_line_1;
	else if (strcmp (argv[1], "line2") == 0)
	  Input = AP_in_line_2;
	else if (strcmp (argv[1], "line3") == 0)
	  Input = AP_in_line_3;
	else {
	  Interp->result = "Bad option, should be "
		"lowgainmike, highgainmike, line1, line2, or line3";
	  return -1;
	  }	
	ahw_set (AD_RECORD, AS_port, Input);
	return 2;
} /* ScanInput */
/*-------------------------------------------------------------------------*/
static int ScanOutput (Tcl_Interp *Interp, int argc, char *argv[])
{
	int			Output = -1;

	if (argc <= 1) {
	  Interp->result = "Missing option to -output";
	  return -1;
	  }
	if (strcmp (argv[1], "line1") == 0)
	  Output = AP_out_line_1;
	else if (strcmp (argv[1], "line2") == 0)
	  Output = AP_out_line_2;
	else if (strcmp (argv[1], "line2") == 0)
	  Output = AP_out_line_3;
	else {
	  Interp->result = "bad option, should be line1, line2 or line3";
	  return -1;
	  }	
	ahw_set (AD_PLAYBACK, AS_port, Output);
	return 2;
} /* ScanOutput */
/*-------------------------------------------------------------------------*/
static int ScanIntSpeaker (Tcl_Interp *Interp, int argc, char *argv[])
{
	int			Val = -1;

	if (argc <= 1) {
	  Interp->result = "Missing option to -intspeaker";
	  return -1;
	  }
	if (strcmp (argv[1], "on") == 0)
	  Val = 1;
	else if (strcmp (argv[1], "off") == 0)
	  Val = 0;
	if (strcmp (argv[1], "1") == 0)
	  Val = 1;
	else if (strcmp (argv[1], "0") == 0)
	  Val = 0;
	else {
	  Interp->result = "bad option, should be on, 1, off or 0";
	  return -1;
	  }	
	ahw_set (AD_PLAYBACK, AS_port, AP_out_speaker);
	if (Val)
	  ahw_set (AD_PLAYBACK, AS_gain, 255);
	else
	  ahw_set (AD_PLAYBACK, AS_gain, 0);
	return 2;
} /* ScanIntSpeaker */
/*-------------------------------------------------------------------------*/
static int ScanVolume (Tcl_Interp *Interp, int argc, char *argv[])
{
	int			Val = -1;

	if (argc <= 1) {
	  Interp->result = "Missing option to -volume";
	  return -1;
	  }
	Val = atoi (argv[1]);
	if (Val < 0)
	  Val = 0;
	else if (Val > 100)
	  Val = 100;
	ahw_set (AD_PLAYBACK, AS_gain, 256 * Val / 100);
	return 2;
} /* ScanVolume */
/*-------------------------------------------------------------------------*/
static int ScanBalance (Tcl_Interp *Interp, int argc, char *argv[])
{
	int			Val = -1;

	if (argc <= 1) {
	  Interp->result = "Missing option, to -balance";
	  return -1;
	  }
	Val = atoi (argv[1]);
	if (Val < -100)
	  Val = -100;
	else if (Val > 100)
	  Val = 100;
	ahw_set (AD_PLAYBACK, AS_balance, 256 * Val / 100);
	return 2;
} /* ScanBalance */
/*-------------------------------------------------------------------------*/
static int ScanGain (Tcl_Interp *Interp, int argc, char *argv[])
{
	int			Val = -1;

	if (argc <= 1) {
	  Interp->result = "Missing option to -gain";
	  return -1;
	  }
	Val = atoi (argv[1]);
	if (Val < 0)
	  Val = 0;
	else if (Val > 100)
	  Val = 100;
	ahw_set (AD_RECORD, AS_gain, 256 * Val / 100);
	return 2;
} /* ScanGain */
/*-------------------------------------------------------------------------*/
static int ScanLeftGain (Tcl_Interp *Interp, int argc, char *argv[])
{
	int			Val = -1;

	if (argc <= 1) {
	  Interp->result = "Missing option, to -leftgain";
	  return -1;
	  }
	Val = atoi (argv[1]);
	if (Val < 0)
	  Val = 0;
	else if (Val > 100)
	  Val = 100;
	ahw_set (AD_RECORD, AS_gain_left, 256 * Val / 100);
	return 2;
} /* ScanLeftGain */
/*-------------------------------------------------------------------------*/
static int ScanRightGain (Tcl_Interp *Interp, int argc, char *argv[])
{
	int			Val = -1;

	if (argc <= 1) {
	  Interp->result = "Missing option, to -rightgain";
	  return -1;
	  }
	Val = atoi (argv[1]);
	if (Val < 0)
	  Val = 0;
	else if (Val > 100)
	  Val = 100;
	ahw_set (AD_RECORD, AS_gain_right, 256 * Val / 100);
	return 2;
} /* ScanRightGain */
/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
char cmd_audioctl (ClientData *cd, Tcl_Interp *Interp, int argc, char *argv[])
/*
 * Implements the new tcl/tk audio command:
 * audioctl <option> <value> [...]
 * 	-input		line1, line2, line3, lowgainmike, highgainmike,
 * 	-gain		0..100 (sets left and right gain)
 *	-leftgain	0..100
 *	-rightgain	0..100
 *
 * 	-output		line1, line2, line3
 *	-intspeaker	on, off, 0, 1
 * 	-volume		0..100
 * 	-balance	-100..0..100
 */
{
	int			n;

	if (argc <= 1) {
	  Interp->result = "Usage: audio ...";
	  return TCL_ERROR;
	  }
	argc--, argv++;
	while (argc > 0) {
	  if ((n = ScanOption (Interp, argc, argv)) <= 0)
	    return TCL_ERROR;
	  argc -= n;
	  argv += n;
	  } /* while */
	return TCL_OK;
} /* cmd_audioctl */
