# Do not add this file to tclIndex with auto_mkindex as it will be
# loaded twice otherwise.
#
# $Id: nevot.tcl,v 1.3 1998/07/30 11:48:11 dor Exp $
#

proc usage {} {
  global argv0 errorFatal

  set errorFatal 1
  return "Usage: $argv0 \[-p tcp/localhost/port | udp/MC-address/port\]"
}

proc nevot {} {
  global env
  global nevot_library vu_size msg
  global bg argv0
  global controller

  wm title      . "NeVoT [infon version]@[exec uname -n]"
  wm iconbitmap . nevot_icon
  wm iconmask   . nevot_mask
  wm iconname   . "NeVoT"

  # delete old ones
  catch {destroy .base .msg}

  # get background color
  set bg [lindex [. configure -background] 4]

  # overall base panel
  frame .base
  frame .base.1 -borderwidth 10
  frame .base.2 -borderwidth 10

  # Join/Settings
  pack \
    [menubutton .base.settings -text "Settings" -padx 4 \
     -menu .base.settings.menu -underline 0] \
    [button .base.quit   -text "Quit"   -padx 4 -command nevot_exit] \
    -in .base.1 -side left -expand true -fill both

  set m [menu .base.settings.menu -tearoff 0]
  $m add command -label "local audio" \
       -command audio_panel
  $m add command -label "automatic gain control" \
       -command agc_panel
  $m add command -label "echo suppression" \
       -command echo_panel
  $m add command -label "silence detection" \
       -command silence_panel
  $m add command -label "volume meter" \
       -command vu_panel
  $m add command -label "timeouts" \
       -command timeout_panel
  $m add command -label "debugging" \
       -command debug_panel
#  $m add command -label "audio control" \
#       -command audioctl_panel

  # VU meters
  pack \
    [canvas .base.vu -width 100 -height 40] \
    -in .base.2 -side top -pady 4 -anchor w -fill x

  .base.vu create line 0  5 0  5 -width 10 -fill green  -tag in_green 
  .base.vu create line 0  5 0  5 -width 10 -fill yellow -tag in_yellow
  .base.vu create line 0  5 0  5 -width 10 -fill red    -tag in_red
  .base.vu create line 0 35 0 35 -width 10 -fill green  -tag out_green 
  .base.vu create line 0 35 0 35 -width 10 -fill yellow -tag out_yellow
  .base.vu create line 0 35 0 35 -width 10 -fill red    -tag out_red
  .base.vu create line 0  0 0  0 -width  1              -tag max

  bind .base.vu <Configure> {vu_resize %w}

  # frames
  pack .base.1 .base.2 -side top -expand true -fill both
  pack .base -side top -fill both

  # determine dimensions
  tkwait visibility .base
  bind . <Destroy> {nevot_exit}

  update

  # constrain to changing width
  wm minsize    . [winfo width .] [winfo height .]
  wm maxsize    . [lindex [wm maxsize .] 0] [winfo height .] 
}

proc vu_resize {size} {
  global vu_size

  set vu_size $size

  # delete old text
  .base.vu delete labels

  # print numeric labels on VU meter
  foreach dB {{60 0.05} 
              {50 0.12}
              {40 0.24}
              {30 0.36}
              {20 0.48}
              {10 0.60}
              { 7 0.71}
              { 4 0.82}
              { 2 0.90}
              { 0 0.97}} {
    .base.vu create text [expr [lindex $dB 1]*$vu_size] 20 \
      -anchor c -justify c -font *helvetica-*-r-*-*-12* \
      -text [lindex $dB 0] -tag labels
  }
}

#
# Called to terminate application. Only act when root window is closed.
#
proc nevot_exit {} {
  global env
    # close all streams
    terminate

    # let others know we are gone
    control_send "agent NeVoT[infon version] detached [pid]"

    # remove .nevotpid
    catch {exec rm $env(HOME)/.nevotpid}
  exit  
}

set auto_path "$nevot_library $auto_path"

# check
if {![info exists nevot_library]} {
  error "Environment variable NEVOT_LIBRARY is undefined."
  nevot_exit
}

#
# Tk error handler
#
proc tkerror err {
  global tk_library errorInfo

  if [info exists tk_library] {tkerror.nevot $err} else {error $errorInfo}
}

#
# set variables from .Xdefaults/RESOURCE_MANAGER
#

# Variable to make options in script /dev/null available which was started to aviod interactive modus
set top Null  

option add $top.debug      ""        startupFile
option add $top.debugfile  nevot.log startupFile

set debug      [option get . debug      Nevot]
set debugfile  [option get . debugfile  Nevot]
# symbolic font names
set helvetica10  "-*-helvetica-medium-r-normal--*-100-*"
set helvetica10b "-*-helvetica-bold-r-normal--*-100-*"
  global argv0 argv argc
set helvetica10o "-*-helvetica-bold-o-normal--*-100-*"
set helvetica12  "-*-helvetica-medium-r-normal--*-120-*"
set helvetica12b "-*-helvetica-bold-r-normal--*-120-*"
set helvetica14  "-*-helvetica-medium-r-normal--*-140-*"
set helvetica14b "-*-helvetica-bold-r-normal--*-140-*"
set times14      "-*-times-medium-r-normal--*-140-*"

# X options
option add $top.*background lightgray startupFile
option add $top.*foreground black startupFile
option add $top.*activeBackground gray50 startupFile
option add $top.*activeForeground white startupFile
option add $top.*selectBackground gray95 startupFile
option add $top.*Font $helvetica14b startupFile

# definitions and defaults
nevot_defs

if {[catch {source ~/.nevotrc-$env(HOST)}]} {
  catch {source ~/.nevotrc}
}

# initialize controller
control_init $controller
control_send "+ all/all/all"
control_send "+ session */audio/*"
control_send "agent NeVoT[infon version] attached [pid]"

# start the activity signals
proc update_active { } {
    active_all
    after 10000 update_active
}

update_active



# start up base window
. configure -bg [option get . background Background] 
nevot
