/*
 * $Id: ahw.h,v 1.1.1.1 1997/11/12 17:16:23 dor Exp $
 *
 * $Log: ahw.h,v $
 * Revision 1.1.1.1  1997/11/12 17:16:23  dor
 *
 * Revision 1.5  1995/08/30  11:12:52  hgs
 * 3.32
 *
 * Revision 1.4  1995/07/12  12:02:36  hgs
 * 3.30
 *
 * Revision 1.3  1995/06/24  10:42:58  hgs
 * 3.30
 * AIX additions.
 *
 * Revision 1.2  1995/06/22  06:59:19  hgs
 * Additional ports for AIX.
 * audio_dev_t definition.
 *
 */

#ifndef _ahw_h
#define _ahw_h
#include "audio_descr.h"

/*
Hardware-independent interface to low-level audio.

int ahw_open(int ndelay)
  Open audio hardware. Return file descriptor number to be used in
  select() for audio input (record).

int ahw_close(void)
  Close audio hardware in the given direction. Return file descriptor
  number for audio input (record). Don't close record mode if play
  mode is still on (as the latter can't work without the former).

int ahw_set()
  Set audio device characteristics from audio_state.
  To set initial values, this should be called before ahw_open().
  ahw_open() does not be called before using this routine.
  Return -1 on failure, 0 on success.
  Examples:
    To set input to line level:
      ahw_set(0, AS_port, AP_in_line);
    To set output port to line-level and speaker:
      ahw_set(1, AS_port, AP_out_line|AP_out_speaker);
    To set input gain (legal values are between 0 and 255):
      ahw_set(0, AS_gain, 100);
    To set output gain (legal values are between 0 and 255):
      ahw_set(1, AS_gain, 100);

int ahw_fmt(audio_descr_t *rec, audio_descr_t *play)
  Given the audio formats 'rec' and 'play', set the corresponding
  device parameters to the closest match. 'rec' and 'play' are
  modified to the actual device setting. Some systems require that
  'rec' and 'play' are the same; in that case, 'rec' takes precedence.
  Returns 0 on success, -1 on error.

int ahw_get()
  Obtain current hardware audio parameters and update error counts.
  Return -1 on failure, 0 on success.

int ahw_write(void *buffer, u_int32 framecount);
  Play 'framecount' audio frames. Return number of frames written
  (should be equal to framecount).

int ahw_read(void *buffer, u_int32 framecount);
  Record 'framecount' audio frames into buffer. The number of frames
  read may be less than that requested to avoid blocking.
  Returns number of frames actually read.

int ahw_write_queue()
  Return the number of frames currently in the play queue. A negative
  number indicates that a buffer overflow has occurred.

int ahw_read_queue()
  Return the number of frames recorded, but not yet read by audio_rec().
  A negative number indicates that a buffer underflow has occurred.

int ahw_waiting(void)
  Return -1 if error, 0 if nobody waiting, 1 if other process waiting.

void ahw_track(func)
  Establish call-back function to call when device parameters are
  changed by some other application.  Few systems actually support
  this mechanism.

int ahw_sync(void)
  Only used for DEC Audiofile, returns difference between audioclock and
  and current system timestamp, in milliseconds. This equals the time

$Id: ahw.h,v 1.1.1.1 1997/11/12 17:16:23 dor Exp $
*/

typedef void (*signal_func_t)(int);

typedef enum {              /* commands passed to ahw_set */
  AS_gain = 3,              /* set left and right gain/vol */
  AS_port,
  AS_fillpoint,
  AS_gain_step,
  AS_gain_left,             /* sets left gain/vol */
  AS_gain_right,            /* sets right gain/vol */
  AS_balance                /* sets balance */
} audio_set_t;

/*** definition for the various input and output ports ***/
typedef enum {
  AP_default       =    0,
  /* ouput: several active concurrently (bitmask) */
  AP_out_line      =    1,  /* for compatibility */
  AP_out_line_1    =    1,
  AP_out_line_2    =    2,
  AP_out_line_3    =    4,
  AP_out_headphone =    8,
  AP_out_digital   = 0x10,
  AP_out_speaker   = 0x20,
  AP_out_max       = 0x20,
  /* input: values (only one active at a time) */
  AP_in_line            =    1,  /* for compatibility */
  AP_in_line_1          =    1,
  AP_in_line_2          =    2,
  AP_in_line_3          =    4,
  AP_in_microphone      =    8,  /* for compatibility */
  AP_in_microphone_low  = 0x10,
  AP_in_microphone_high = 0x20,
  AP_in_digital         = 0x40,
  AP_in_max             = 0x40
} audio_port_t;

/*** definitions for record / playback / monitor ***/
typedef enum {
  AD_RECORD   =  0,
  AD_PLAYBACK =  1,
  AD_MONITOR  =  2
} audio_dev_t;

/* workstation audio I/O descriptor */
typedef struct {
  audio_descr_t ad;  /* generic audio description */
  short gain[2];     /* left/right gain level: 0-255 */
  int port;          /* current port (bitmask) */
  int fillpoint;     /* select() threshold */
  int buffer_size;   /* audio buffer size */
  int samples;       /* number of samples read/written */
  int errors;        /* overflow/underflow count */
  int waiting;       /* somebody else waiting for audio device */
  char active;       /* audio device is active */
} audio_ws_t;

typedef struct {
  audio_ws_t play, record;
  short monitor_gain[2]; /* per-channel monitor gain */
  int track;             /* flag: track changes to gain etc. by other processes */
} audio_state_t;

extern int  ahw_get(audio_state_t *);
extern int  ahw_open(int ndelay);
extern int  ahw_close(void);
extern int  ahw_fmt(audio_descr_t *record, audio_descr_t *play);
extern void ahw_track(signal_func_t func);
extern int  ahw_write(char *buffer, int framecount);
extern int  ahw_read(char *buffer, int framecount);
extern int  ahw_set(int pr, audio_set_t p, int value);
extern int  ahw_write_queue(int *errors);
extern int  ahw_write_flush(void);
extern int  ahw_read_queue(int *errors);
extern int  ahw_waiting(void);

extern int ahw_changed;  /* flag: audio device has changed */
extern int ahw_errno;    /* error indication */
#endif
